/*
 * ----------------------------------------------------------------------------------
 * dc_pcm.c
 *
 * Functions to display the memory snapshots slots
 * 
 * Copyright (C) Kroustibat
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 */


#define _arch_dreamcast
#include <kos.h>
#include "pcx.h"
#include "dc_snap.h"


char * my_strrchr(const char * s, int c)
{
       const char *p = s + strlen(s);
       do {
           if (*p == (char)c)
               return (char *)p;
       } while (--p >= s);
       return NULL;
}


//-----------------------------------------------------------------------------
// Adapted from Allegro library, credits goes to Shawn

int LoadPcx2Buffer(char* filename,unsigned short *buffer)
{
	unsigned int	fd;
	pcx_hdr			*pcx_header;
	unsigned char	pcx_file[0x10000];
	unsigned char   *pcx;
	int	j;
	unsigned char	byte;
	int w,h;
	int i,runlen,cnt,size;
	uint16	pcx_pal[256];		
	uint8	*data=NULL;
	int		x,xx,y,po;
	int     bytes_per_line=0;
	unsigned char bpp ;

	cnt=0;

	//-- check if a pcx --
	if (stricmp(my_strrchr(filename, '.'), ".pcx") != 0)
	{	return 0;}

	//-- open file --
	fd=fs_open(filename, O_RDONLY);
	if(fd==0)
	{	return 0;}

	//-- Get the length of the file --
	size=fs_total(fd);	
	if(size>0x10000) return 0;		

	//-- read header --
	pcx= (unsigned char*) pcx_file;
	if(!pcx)
	{	fs_close ( fd );
		return 0;
	}	

	if(fs_read(fd,pcx, size)!=size)
	{	fs_close ( fd );
		return(0);	
	}

	fs_close ( fd );

	pcx_header = (pcx_hdr*) (pcx);
	
	w=pcx_header->Xmax+1;
	h=pcx_header->Ymax+1;
	bytes_per_line=pcx_header->Bpl;
	bpp== pcx_header->ClrPlanes *8 ;
	
	// allocate memory
	//-- check of attributes --
	if(pcx_header->Bpp!=8 || !(pcx_header->ClrPlanes==1 || pcx_header->ClrPlanes==3)  )
	{	return 0;}	
	
	if(pcx_header->ClrPlanes==1)
		data=malloc(w*h*pcx_header->ClrPlanes);
	
	
	//-- skip header --
	pcx= pcx+ sizeof(pcx_hdr);

	
	// RLE decoding
    for(y=0;y<h;y++)
    {   x=xx=0;
        po = 0;
        while (x < (bytes_per_line*pcx_header->ClrPlanes))
        {
            byte = *pcx++;
            if ((byte & 0xC0) == 0xC0)
            {   runlen = (byte & 0x3F);
                byte = *pcx++;
            }
            else
            {   runlen=1;
            }

            if(pcx_header->ClrPlanes==1)
            {   while (runlen--)
                {   if (x < w)
                        data[(y*w)+x] = byte;
                    x++;
	            }
            }
            else
            {
                while (runlen--)
	 	        {
                    // red
	    	        if (xx < w)
                    {
                        switch(po)
                        {
                            case 0:
                            {   buffer[(y*w)+(xx)] = ((byte>>3) & 0x1f)<<10;
                            }break;

                            case 1:
                            {   buffer[(y*w)+(xx)] |=  (byte>>3<<5);
                            }break;

                            case 2:
                            {   buffer[(y*w)+(xx)] |= ((byte>>3) & 0x1f);

								if(buffer[(y*w)+(xx)]!= 0x7c1f)
								{	buffer[(y*w)+(xx)] |=0x8000;									
								}

                            }break;
                        }
                    }

        	        x++;
        	        if (x == bytes_per_line)
	                {
                        xx = 0;
            		  	po = 1;
        	        }
	                else
        	        {
                        if (x == bytes_per_line*2)
            	      	{   xx = 0;
		  	    	        po = 2;
        	            }
	                    else
		  	                xx++;
                    }
                }
	        }
        }
    }


	if(pcx_header->ClrPlanes ==1)
	{
		//-- Load the palette --
		pcx++;

		//-- load the pal --
		for (i=0; i<256; i++)
		{	pcx_pal[i]= ARGB1555(1,pcx[i*3+0],pcx[i*3+1],pcx[i*3+2]);
		}

		// convertit
		for(i=0;i<h;i++)
		{	
			for(j=0;j<w;j++)
				buffer[(i*w)+j]= data[(i*w)+j]!=0 ? pcx_pal[data[(i*w)+j]] : 0;
		}
	}

	if(data)
		free(data);

	//-- close the pcx --
	return 1;

}


int LoadPcx2BufferNoMask(char* filename,unsigned short *buffer)
{
	unsigned int	fd;
	pcx_hdr			*pcx_header;
	unsigned char	pcx_file[0x20000];
	unsigned char   *pcx;
	int	j;
	unsigned char	byte;
	int w,h;
	int i,runlen,cnt,size;
	uint16	pcx_pal[256];		
	uint8	*data=NULL;
	int		x,xx,y,po;
	int     bytes_per_line=0;
	unsigned char bpp ;

	cnt=0;

	//-- check if a pcx --
	if (stricmp(my_strrchr(filename, '.'), ".pcx") != 0)
	{	return 0;}

	//-- open file --
	fd=fs_open(filename, O_RDONLY);
	if(fd==0)
	{	return 0;}

	//-- Get the length of the file --
	size=fs_total(fd);	
	if(size>0x20000) return 0;		

	//-- read header --
	pcx= (unsigned char*) pcx_file;
	if(!pcx)
	{	fs_close ( fd );
		return 0;
	}	

	if(fs_read(fd,pcx, size)!=size)
	{	fs_close ( fd );
		return(0);	
	}

	fs_close ( fd );

	pcx_header = (pcx_hdr*) (pcx);
	
	w=pcx_header->Xmax+1;
	h=pcx_header->Ymax+1;
	bytes_per_line=pcx_header->Bpl;
	bpp== pcx_header->ClrPlanes *8 ;

	if(h>192 || w>256) 
		return 0;
	
	// allocate memory
	//-- check of attributes --
	if(pcx_header->Bpp!=8 || !(pcx_header->ClrPlanes==1 || pcx_header->ClrPlanes==3)  )
	{	return 0;}	
	
	if(pcx_header->ClrPlanes==1)
		data=malloc(w*h*pcx_header->ClrPlanes);
	
	
	//-- skip header --
	pcx= pcx+ sizeof(pcx_hdr);

	
	// RLE decoding
    for(y=0;y<h;y++)
    {   x=xx=0;
        po = 0;
        while (x < (bytes_per_line*pcx_header->ClrPlanes))
        {
            byte = *pcx++;
            if ((byte & 0xC0) == 0xC0)
            {   runlen = (byte & 0x3F);
                byte = *pcx++;
            }
            else
            {   runlen=1;
            }

            if(pcx_header->ClrPlanes==1)
            {   while (runlen--)
                {   if (x < w)
                        data[(y*w)+x] = byte;
                    x++;
	            }
            }
            else
            {
                while (runlen--)
	 	        {
                    // red
	    	        if (xx < w)
                    {
                        switch(po)
                        {
                            case 0:
                            {   buffer[(y*256)+(xx)] = ((byte>>3) & 0x1f)<<10;
                            }break;

                            case 1:
                            {   buffer[(y*256)+(xx)] |=  (byte>>3<<5);
                            }break;

                            case 2:
                            {   buffer[(y*256)+(xx)] |= ((byte>>3) & 0x1f);
								buffer[(y*256)+(xx)] |=0x8000;									

								/*if(buffer[(y*w)+(xx)]!= 0x7c1f)
								{	buffer[(y*w)+(xx)] |=0x8000;									
								}*/

                            }break;
                        }
                    }

        	        x++;
        	        if (x == bytes_per_line)
	                {
                        xx = 0;
            		  	po = 1;
        	        }
	                else
        	        {
                        if (x == bytes_per_line*2)
            	      	{   xx = 0;
		  	    	        po = 2;
        	            }
	                    else
		  	                xx++;
                    }
                }
	        }
        }
    }


	if(pcx_header->ClrPlanes ==1)
	{
		//-- Load the palette --
		pcx++;

		//-- load the pal --
		for (i=0; i<256; i++)
		{	pcx_pal[i]= ARGB1555(1,pcx[i*3+0],pcx[i*3+1],pcx[i*3+2]);
		}

		// convertit
		for(i=0;i<h;i++)
		{	
			for(j=0;j<w;j++)
				buffer[(i*256)+j]= pcx_pal[data[(i*w)+j]];
			//	buffer[(i*w)+j]= data[(i*w)+j]!=0 ? pcx_pal[data[(i*w)+j]] : 0;
		}
	}

	if(data)
		free(data);

	//-- close the pcx --
	return 1;

}
