/*
 * ----------------------------------------------------------------------------------
 * dc_pcm.c
 *
 * Functions to display the memory snapshots slots
 * 
 * Copyright (C) Takayama Fumihiko
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 */

#define _arch_dreamcast
#include <kos.h>

//#include <shared.h>
#include <dc/g2bus.h>
#include "dc_sound.h"

#define ARM_OFFSET 0xa0800000
static volatile uint32 *start  = (uint32*)(0x10000 + ARM_OFFSET);
static volatile uint32 *init_done = (uint32*)(0x10004 + ARM_OFFSET);
static volatile uint32 *play_start = (uint32*)(0x10008 + ARM_OFFSET);
static volatile uint32 *rate = (uint32*)(0x10010 + ARM_OFFSET);
static volatile uint32 *bits = (uint32*)(0x10014 + ARM_OFFSET);
static volatile uint32 *buf_len = (uint32*)(0x10018 + ARM_OFFSET);
static volatile uint32 *position  = (uint32*)(0x10020 + ARM_OFFSET);

#define BASE_ADDRESS 0x11000

static const uint8 stream_driver[] = {
#include "sound/aica_fw.h"
};


void
dc_sound_init (int sample_rate, int sample_bits, int buffer_len)
{
  *init_done = 0; /* need before spu_memload */
  g2_fifo_wait ();
  
  spu_disable();
  spu_memload (0, stream_driver, sizeof(stream_driver));
  spu_enable ();
  
  *rate = sample_rate;
  *bits = sample_bits;
  *buf_len = buffer_len;
  g2_fifo_wait ();
  
  for (;;)
  {
    if (*init_done) break;
    timer_spin_sleep (100);
  }
  
  *start = true;
  g2_fifo_wait ();
  
  for (;;)
  {
    if (*play_start) break;
    timer_spin_sleep (100);
  }

}


void
dc_sound_shutdown (void)
{
  spu_disable();
}


uint32
dc_sound_get_baseaddr (void)
{
  return BASE_ADDRESS;
}


uint32 
dc_sound_get_position (void)
{
  uint32 pos = *position;
  g2_fifo_wait ();
  return pos;
}


