/*
 * ----------------------------------------------------------------------------------
 * speccyal.c
 * 
 * Copyright (C) 2001 Stephane Schmitz
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 * ----------------------------------------------------------------------------------
 * Version  	: 1.0
 * Author	: S.Schmitz
 * History  	: 	03/05/2003, V0.1
 *			- First release
 * Date		: 03/05/2003
 * ----------------------------------------------------------------------------------
 */

#define _arch_dreamcast
#include <kos.h>
#include <stdio.h>
#include "file_dialog.h"
#include "snaps.h"
#include "dc_pcm.h"
#include "menu.h"
#include "ay8910.h"
#include "speccyal.h"
#include "screen.h"
#include "slots.h"
#include "z80_mame.h"
#include "osd_cpu.h"
#include "tapreader.h"

char msg2[256];

t_snd snd;

extern int g_b48kmode;
extern char g_szSnapFileName[];

#define GETVIDEOPAGE()		(GetMemPtrEx()).videoPage
#define POKE(addr, b)		(GetMemPtrEx()).MemPoke (addr,b)

//void writeprt (unsigned short adr,byte data);
//unsigned char readprt (unsigned short adr);
int irq_callback(int param);

__inline void DrawScreen_full (void);
__inline void DrawScreen (void);
__inline void DrawLine (int line);
__inline void DrawLine3 (int skip);
static void DrawHorizLine (uint16 *screen, unsigned int x, unsigned int y, unsigned int length, unsigned int bg);
static void sound_callback ();
unsigned short *CreateSpectrumPalette (void);
void clear_sound_buffer (void);

static void RectFill2 (uint16 *screen, unsigned int x, unsigned int y, unsigned int x2, unsigned int y2, unsigned int bg);

unsigned short int PixelAddressArray[256*192];
unsigned short int *ptrPixelAddressArray = &PixelAddressArray[0];
unsigned short int AttribArray[256*192];
unsigned short int *ptrAttribArray = &AttribArray[0];
int nflip, fps=0;
unsigned char	border[312], border2[312];

// Sound related
// Comment the following line to render sound in 22 khz

#define 	SOUND44100	1

#ifdef SOUND44100
#define 	FREQUENCY	44100
#else
#define 	FREQUENCY	22050
#endif

#define	RATE		50
#define	BUFFERLEN	(FREQUENCY/RATE)

static unsigned short Buffer1[BUFFERLEN], Buffer2[BUFFERLEN], Buffer3[BUFFERLEN], Buffer4[BUFFERLEN], buffmix[BUFFERLEN];
unsigned int g_bEarBitOn;
unsigned int g_unBeeperBufferPartPos = 0;
unsigned short *buf[3];
void RenderBeeper44100 ();
void RenderBeeper22050 ();
void RenderBeeper2 ();
unsigned int g_bWrite8910 = 0;

// Misc.

char virtual_key = 0;
unsigned int ntime = 600;
unsigned int bLoadCommand = 0, step = 0;
unsigned char ucFePortValue;
void set_options (void);
void selectKey (unsigned char *kBuffer, char key);

void cpu_writemem16(unsigned short address, unsigned char data);

struct options emu_options;

int pauseGraphics;

unsigned short *palEntries;

extern uint16 vbuffer[] __attribute__ ((aligned (32)));

myRECT      rc;

// -------------------------------
// Main emulation loop
// -------------------------------

int main(int argc, char **argv)
{
	int i,j,x,y,bFileLoaded=0;
	uint32 n_tick, ticks, n_wait_tick, n_tick_virtual_kb, towait, frames=0, tmpTicks;
	int cnt_surface;
	char outpath[256] = "/cd/", inpath[256];
	char *slash_pos;
	int ExKey;
	int skip = 0;
	char buff[256];
	char slotFileName[10];
	uint16	ret;
	cont_cond_t cond;
	uint8	down_timer=0;
	uint32 secs, msecs;
	int bExit = 0, length;
	char *pszFileExtension;
	char *pszSlashPosition;
	char pszShortDesc[128];

	strcpy (g_szSnapFileName, "");

	rc.w = 2*512;
	rc.h = 2*512;
	rc.x = 0;
	rc.y = 0;
	
	//tape_init ();
	setup_screen ();
    
	// Loads roms, init spectrum 48k / 128k memory map
	
	if (Init())
	{

		// Prepare the spectrum screen pixel coordinates arrays
	
		for (y=0;y < 192;y++)
		{
			x=0;
			while (x < 256)
			{
				// Computes spectrum pixel address (8 bits groups)
	
				PixelAddressArray[(y<<8)+x] = ((y>>6)*2048) + (((y>>3) & 0x07)<<5) + ((y & 0x07)<<8) + ((x>>3) & 0x1f);
				AttribArray[(y<<8)+x] = (ATTR_LOW-0x4000) + ((x>>3)&0x1f) + ((y>>3)<<5);
				x++;
			}
		}
		
		load_options();
		set_options();
	
		palEntries = CreateSpectrumPalette();
		
		// Init mame z80 core
	
		z80_reset (NULL);
	
		z80_set_irq_callback(irq_callback);
	
		AY8910_init(0,1774000/4,0xaa,FREQUENCY,NULL,NULL,NULL,NULL);

		// Sound init

		snd.bufsize = BUFFERLEN;
		pcm_init();

		buf[0] = &Buffer2[0];
		buf[1] = &Buffer3[0];
		buf[2] = &Buffer4[0];

		timer_ms_gettime(&secs, &n_tick_virtual_kb);
		n_tick = n_tick_virtual_kb;

		while(1)
		{
			timer_ms_gettime(&secs, &n_wait_tick);
						
			z80_set_irq_line (0x00, CLEAR_LINE);

			if (emu_options.rendering == 1)
			{
				for (j = 0;j < 312; j++)
				{
					z80_execute(224);

					#ifdef SOUND44100
					RenderBeeper44100 ();
					#else
					RenderBeeper22050 ();
					#endif
					*(border+j) = ucFePortValue;
				}
			}
			
			g_unBeeperBufferPartPos = 0;
			
			if (emu_options.rendering == 0) 
			{
				DrawLine3 (skip);
				sound_callback();
			}
			else
			{
				if (emu_options.fullscreen == 1)
				{
					DrawScreen_full();
				}
				else				
				if (emu_options.frameskip == 1)
				{
					frames++;
					if (skip == 0) 
					{	
						DrawScreen();
						skip++;
					}
					else
					{
						skip = 0;
					}
				}
				else
				{
					DrawScreen();
				}
				sound_callback();
			}

			z80_set_irq_line (0x00, ASSERT_LINE);

			// FPS limiter

			timer_ms_gettime(&secs, &ticks);
			if (ticks - n_wait_tick < 20)
			{
				towait = 20 - (ticks - n_wait_tick); 
				timer_ms_gettime(&secs, &n_wait_tick);

				bExit = 0;
				while (!bExit)
				{
					timer_ms_gettime(&secs, &tmpTicks);
					if (tmpTicks - n_wait_tick >= towait) bExit = 1;
				}
			}

			if(cont_get_cond(maple_first_controller(), &cond) <0) break;
	
			// Save snapshot  = START + left trigger 
		
			if ( (!(cond.buttons & CONT_START)) && (cond.ltrig))
			{
				pcm_close();
				ret=displaySlots (1, 4, 7, 0);
				if ( ret != 0xffff) 
				{
					// Build the name of the file to save on VMU ...

					sprintf (&buff[0], "%d", (ret >= 4) ? ret-4 : ret);
					strcpy (slotFileName, "spkslot");
					strcat (slotFileName, buff);
					strcat (slotFileName, ".mem");

					// Get the string to put in the savefile short description 
					// (get the currently loaded file and remove the path as well as the extension)

					if (g_szSnapFileName != NULL && strlen (g_szSnapFileName) > 0)
					{
						pszFileExtension = (char *) (strrchr(g_szSnapFileName, '.') + 1);
						pszSlashPosition = (char *) (strrchr(g_szSnapFileName, '/') + 1);
						length = pszFileExtension - pszSlashPosition - 1;
						strncpy (pszShortDesc, pszSlashPosition, length);
						pszShortDesc[length] = '\0';
					}
					else
					{
						strcpy (pszShortDesc, "");
					}

					if (ret < 4)
					{
						// save
						
						SaveSnapShot (slotFileName, pszShortDesc);
					}
					else
					{
						// load
						LoadFileFromVmu (slotFileName);
					}
				}

				pcm_init();
			}

			// File browser = START + right trigger

			else if ( (!(cond.buttons & CONT_START)) && (cond.rtrig))
			{
				pcm_close();

				if (strlen (outpath) > 0)
				{
					strcpy (inpath, outpath);
					slash_pos = (char *)strrchr (inpath, '/');
					if (slash_pos != NULL) *slash_pos = '\0';
				}

				if (SelectFileDialog(inpath, 0, outpath) == 0) 
				{
					LoadFile (outpath);
					load_game_keys (outpath); // loads a key files, if present
					bFileLoaded = 1;
				}

				set_options(); // because rc (depending on the fullscreen option) has to be reinitialized
				pcm_init();
			}			

			// Reset = START + Y

			else if ( !(cond.buttons & CONT_START) && !(cond.buttons & CONT_Y) ) 
			{
				strcpy (g_szSnapFileName, "");
				g_bWrite8910 = 0;
				clear_sound_buffer();
				InitPaging();
				z80_set_reg (Z80_PC, 0x0000);
				bFileLoaded = 0;
			}

			// Tape rewind = START + B

			else if ( !(cond.buttons & CONT_START) && !(cond.buttons & CONT_B) ) 
			{
				tape_rewind();
			}

			// Quit

			else if ( (cond.ltrig) && (cond.rtrig) )
			{
				return;
			}

			// Virtual keyboard = START + RIGHT trig

			else if (cond.ltrig) 
			{
				pcm_close();
				virtual_key = display_virtual_keyboard2 (1, 4);
				timer_ms_gettime(&secs, &n_tick_virtual_kb);
				ntime = 600;
				pcm_init();
				set_options(); // because rc (depending on the fullscreen option) has to be reinitialized
				/*
				virtual_key = 'J';
				bLoadCommand = 1;
				*/
			}

			// Options screen = RIGHT trigger

			else if ( cond.rtrig ) 
			{
				pcm_close();
				
				if (draw_options (1, 4, 7) == 1) 
				{
					save_options();
					set_options();
					
					// Only save the custom keys if a file has been previously loaded
					
					if (bFileLoaded == 1) 
					{
						// only save keys (to save blocks) if joystick emyulation = custom keys

						if (emu_options.joy == CUSTOM_KEYS)
						{
							save_game_keys (outpath); // per-game key profile
						}
					}
				}
				else
				{
					set_options();
				}
				
				pcm_init();

			}

			// Limit the time a virtual key is buffered ...

			timer_ms_gettime(&secs, &tmpTicks);
			if (tmpTicks - n_tick_virtual_kb >= ntime)
			{
				if (bLoadCommand != 1) virtual_key = 0;
				else
				{
					// Simulate the 3 key presses (J P P = LOAD ""), one key press each 200 ms

					if (step == 0)
					{
						bLoadCommand = 0;
						ntime = 600;
					}
					
					/*if (step == 0)
					{
						virtual_key = 'P';
						step++;
					}
					else if (step == 1)
					{
						virtual_key = 'P';
						if (emu_options.frameskip == 0) ntime = 600; else ntime = 1200;
						step++;
					}
					else if (step == 2)
					{
						virtual_key = 3; // enter
						step++;
					}				
					else if (step == 3)
					{
						bLoadCommand = 0;
						ntime = 600;
						step = 0;
					}
					*/
				}
				timer_ms_gettime(&secs, &n_tick_virtual_kb);
			}
	
		}
	
		// Release 48k/128k allocated memory
	
		Release();
		
	}
}

static void sound_callback ()
{
	int i = 0, j = 0;
	unsigned short mix = 0;

	//int size = 44100/60;

	if (g_bWrite8910) AY8910Update(0, buf, snd.bufsize);

	for (i=0; i<BUFFERLEN; i++)
	{
		mix = Buffer1[i];
		
		if (g_bWrite8910) 
		{
			mix = Buffer2[i] + Buffer3[i] + Buffer4[i];
		}
		if (mix > 65535) mix = 65535;
		mix = mix >> 2; // to avoid clicking

		buffmix[i] = (unsigned short)mix;
	}

	// for (i=0; i<882; i++) buffmix[i] = Buffer1[i] >> 2 + Buffer2[i] >> 2 + Buffer3[i] >> 2 + Buffer4[i] >> 2;

	memcpy (&snd.psg_buffer[0][0], buffmix, BUFFERLEN*2); // *2 because buffer contains words = 2 bytes
	render_pcm_ntsc();
}

void clear_sound_buffer (void)
{
	memset (Buffer1, 0, BUFFERLEN*2); // *2 because buffer contains words = 2 bytes
	memset (Buffer2, 0, BUFFERLEN*2);
	memset (Buffer3, 0, BUFFERLEN*2);
	memset (Buffer4, 0, BUFFERLEN*2);
	memset (buffmix, 0, BUFFERLEN*2);
}

void set_options (void)
{
	if (get_option_state (MACHINE48K) == OPT_STATE_ON) 
	{
		emu_options.machine = MACHINE48K;	
		g_b48kmode = 1;
	}
	else
	{
		emu_options.machine = MACHINE128K;
		g_b48kmode = 0;
	}

	if (get_option_state (BORDER) == OPT_STATE_ON) 
	{
		emu_options.border = 1;	
	}
	else
	{
		emu_options.border = 0;
	}

	if (get_option_state (FPS) == OPT_STATE_ON) 
	{
		emu_options.fps = 1;	
	}
	else
	{
		emu_options.fps = 0;
	}

	if (get_option_state (RENDERING) == OPT_STATE_ON) 

	{
		emu_options.rendering = 1;	
	}
	else
	{
		emu_options.rendering = 0;
	}

	if (get_option_state (FRAMESKIP) == OPT_STATE_ON) 

	{
		emu_options.frameskip = 1;	
	}
	else
	{
		emu_options.frameskip = 0;
	}

	if (get_option_state (FILTERING) == OPT_STATE_ON) 
	{
		emu_options.filtering = 1;	
	}
	else
	{
		emu_options.filtering = 0;
	}

	emu_options.k_up = (char)get_option_state (KEY_UP);
	emu_options.k_down = (char)get_option_state (KEY_DOWN);
	emu_options.k_left = (char)get_option_state (KEY_LEFT);
	emu_options.k_right = (char)get_option_state (KEY_RIGHT);
	emu_options.k_fire1 = (char)get_option_state (KEY_FIREA);
	emu_options.k_fire2 = (char)get_option_state (KEY_FIREB);

	if (get_option_state (SINCLAIR1) == OPT_STATE_ON) 
	{
		emu_options.joy = SINCLAIR1;	
	}
	else if (get_option_state (SINCLAIR2) == OPT_STATE_ON)
	{
		emu_options.joy = SINCLAIR2;
	}
	else if (get_option_state (CURSOR) == OPT_STATE_ON)
	{
		emu_options.joy = CURSOR;
	}
	else
	{
		emu_options.joy = CUSTOM_KEYS;
	}
	
	if (get_option_state (FULLSCREEN) == OPT_STATE_ON) 
	{
		rc.w = 2.5*512;
		rc.h = 2.5*512;
		rc.x = 0;
		rc.y = 0;
		emu_options.fullscreen = 1;	
	}
	else
	{
		rc.w = 2*512;
		rc.h = 2*512;
		rc.x = 0;
		rc.y = 0;
		emu_options.fullscreen = 0;
	}	
}

void RenderBeeper44100 ()
{
	unsigned long	unBufferPos=0;
	static char	cCounter=0;
	char		cNumberOfBytes;

		cCounter++;
		if (cCounter == 8)
		{
			cCounter=0;
			cNumberOfBytes = 2;
		}
		else cNumberOfBytes = 3;

		if ( (g_unBeeperBufferPartPos+cNumberOfBytes) <= BUFFERLEN)
		{
			if (g_bEarBitOn == 1)
			{
				// Fills beeper sound buffer with 3 equal bytes (0xff) = beeper state high

				if (cNumberOfBytes == 3)
				{
					Buffer1[g_unBeeperBufferPartPos]=0x7000;
					Buffer1[g_unBeeperBufferPartPos+1]=0x7000;
					Buffer1[g_unBeeperBufferPartPos+2]=0x7000;
				}
				else
				{
					Buffer1[g_unBeeperBufferPartPos]=0x7000;
					Buffer1[g_unBeeperBufferPartPos+1]=0x7000;
				}
			}
			else
			{	// Fills beeper sound buffer with 3 equal bytes (0x00) = beeper state low

				if (cNumberOfBytes == 3)
				{
					Buffer1[g_unBeeperBufferPartPos]=0x00;
					Buffer1[g_unBeeperBufferPartPos+1]=0x00;
					Buffer1[g_unBeeperBufferPartPos+2]=0x00;
				}
				else
				{

					Buffer1[g_unBeeperBufferPartPos]=0x00;
					Buffer1[g_unBeeperBufferPartPos+1]=0x00;
				}
			}
			g_unBeeperBufferPartPos+=cNumberOfBytes;
		}
}

void RenderBeeper22050 ()
{
	unsigned long	unBufferPos=0;
	static char	cCounter=0;
	char		cNumberOfBytes;

		cCounter++;
		if (cCounter == 8)
		{
			cCounter=0;
			cNumberOfBytes = 1;
		}
		else cNumberOfBytes = 2;

		if ( (g_unBeeperBufferPartPos+cNumberOfBytes) <= BUFFERLEN)
		{
			if (g_bEarBitOn == 1)
			{
				// Fills beeper sound buffer with 3 equal bytes (0xff) = beeper state high

				if (cNumberOfBytes == 2)
				{
					Buffer1[g_unBeeperBufferPartPos]=0x7000;
					Buffer1[g_unBeeperBufferPartPos+1]=0x7000;
				}
				else
				{
					Buffer1[g_unBeeperBufferPartPos]=0x7000;
				}
			}
			else
			{	// Fills beeper sound buffer with 3 equal bytes (0x00) = beeper state low

				if (cNumberOfBytes == 2)
				{
					Buffer1[g_unBeeperBufferPartPos]=0x00;
					Buffer1[g_unBeeperBufferPartPos+1]=0x00;
				}
				else
				{

					Buffer1[g_unBeeperBufferPartPos]=0x00;
				}
			}
			g_unBeeperBufferPartPos+=cNumberOfBytes;
		}
}

void RenderBeeper2 ()
{
	int i;

	for (i=0; i<BUFFERLEN; i++)
	{
		if (g_bEarBitOn == 1)
		{
			Buffer1[i] = 0x7000;
		}
		else
		{	
			Buffer1[i] = 0x0000;
		}
	}
}

__inline void DrawScreen (void)
{
	unsigned short int	w, x, y, i, xbord=32, ybord=32;
	char			PixelGroup, Attrib, FlashColorIndex1, FlashColorIndex2;
	static int		FlashCount=0;
	unsigned char	*Screen;
	static int 		FlashOn = 0;
	char			buff[15];
	int				borderIndex = 0;
	uint32			col;

	Screen = GETVIDEOPAGE();
  
	FlashCount++;

	// The cursor must (should) blink every 312 ms
	// an approximation is 16/50 = 320 ms

	if ( FlashCount == 16 ) // 25/50 = half a second
    {
		if (FlashOn==0) FlashOn = 1; else FlashOn = 0;
		FlashCount=0;
    }	

	for (y=0;y < 192;y++)
	{
		x=0;
		w=0;

		while (x < 256)
		{
			// Computes spectrum pixel address (8 bits groups)
									
			PixelGroup = *(Screen + *(PixelAddressArray+(y<<8) + x));
			Attrib= *(Screen + *(AttribArray+(y<<8) + x) );

			if (Attrib & 0x80)
			{
				if (FlashOn==1)
				{
					FlashColorIndex1=(Attrib & 0x38) >> 3;
					FlashColorIndex2=Attrib & 0x07;
				}
				else
				{
					FlashColorIndex1=Attrib & 0x07;
					FlashColorIndex2=(Attrib & 0x38) >> 3;
				}
			}
			else
			{
				FlashColorIndex1=Attrib & 0x07;
				FlashColorIndex2=(Attrib & 0x38) >> 3;				
			}

			for (i=0;i < 8;i++)
			{
				// Bright ?
	
				if (Attrib & 0x40)
				{
					if (PixelGroup & 0x80) 
							*(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex1+8);
								else *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex2+8);
				}
				else
				{
					// Inverts paper and ink colors each 500 ms
	
					if (PixelGroup & 0x80) *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex1);
								else *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex2);
				}

				PixelGroup = PixelGroup << 1;
				x++;
			}

		}


	}

	// Border rendering

	borderIndex = 32;
	y = 0;
	for (i=0;i<32;i++)
	{
		col = *(palEntries+*(border+borderIndex))<<16|*(palEntries+*(border+borderIndex));
		sq_set32 (vbuffer + (y++*(DISPLAY_WIDTH)), col, 640);

		borderIndex++;
	}

	for (i=0;i<192;i++)
	{
		col = *(palEntries+*(border+borderIndex))<<16|*(palEntries+*(border+borderIndex));
		sq_set32 (vbuffer + (y*(DISPLAY_WIDTH)), col, 64);
		sq_set32 (vbuffer + (y++*(DISPLAY_WIDTH)+288), col, 64);

		borderIndex++;
	}

	for (i=0;i<16;i++)
	{
		col = *(palEntries+*(border+borderIndex))<<16|*(palEntries+*(border+borderIndex));
		sq_set32 (vbuffer + (y++*(DISPLAY_WIDTH)), col, 640);

		borderIndex++;
	}

	renderscene(rc);
}

static void RectFill2 (uint16 *screen, unsigned int x, unsigned int y, unsigned int x2, unsigned int y2, unsigned int bg)
{
	int i, j;
	uint32 color;
	uint16 *screen2;
	
	color = *(palEntries+bg);

	for (i=y; i<y2; i++)
	{
		screen2 = screen + (i*(DISPLAY_WIDTH));
		sq_set16 (screen2+x, color, (x2-x)*2);	
	}
}

__inline void DrawHorizLine(uint16 *screen, unsigned int x, unsigned int y, unsigned int length, unsigned int bg)
{
	uint32 col;

	col = *(palEntries+bg)<<16|*(palEntries+bg);

	sq_set32 (screen + (y*(DISPLAY_WIDTH)) + x, col, length);
}

__inline void DrawScreen_full (void)
{
	unsigned short int	w, x, y, i, xbord=0, ybord=0;
	char			PixelGroup, Attrib, FlashColorIndex1, FlashColorIndex2;
	static int		FlashCount=0;
	unsigned char	*Screen;
	static int 		FlashOn = 0;
	char			buff[15];

	Screen = GETVIDEOPAGE();
  
	FlashCount++;

	// The cursor must (should) blink every 312 ms
	// an approximation is 16/50 = 320 ms

	if ( FlashCount == 16 ) // 25/50 = half a second
    {
		if (FlashOn==0) FlashOn = 1; else FlashOn = 0;
		FlashCount=0;
    }	

	for (y=0;y < 192;y++)
	{
		x=0;
		w=0;

		while (x < 256)
		{
			// Computes spectrum pixel address (8 bits groups)
						
			PixelGroup = *(Screen + *(PixelAddressArray+(y<<8) + x));
			Attrib= *(Screen + *(AttribArray+(y<<8) + x) );

			if (Attrib & 0x80)
			{
				if (FlashOn==1)
				{
					FlashColorIndex1=(Attrib & 0x38) >> 3;
					FlashColorIndex2=Attrib & 0x07;
				}
				else
				{
					FlashColorIndex1=Attrib & 0x07;
					FlashColorIndex2=(Attrib & 0x38) >> 3;
				}
			}
			else
			{
				FlashColorIndex1=Attrib & 0x07;
				FlashColorIndex2=(Attrib & 0x38) >> 3;				
			}

			for (i=0;i < 8;i++)
			{
				// Bright ?
	
				if (Attrib & 0x40)
				{
					if (PixelGroup & 0x80) 
							*(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex1+8);
								else *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex2+8);
				}
				else
				{
					// Inverts paper and ink colors each 500 ms
	
					if (PixelGroup & 0x80) *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex1);
								else *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = *(palEntries+FlashColorIndex2);
				}

				PixelGroup = PixelGroup << 1;
				x++;
			}

		}


	}

	renderscene(rc);
}

__inline void DrawLine3 (int skip)
{
	unsigned short int	i,xbord=32,ybord=32,line,test,x,y;
	char			PixelGroup,Attrib,FlashColorIndex1,FlashColorIndex2;
	char			*ptrFlashColorIndex1 = &FlashColorIndex1, *ptrFlashColorIndex2 = &FlashColorIndex2;
	unsigned char	*Screen;
	static int		FlashCount=0;
	static int 		FlashOn = 0;
	char			buff[256];
	uint32 col;	
	int				borderIndex=0;

	Screen = GETVIDEOPAGE();
  
	// The cursor must (should) blink every 312 ms
	// an approximation is 16/50 = 320 ms

	FlashCount++;

	if ( FlashCount == 16 ) // 25/50 = half a second
    {
		if (FlashOn==0) FlashOn = 1; else FlashOn = 0;
		FlashCount=0;
    }	

	if (emu_options.fullscreen == 1)
	{
		xbord=0;
		ybord=0;
	}

	for (line=0; line<312; line++)
	{

	z80_execute(224);

	#ifdef SOUND44100
	RenderBeeper44100 ();
	#else
	RenderBeeper22050 ();
	#endif
	
	y=line-32;

	x=0;

	if (y >= 32 && y < 192+32)
	{
		*(border+(borderIndex++)) = ucFePortValue;

		y -= 32;

		while (x < 256)
		{
			// Computes spectrum pixel address (8 bits groups)
									
			PixelGroup = *(Screen + *(PixelAddressArray+(y<<8) + x));
			Attrib= *(Screen + *(AttribArray+(y<<8) + x) );

			if (Attrib & 0x80)
			{
				if (FlashOn==1)
				{
					FlashColorIndex1=(Attrib & 0x38) >> 3;
					FlashColorIndex2=Attrib & 0x07;
				}
				else
				{
					FlashColorIndex1=Attrib & 0x07;
					FlashColorIndex2=(Attrib & 0x38) >> 3;
				}
			}
			else
			{
				FlashColorIndex1=Attrib & 0x07;
				FlashColorIndex2=(Attrib & 0x38) >> 3;				
			}

			for (i=0;i < 8;i++)
			{
				// Bright ?
	
				if (Attrib & 0x40)
				{

					if (PixelGroup & 0x80) 
							*(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = palEntries[FlashColorIndex1+8];
								else *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = palEntries[FlashColorIndex2+8];
				}
				else
				{
					// Inverts paper and ink colors each 500 ms
	
					if (PixelGroup & 0x80) *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = palEntries [FlashColorIndex1];
								else *(vbuffer+(DISPLAY_WIDTH*(y+ybord)) + x + xbord) = palEntries [FlashColorIndex2];
				}

				PixelGroup = PixelGroup << 1;
				x++;
			}
		}
	}
	else
	{
		*(border+(borderIndex++)) = ucFePortValue;
	}


	} // for ()

	// Border colors rendering

	borderIndex = 32;
	y = 0;
	for (i=0;i<32;i++)
	{
		col = *(palEntries+*(border+borderIndex))<<16|*(palEntries+*(border+borderIndex));
		sq_set32 (vbuffer + (y++*(DISPLAY_WIDTH)), col, 640);

		borderIndex++;
	}

	for (i=0;i<192;i++)
	{
		col = *(palEntries+*(border+borderIndex))<<16|*(palEntries+*(border+borderIndex));
		sq_set32 (vbuffer + (y*(DISPLAY_WIDTH)), col, 64);
		sq_set32 (vbuffer + (y++*(DISPLAY_WIDTH)+288), col, 64);

		borderIndex++;
	}

	for (i=0;i<16;i++)
	{
		col = *(palEntries+*(border+borderIndex))<<16|*(palEntries+*(border+borderIndex));
		sq_set32 (vbuffer + (y++*(DISPLAY_WIDTH)), col, 640);

		borderIndex++;
	}

	if (emu_options.frameskip == 1)
	{
		if (skip == 0) 
		{	
			renderscene(rc);
			skip++;
		}
		else
		{
			skip = 0;
		}
	}
	else
	{
		renderscene(rc);
	}
	
}

unsigned short *CreateSpectrumPalette (void)
{
	unsigned short int* redLUT;
	unsigned short int* greenLUT;
	unsigned short int* blueLUT;
	unsigned short int globalColourTable[256 * 3];
	unsigned short int rMask;
	unsigned short int gMask;
	unsigned short int bMask;
	long palTable = 0;
	static unsigned short *palEntries;

	redLUT = globalColourTable;     // Get a pointer to each lookup table
	greenLUT = globalColourTable + 256;
	blueLUT = globalColourTable + 512;

	// mask for RGB mode 565

	rMask = (unsigned short)0xf800;
	gMask = (unsigned short)0x07e0;
	bMask = (unsigned short)0x001f;

    	// Create the lookup table

	for (palTable = 0; palTable < 256; palTable++)
	{
        		redLUT[palTable] = (unsigned short int)((( (long)(rMask) * palTable) >> 8) & rMask);
        		greenLUT[palTable] = (unsigned short int)((( (long)(gMask * palTable)) >> 8) & gMask);
        		blueLUT[palTable] = (unsigned short int)((( (long)(bMask) * palTable) >> 8) & bMask);
    	}

	palEntries = (unsigned short *)malloc ( (sizeof(unsigned short) * 256) );

	// BRIGHT 0 colors

	palEntries[0] = (unsigned short)(0x0000);
	palEntries[1] = (unsigned short)redLUT[0] | greenLUT[0] | blueLUT[159];
	palEntries[2] = (unsigned short)redLUT[223] | greenLUT[0] | blueLUT[0];
	palEntries[3] = (unsigned short)redLUT[224] | greenLUT[0] | blueLUT[176];
	palEntries[4] = (unsigned short)redLUT[0] | greenLUT[208] | blueLUT[0];
	palEntries[5] = (unsigned short)redLUT[0] | greenLUT[208] | blueLUT[208];
	palEntries[6] = (unsigned short)redLUT[207] | greenLUT[207] | blueLUT[0];
	palEntries[7] = (unsigned short)redLUT[192] | greenLUT[199] | blueLUT[192];

	// Spectrum BRIGHT 1 colors

	palEntries[8] = (unsigned short)(0x0000);
	palEntries[9] = (unsigned short)redLUT[0] | greenLUT[0] | blueLUT[175];
	palEntries[10] = (unsigned short)redLUT[239] | greenLUT[0] | blueLUT[0];
	palEntries[11] = (unsigned short)redLUT[255] | greenLUT[0] | blueLUT[223];
	palEntries[12] = (unsigned short)redLUT[0] | greenLUT[239] | blueLUT[0];
	palEntries[13] = (unsigned short)redLUT[0] | greenLUT[248] | blueLUT[255];
	palEntries[14] = (unsigned short)redLUT[255] | greenLUT[248] | blueLUT[0];
	palEntries[15] = (unsigned short)redLUT[255] | greenLUT[248] | blueLUT[255];

	// File selector

	palEntries[16] = (unsigned short)redLUT[255] | greenLUT[255] | blueLUT[255];		// File Size
	palEntries[17] = (unsigned short)redLUT[255] | greenLUT[255] | blueLUT[255];		// File Name
	palEntries[240] = (unsigned short)redLUT[255] | greenLUT[255] | blueLUT[255];		// Previous <..>
	palEntries[160] = (unsigned short)redLUT[255] | greenLUT[255] | blueLUT[255];		// Root <.>
	palEntries[70] = (unsigned short)redLUT[255] | greenLUT[255] | blueLUT[255];		// Folder 
	
	palEntries[27] = (unsigned short)redLUT[255] | greenLUT[0] | blueLUT[0];	// Frame background around "Select File"
	palEntries[255] = (unsigned short)redLUT[0] | greenLUT[0] | blueLUT[0];	 	// Background 
	
	palEntries[123] = (unsigned short)redLUT[0] | greenLUT[0] | blueLUT[255];    // Frame background around the selected file
	
	return (palEntries);
}

void cpu_writemem16(unsigned short addr, unsigned char data)
{
	if (addr >= 0x4000)
	{
		POKE (addr, data);
	}
}

void cpu_writeport (int port, int data)
{
	if ( (port & 0x00ff) == 0xfe || (port >> 8) == 0xfe)
	{			
		ucFePortValue = data & 0x07;
		
		if (data & 0x10) // EAR bit set
		{
			g_bEarBitOn = 1;
		}
		else // EAR bit not set then check MIC bit
		{
			g_bEarBitOn = 0;
		}
	}	

	// AY-3-8910 Sound emulation

	if (port == 0xfffd)
	{
		// Selects a register

		AY8910Write(0, 0, data);
	}
	else
	if (port == 0xbffd)
	{
		// Writes to the selected register

		AY8910Write(0, 1, data);

		g_bWrite8910 = 1;
	}

	// 128k paging port

	if (port == 0x7FFD)
	{
		PageControl (data);
	}
}

unsigned char cpu_readport (int port)
{
	unsigned char Port, Val2, KeysPressed[8];
	int i=0,counter=0;
	unsigned char kBuffer[256];	
	unsigned char Port7F,Port1F;
	int ExKey;
	unsigned char PortHigh, PortLow;
	unsigned int reg;
	unsigned char regaf;
	uint16	ret;
	cont_cond_t cond;
	kbd_cond_t kbcond;
	uint8	down_timer=0;
	kbd_state_t *m_Keyboard;
	maple_device_t *keybDevice;
	
	PortHigh = (port & 0xff00) >> 8;
	PortLow = (port & 0x00ff);

	switch (PortLow)
	{		
		case 0xfe:

			memset (&kBuffer[0], 0, 256);

			cond.buttons = 0;

			// A key has been pressed on the virtual keyboard ? Then emulate the keypress...

			if (virtual_key != 0)
			{
				switch (virtual_key)
				{
					case '1': kBuffer[DIK_1] = 0x80;
						break;
					case '2': kBuffer[DIK_2] = 0x80;
						break;
					case '3': kBuffer[DIK_3] = 0x80;
						break;
					case '4': kBuffer[DIK_4] = 0x80;
						break;
					case '5': kBuffer[DIK_5] = 0x80;
						break;
					case '6': kBuffer[DIK_6] = 0x80;
						break;
					case '7': kBuffer[DIK_7] = 0x80;
						break;

					case '8': kBuffer[DIK_8] = 0x80;
						break;
					case '9': kBuffer[DIK_9] = 0x80;
						break;
					case '0': kBuffer[DIK_0] = 0x80;
						break;
					case 'A': kBuffer[DIK_A] = 0x80;
						break;
					case 'B': kBuffer[DIK_B] = 0x80;
						break;
					case 'C': kBuffer[DIK_C] = 0x80;
						break;
					case 'D': kBuffer[DIK_D] = 0x80;
						break;
					case 'E': kBuffer[DIK_E] = 0x80;
						break;
					case 'F': kBuffer[DIK_F] = 0x80;
						break;
					case 'G': kBuffer[DIK_G] = 0x80;
						break;
					case 'H': kBuffer[DIK_H] = 0x80;
						break;
					case 'I': kBuffer[DIK_I] = 0x80;
						break;
					case 'J': kBuffer[DIK_J] = 0x80;
						break;
					case 'K': kBuffer[DIK_K] = 0x80;
						break;
					case 'L': kBuffer[DIK_L] = 0x80;
						break;
					case 'M': kBuffer[DIK_M] = 0x80;
						break;
					case 'N': kBuffer[DIK_N] = 0x80;
						break;
					case 'O': kBuffer[DIK_O] = 0x80;
						break;
					case 'P': kBuffer[DIK_P] = 0x80;

						 // Simulate a double quote by pressing the 'control' key

						 if (bLoadCommand == 1)
						 {
							kBuffer[DIK_LCONTROL] = 0x80;
							//kBuffer[DIK_RCONTROL] = 0x80;
						 }
						break;
					case '"': kBuffer[DIK_P] = 0x80;

						 // Simulate a double quote by pressing the 'control' key

						kBuffer[DIK_LCONTROL] = 0x80;

						break;						
					case 'Q': kBuffer[DIK_Q] = 0x80;
						break;
					case 'R': kBuffer[DIK_R] = 0x80;
						break;
					case 'S': kBuffer[DIK_S] = 0x80;
						break;
					case 'T': kBuffer[DIK_T] = 0x80;
						break;
					case 'U': kBuffer[DIK_U] = 0x80;
						break;
					case 'V': kBuffer[DIK_V] = 0x80;
						break;
					case 'W': kBuffer[DIK_W] = 0x80;
						break;
					case 'X': kBuffer[DIK_X] = 0x80;
						break;
					case 'Y': kBuffer[DIK_Y] = 0x80;
						break;
					case 'Z': kBuffer[DIK_Z] = 0x80;
						break;
					case ' ': kBuffer[DIK_SPACE] = 0x80;
						break;
					case 1 :
						// caps shift
						
						kBuffer[DIK_LCONTROL] = 0x80; 
						kBuffer[DIK_RCONTROL] = 0x80; 
						break;
					case 2 :
						// symbol shift
						
						kBuffer[DIK_LSHIFT] = 0x80;
						kBuffer[DIK_RSHIFT] = 0x80;
						break;
					case 3 :
						// Enter
						 
						kBuffer[DIK_RETURN] = 0x80; // enter
						break;
					case 4 : 
						// back space
						
						kBuffer[DIK_LSHIFT]|=0x80;
						kBuffer[DIK_RSHIFT]|=0x80;
						kBuffer[DIK_0]|=0x80;
						break;							
						
				}		

			}
			else
			{
				// Test keyboard keys

				keybDevice = maple_enum_type(0, MAPLE_FUNC_KEYBOARD);
				// printf ("keybDevice = %d\n", keybDevice);
				if (keybDevice != 0)
				{
					m_Keyboard = (kbd_state_t *)maple_dev_status(keybDevice); 
					//printf ("m_Keyboard = %d\n", m_Keyboard);

					for (i=0; i<6; i++)
					{
						if (m_Keyboard->cond.keys[i] != 0)
						{
							switch (m_Keyboard->cond.keys[i])
							{
								case KBD_KEY_1: kBuffer[DIK_1] = 0x80;
									break;
								case KBD_KEY_2: kBuffer[DIK_2] = 0x80;
									break;
								case KBD_KEY_3: kBuffer[DIK_3] = 0x80;
									break;
								case KBD_KEY_4: kBuffer[DIK_4] = 0x80;
									break;
								case KBD_KEY_5: kBuffer[DIK_5] = 0x80;
									break;
								case KBD_KEY_6: kBuffer[DIK_6] = 0x80;
									break;
								case KBD_KEY_7: kBuffer[DIK_7] = 0x80;
									break;
								case KBD_KEY_8: kBuffer[DIK_8] = 0x80;
									break;
								case KBD_KEY_9: kBuffer[DIK_9] = 0x80;
									break;
								case KBD_KEY_0: kBuffer[DIK_0] = 0x80;
									break;
								case KBD_KEY_A: kBuffer[DIK_A] = 0x80;
									break;
								case KBD_KEY_B: kBuffer[DIK_B] = 0x80;
									break;
								case KBD_KEY_C: kBuffer[DIK_C] = 0x80;
									break;
								case KBD_KEY_D: kBuffer[DIK_D] = 0x80;
									break;
								case KBD_KEY_E: kBuffer[DIK_E] = 0x80;
									break;
								case KBD_KEY_F: kBuffer[DIK_F] = 0x80;
									break;
								case KBD_KEY_G: kBuffer[DIK_G] = 0x80;
									break;
								case KBD_KEY_H: kBuffer[DIK_H] = 0x80;
									break;
								case KBD_KEY_I: kBuffer[DIK_I] = 0x80;
									break;
								case KBD_KEY_J: kBuffer[DIK_J] = 0x80;
									break;
								case KBD_KEY_K: kBuffer[DIK_K] = 0x80;
									break;
								case KBD_KEY_L: kBuffer[DIK_L] = 0x80;
									break;
								case KBD_KEY_M: kBuffer[DIK_M] = 0x80;
									break;
								case KBD_KEY_N: kBuffer[DIK_N] = 0x80;
									break;
								case KBD_KEY_O: kBuffer[DIK_O] = 0x80;
									break;
								case KBD_KEY_P: kBuffer[DIK_P] = 0x80;

									 // Simulate a double quote by pressing the 'control' key

									 if (bLoadCommand == 1)
									 {
										kBuffer[DIK_LCONTROL] = 0x80;
										//kBuffer[DIK_RCONTROL] = 0x80;
									 }
									break;
								case KBD_KEY_QUOTE: kBuffer[DIK_P] = 0x80;

									 // Simulate a double quote by pressing the 'control' key

									kBuffer[DIK_LCONTROL] = 0x80;

									break;						
								case KBD_KEY_Q: kBuffer[DIK_Q] = 0x80;
									break;
								case KBD_KEY_R: kBuffer[DIK_R] = 0x80;
									break;
								case KBD_KEY_S: kBuffer[DIK_S] = 0x80;
									break;
								case KBD_KEY_T: kBuffer[DIK_T] = 0x80;
									break;
								case KBD_KEY_U: kBuffer[DIK_U] = 0x80;
									break;
								case KBD_KEY_V: kBuffer[DIK_V] = 0x80;
									break;
								case KBD_KEY_W: kBuffer[DIK_W] = 0x80;
									break;
								case KBD_KEY_X: kBuffer[DIK_X] = 0x80;
									break;
								case KBD_KEY_Y: kBuffer[DIK_Y] = 0x80;
									break;
								case KBD_KEY_Z: kBuffer[DIK_Z] = 0x80;
									break;
								case KBD_KEY_SPACE: kBuffer[DIK_SPACE] = 0x80;
									break;

								case KBD_KEY_ENTER :
									// Enter
									 
									kBuffer[DIK_RETURN] = 0x80; // enter
									break;
								case KBD_KEY_BACKSPACE : 
									// back space
									
									kBuffer[DIK_LSHIFT]|=0x80;
									kBuffer[DIK_RSHIFT]|=0x80;
									kBuffer[DIK_0]|=0x80;
									break;				
								// Simulate joystick
								case KBD_KEY_UP : 
									cond.buttons ^= ~CONT_DPAD_UP;
									break;
								case KBD_KEY_DOWN : 
									cond.buttons ^= ~CONT_DPAD_DOWN;
									break;
								case KBD_KEY_LEFT : 
									cond.buttons ^= ~CONT_DPAD_LEFT;
									break;
								case KBD_KEY_RIGHT : 
									cond.buttons ^= ~CONT_DPAD_RIGHT;
									break;
							}
						}
					}

					if (m_Keyboard->cond.modifiers & KBD_MOD_LSHIFT)
					{
							kBuffer[DIK_LCONTROL] = 0x80; 
							kBuffer[DIK_RCONTROL] = 0x80; 
					}

					if (m_Keyboard->cond.modifiers & KBD_MOD_RSHIFT)
					{
							kBuffer[DIK_LSHIFT] = 0x80;
							kBuffer[DIK_RSHIFT] = 0x80;
					}

				}


			}

			// no directional kb key pressed ? then read joy
			if (cond.buttons == 0) if(cont_get_cond(maple_first_controller(), &cond) <0) break;
		
			/* Check for start on the controller */
				
			// Simulate shift 0 = backspace

/*			if (kBuffer[DIK_BACK] & 0x80) 
			{
				kBuffer[DIK_LSHIFT]|=0x80;
				kBuffer[DIK_RSHIFT]|=0x80;
				kBuffer[DIK_0]|=0x80;
			}
*/
			// Sinclair 1 joystick emulation
			
			if (emu_options.joy == SINCLAIR1)
			{
				if ( !(cond.buttons & CONT_DPAD_UP))
				{	
					kBuffer[DIK_9]|=0x80;
				}
			
				if ( !(cond.buttons & CONT_DPAD_DOWN))
				{	kBuffer[DIK_8]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_LEFT))
				{	kBuffer[DIK_6]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_RIGHT))
				{	kBuffer[DIK_7]|=0x80;}
			
				if ( !(cond.buttons & CONT_X))
				{	kBuffer[DIK_0]|=0x80;}
			
				if ( !(cond.buttons & CONT_A))
				{	kBuffer[DIK_RETURN]|=0x80;}

			}
			else if (emu_options.joy == SINCLAIR2)
			{
				if ( !(cond.buttons & CONT_DPAD_UP))
				{	kBuffer[DIK_4]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_DOWN))
				{	kBuffer[DIK_3]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_LEFT))
				{	kBuffer[DIK_1]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_RIGHT))
				{	kBuffer[DIK_2]|=0x80;}
			
				if ( !(cond.buttons & CONT_X))
				{	kBuffer[DIK_5]|=0x80;}
			
				if ( !(cond.buttons & CONT_A))
				{	kBuffer[DIK_RETURN]|=0x80;}

			}
			else if (emu_options.joy == CURSOR)
			{
				// Cursor

				if ( !(cond.buttons & CONT_DPAD_UP))
				{	kBuffer[DIK_7]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_DOWN))
				{	kBuffer[DIK_6]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_LEFT))
				{	kBuffer[DIK_5]|=0x80;}
			
				if ( !(cond.buttons & CONT_DPAD_RIGHT))
				{	kBuffer[DIK_8]|=0x80;}
			
				if ( !(cond.buttons & CONT_X))
				{	kBuffer[DIK_0]|=0x80;}
			
				if ( !(cond.buttons & CONT_A))
				{	kBuffer[DIK_RETURN]|=0x80;}

			}
			else
			{
				// Custom keys

				if ( !(cond.buttons & CONT_DPAD_UP))
				{	selectKey (&kBuffer[0], emu_options.k_up);}
			
				if ( !(cond.buttons & CONT_DPAD_DOWN))
				{	selectKey (&kBuffer[0], emu_options.k_down);}
			
				if ( !(cond.buttons & CONT_DPAD_LEFT))
				{	selectKey (&kBuffer[0], emu_options.k_left);}
			
				if ( !(cond.buttons & CONT_DPAD_RIGHT))
				{	selectKey (&kBuffer[0], emu_options.k_right);}
			
				if ( !(cond.buttons & CONT_X))
				{	selectKey (&kBuffer[0], emu_options.k_fire1);}
			
				if ( !(cond.buttons & CONT_A))
				{	selectKey (&kBuffer[0], emu_options.k_fire2);}
			}

			KeysPressed[ASDFG]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_A] & 0x80) KeysPressed[ASDFG]&=0xfe;
			if (kBuffer[DIK_S] & 0x80) KeysPressed[ASDFG]&=0xfd;
			if (kBuffer[DIK_D] & 0x80) KeysPressed[ASDFG]&=0xfb;
			if (kBuffer[DIK_F] & 0x80) KeysPressed[ASDFG]&=0xf7;
			if (kBuffer[DIK_G] & 0x80) KeysPressed[ASDFG]&=0xef;

			KeysPressed[SZXCV]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_LSHIFT] & 0x80 || kBuffer[DIK_RSHIFT] & 0x80) KeysPressed[SZXCV]&=0xfe;
			if (kBuffer[DIK_Z] & 0x80) KeysPressed[SZXCV]&=0xfd;
			if (kBuffer[DIK_X] & 0x80) KeysPressed[SZXCV]&=0xfb;
			if (kBuffer[DIK_C] & 0x80) KeysPressed[SZXCV]&=0xf7;
			if (kBuffer[DIK_V] & 0x80) KeysPressed[SZXCV]&=0xef;

			KeysPressed[QWERT]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_Q] & 0x80) KeysPressed[QWERT]&=0xfe;
			if (kBuffer[DIK_W] & 0x80) KeysPressed[QWERT]&=0xfd;
			if (kBuffer[DIK_E] & 0x80) KeysPressed[QWERT]&=0xfb;
			if (kBuffer[DIK_R] & 0x80) KeysPressed[QWERT]&=0xf7;
			if (kBuffer[DIK_T] & 0x80) KeysPressed[QWERT]&=0xef;
					
			KeysPressed[N12345]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_1] & 0x80) KeysPressed[N12345]&=0xfe;
			if (kBuffer[DIK_2] & 0x80) KeysPressed[N12345]&=0xfd;	
			if (kBuffer[DIK_3] & 0x80) KeysPressed[N12345]&=0xfb;
			if (kBuffer[DIK_4] & 0x80) KeysPressed[N12345]&=0xf7;
			if (kBuffer[DIK_5] & 0x80) KeysPressed[N12345]&=0xef;

			KeysPressed[N09876]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_0] & 0x80) KeysPressed[N09876]&=0xfe;
			if (kBuffer[DIK_9] & 0x80) KeysPressed[N09876]&=0xfd;
			if (kBuffer[DIK_8] & 0x80) KeysPressed[N09876]&=0xfb;
			if (kBuffer[DIK_7] & 0x80) KeysPressed[N09876]&=0xf7;
			if (kBuffer[DIK_6] & 0x80) KeysPressed[N09876]&=0xef;

			KeysPressed[POIUY]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_P] & 0x80) KeysPressed[POIUY]&=0xfe;
			if (kBuffer[DIK_O] & 0x80) KeysPressed[POIUY]&=0xfd;
			if (kBuffer[DIK_I] & 0x80) KeysPressed[POIUY]&=0xfb;
			if (kBuffer[DIK_U] & 0x80) KeysPressed[POIUY]&=0xf7;
			if (kBuffer[DIK_Y] & 0x80) KeysPressed[POIUY]&=0xef;

			KeysPressed[ELKJH]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_RETURN] & 0x80) KeysPressed[ELKJH]&=0xfe;		
			if (kBuffer[DIK_L] & 0x80) KeysPressed[ELKJH]&=0xfd;
			if (kBuffer[DIK_K] & 0x80) KeysPressed[ELKJH]&=0xfb;
			if (kBuffer[DIK_J] & 0x80) KeysPressed[ELKJH]&=0xf7;
			if (kBuffer[DIK_H] & 0x80) KeysPressed[ELKJH]&=0xef;

			KeysPressed[SPSSMN]=(char)0xff;	// No key pressed
			if (kBuffer[DIK_SPACE] & 0x80) KeysPressed[SPSSMN]&=0xfe;
			if (kBuffer[DIK_LCONTROL] & 0x80 || kBuffer[DIK_RCONTROL] & 0x80) KeysPressed[SPSSMN]&=0xfd;
			if (kBuffer[DIK_M] & 0x80) KeysPressed[SPSSMN]&=0xfb;
			if (kBuffer[DIK_N] & 0x80) KeysPressed[SPSSMN]&=0xf7;	
			if (kBuffer[DIK_B] & 0x80) KeysPressed[SPSSMN]&=0xef;	

			Port=(char)0xff;
			if (!(PortHigh & 0x01)) Port&=KeysPressed[SZXCV];		
			if (!(PortHigh & 0x02)) Port&=KeysPressed[ASDFG];
			if (!(PortHigh & 0x04)) Port&=KeysPressed[QWERT];		
			if (!(PortHigh & 0x08)) Port&=KeysPressed[N12345];		
			if (!(PortHigh & 0x10)) Port&=KeysPressed[N09876];		
			if (!(PortHigh & 0x20)) Port&=KeysPressed[POIUY];		
			if (!(PortHigh & 0x40)) Port&=KeysPressed[ELKJH];		
			if (!(PortHigh & 0x80)) Port&=KeysPressed[SPSSMN];
			Val2=Port;

			// z80_get_context (&z80regs);
			
			reg = z80_get_reg (Z80_AF);
			regaf = reg & 0xff;

			/*
			if (Port==0) z80regs.AF.b.l |= Z_FLAG; else z80regs.AF.b.l &= CL_Z_FLAG;
			if (Port & 0x80) z80regs.AF.b.l |= S_FLAG; else z80regs.AF.b.l &= CL_S_FLAG;
			while (i<8){if (Val2 & 0x01) counter++;Val2=Val2 >> 1;i++;}
			if (!(counter & 0x01)) z80regs.AF.b.l |= PV_FLAG; else z80regs.AF.b.l &= CL_PV_FLAG;
			*/

			if (Port==0) regaf |= Z_FLAG; else regaf &= CL_Z_FLAG;
			if (Port & 0x80) regaf |= S_FLAG; else regaf &= CL_S_FLAG;
			while (i<8){if (Val2 & 0x01) counter++;Val2=Val2 >> 1;i++;}
			if (!(counter & 0x01)) regaf |= PV_FLAG; else regaf &= CL_PV_FLAG;
			reg = (reg & 0xff00)|regaf;
			z80_set_reg (Z80_AF, reg);

			// z80_set_context (&z80regs);

			return (Port);

			break;		

		// Fuller interface joystick emulation

		case 0x7f:

			Port7F=0xff;

			if(cont_get_cond(maple_first_controller(), &cond) <0) break;

			if ( !(cond.buttons & CONT_DPAD_UP))
			{	Port7F&=~Full_UP;}
			
			if ( !(cond.buttons & CONT_DPAD_DOWN))
			{	Port7F&=~Full_DOWN;}
			
			if ( !(cond.buttons & CONT_DPAD_LEFT))
			{	Port7F&=~Full_LEFT;}
			
			if ( !(cond.buttons & CONT_DPAD_RIGHT))
			{	Port7F&=~Full_RIGHT; }
			
			if ( !(cond.buttons & CONT_X))
			{	Port7F&=~Full_FIRE;}

			Val2=Port7F;

			reg = z80_get_reg (Z80_AF);
			regaf = reg & 0xff;
			if (Port7F==0) regaf |= Z_FLAG; else regaf &= CL_Z_FLAG;
			if (Port7F & 0x80) regaf |= S_FLAG; else regaf &= CL_S_FLAG;
			while (i<8){if (Val2 & 0x01) counter++;Val2=Val2 >> 1;i++;}
			if (!(counter & 0x01)) regaf |= PV_FLAG; else regaf &= CL_PV_FLAG;
			reg = (reg & 0xff00)|regaf;
			z80_set_reg (Z80_AF, reg);			
			
			return Port7F;			

			break;

		// Kempston interface joystick emulation

		case 0x1f:

			Port1F=0x00;

			if(cont_get_cond(maple_first_controller(), &cond) <0) break;
		
			/* Check for start on the controller */
							
			if ( !(cond.buttons & CONT_DPAD_UP))
			{	Port1F|=Kem_UP;}
			
			if ( !(cond.buttons & CONT_DPAD_DOWN))
			{	Port1F|=Kem_DOWN;}
			
			if ( !(cond.buttons & CONT_DPAD_LEFT))
			{	Port1F|=Kem_LEFT;}
			
			if ( !(cond.buttons & CONT_DPAD_RIGHT))
			{	Port1F|=Kem_RIGHT;}
			
			if ( !(cond.buttons & CONT_X))
			{	Port1F|=Kem_FIRE;}
									
			Val2=Port1F;

			reg = z80_get_reg (Z80_AF);
			regaf = reg & 0xff;
			if (Port1F==0) regaf |= Z_FLAG; else regaf &= CL_Z_FLAG;
			if (Port1F & 0x80) regaf |= S_FLAG; else regaf &= CL_S_FLAG;
			while (i<8){if (Val2 & 0x01) counter++;Val2=Val2 >> 1;i++;}
			if (!(counter & 0x01)) regaf |= PV_FLAG; else regaf &= CL_PV_FLAG;
			reg = (reg & 0xff00)|regaf;
			z80_set_reg (Z80_AF, reg);

			return Port1F;
			break;

		default:
			return 0;
	}
}

void selectKey (unsigned char *kBuffer, char key)
{
				switch (key)
				{
					case '1': kBuffer[DIK_1] = 0x80;
						break;
					case '2': kBuffer[DIK_2] = 0x80;
						break;
					case '3': kBuffer[DIK_3] = 0x80;
						break;
					case '4': kBuffer[DIK_4] = 0x80;
						break;
					case '5': kBuffer[DIK_5] = 0x80;
						break;
					case '6': kBuffer[DIK_6] = 0x80;
						break;
					case '7': kBuffer[DIK_7] = 0x80;
						break;
					case '8': kBuffer[DIK_8] = 0x80;
						break;
					case '9': kBuffer[DIK_9] = 0x80;
						break;
					case '0': kBuffer[DIK_0] = 0x80;
						break;
					case 'A': kBuffer[DIK_A] = 0x80;
						break;
					case 'B': kBuffer[DIK_B] = 0x80;
						break;
					case 'C': kBuffer[DIK_C] = 0x80;
						break;
					case 'D': kBuffer[DIK_D] = 0x80;
						break;
					case 'E': kBuffer[DIK_E] = 0x80;
						break;
					case 'F': kBuffer[DIK_F] = 0x80;
						break;
					case 'G': kBuffer[DIK_G] = 0x80;
						break;
					case 'H': kBuffer[DIK_H] = 0x80;
						break;
					case 'I': kBuffer[DIK_I] = 0x80;
						break;
					case 'J': kBuffer[DIK_J] = 0x80;
						break;
					case 'K': kBuffer[DIK_K] = 0x80;
						break;
					case 'L': kBuffer[DIK_L] = 0x80;
						break;
					case 'M': kBuffer[DIK_M] = 0x80;
						break;
					case 'N': kBuffer[DIK_N] = 0x80;
						break;
					case 'O': kBuffer[DIK_O] = 0x80;
						break;
					case 'P': kBuffer[DIK_P] = 0x80;
						break;
					case 'Q': kBuffer[DIK_Q] = 0x80;
						break;
					case 'R': kBuffer[DIK_R] = 0x80;
						break;
					case 'S': kBuffer[DIK_S] = 0x80;
						break;
					case 'T': kBuffer[DIK_T] = 0x80;
						break;
					case 'U': kBuffer[DIK_U] = 0x80;
						break;
					case 'V': kBuffer[DIK_V] = 0x80;
						break;
					case 'W': kBuffer[DIK_W] = 0x80;
						break;
					case 'X': kBuffer[DIK_X] = 0x80;
						break;
					case 'Y': kBuffer[DIK_Y] = 0x80;
						break;
					case 'Z': kBuffer[DIK_Z] = 0x80;
						break;
					case ' ': kBuffer[DIK_SPACE] = 0x80;
						break;
					case 1 :
						// caps shift
						
						kBuffer[DIK_LCONTROL] = 0x80; 
						kBuffer[DIK_RCONTROL] = 0x80; 
						break;
					case 2 :
						// symbol shift
						
						kBuffer[DIK_LSHIFT] = 0x80;
						kBuffer[DIK_RSHIFT] = 0x80;
						break;
					case 3 :
						// Enter
						 
						kBuffer[DIK_RETURN] = 0x80; // enter
						break;
					case 4 : 
						// back space
						
						kBuffer[DIK_LSHIFT]|=0x80;
						kBuffer[DIK_RSHIFT]|=0x80;
						kBuffer[DIK_0]|=0x80;
						break;						
				}
}

int irq_callback(int param)
{	
	return (0xFF);
}
