# ------------------------------------------
# Title:			Default Makefile for use in VC++
# File:		      default.mak
# Revision:             20020210 8:30pm
# 
# Author:			Benjamin D. Hale
# Orignally by:         Jaap Suter
# Contributed to by:    Christer Andersson
# 
# Info:		      This file is contains the make-instructions
#				for the project. Replace DEVDIR with the appropriate
#                       base directory for devkitdc. After you make the 
#                       needed modifications here rename this file to
#                       <project_name>.mak so Visual C++ does not get
#                       confused.
# -------------------------------------------

# -------------------------------------------
# Project name definition makes this whole thing 
# a lot easier to use. Only two total modifications
# (other than what .o files will be needed)
# should be needed to get this working with your
# compiler and Visual C++. Replace "Default" with
# your project's name.
# -------------------------------------------

PROJECT = dctest1

# -------------------------------------------
# Define some directories;
# -------------------------------------------

# -------------------------------------------
# Base Directory for gcc arm-agb-elf replace with 
# wherever you put your gba devkitadv files.
# -------------------------------------------

DEVDIR  = D:\DevKitDc

# -------------------------------------------
# Source directory (where your project's source files are located.)
# -------------------------------------------

SRCDIR  = D:\PROJECTS\dctest1\

# -------------------------------------------
# Compiler Directories for binaries, includes and libs.
# -------------------------------------------

CMPDIR  = $(DEVDIR)\bin
LIBDIR  = $(DEVDIR)\lib
LIBDIR2 = $(DEVDIR)\sh-sega-dreamcast\lib
INCDIR  = $(DEVDIR)\include
INCDIR2 = $(DEVDIR)\kernel\arch\dreamcast\include
INCDIR3 = $(SRCDIR)\libspec

# -------------------------------------------
# END of directory defines
# -------------------------------------------

# -------------------------------------------
# Define what extensions we use;
# -------------------------------------------
.SUFFIXES : .cpp .c .s

# -------------------------------------------
# Define the flags for the compilers;
# -------------------------------------------
CFLAGS  = -I $(INCDIR) -I $(INCDIR2) -I $(INCDIR3) -ml -m4-single-only -O3 -fno-builtin -fno-strict-aliasing -fomit-frame-pointer -nostdlib
#CFLAGS  = -I $(INCDIR) -I $(INCDIR2) -I $(INCDIR3) -Wall -g -ml -m4-single-only -O2 -fbranch-probabilities -funroll-loops -frerun-loop-opt -fstrict-aliasing  -fomit-frame-pointer -fstrength-reduce -fschedule-insns2 -freduce-all-givs  -ffunction-sections -fdata-sections -mieee -falign-loops -fno-builtin -nostdlib 
SFLAGS  = -I $(INCDIR) -I $(INCDIR2) -little
LDFLAGS = -L $(LIBDIR) -L $(LIBDIR2) -lk++ -lbz2 -lgl -lpng -lm -lz -ldcutils -lkallisti -lgcc -Ttext=0x8c010000 

# -------------------------------------------
# Define the list of all O files;
# Just follow the syntax shown to add any 
# other objects your project may need to
# compile properly. You will need to add 
# files to this part to make it work with 
# your project add a \ to the end of all o
# files except the last one. Like below.
# -------------------------------------------

O_FILES	= \
romdisk.o \
startup.o \
Z80_mame.c \
dctest1.c \
ay8910.c \
keyb2.c \
screen1.c \
screen2.c \
dc_sound.c \
dc_vmu.c \
dc_pcm.c \
dc_snap.c \
menu.c \
file_dialog.c \
load+save.c \
screen.c \
snaps.c \
vmu.cpp \
vmutest.cpp \
slots.c \
tapreader.c \
charrom.c

# libspec\gmem.c \
# libspec\gslist.c \
# libspec\libspectrum.c \
# libspec\timings.c \
# libspec\creator.c \
# libspec\plusd.c \
# libspec\sna.c \
# libspec\snp.c \
# libspec\sp.c \
# libspec\tape_accessors.c \
# libspec\snapshot.c \
# libspec\snap_accessors.c \
# libspec\tap.c \
# libspec\tzx_read.c \
# libspec\z80.c \
# libspec\warajevo_read.c \
# libspec\tape_set.c \
# libspec\tape_block.c \
# libspec\tape.c \
# libspec\zlib.c \
# libspec\snapshotread.c \
# libspec\taperead.c \
# libspec\utils.c \
# -------------------------------------------
# There should be no need to modify anything
# below here.
# -------------------------------------------

# -------------------------------------------
# Define the final dependecy;
# -------------------------------------------
all : $(PROJECT).bin

# -------------------------------------------
# Define the copy from .elf to .bin file
# -------------------------------------------
$(PROJECT).bin : $(PROJECT).elf
	$(CMPDIR)\objcopy -v -O binary $(PROJECT).elf $(PROJECT).bin
	-@echo ------------------------------------------ 
	-@echo Done 
	-@echo ------------------------------------------
	-@scramble $(PROJECT).bin 1ST_READ.BIN
	-@echo ------------------------------------------ 
	-@echo Scrambling done. 1ST_READ.BIN file created
	-@echo ------------------------------------------ 
	-@copy /Y 1ST_READ.BIN j:\temp\compile
	-@echo ------------------------------------------ 
	-@echo 1ST_READ.BIN copied to the DC disk
	-@echo ------------------------------------------ 
	
# -------------------------------------------
# Define the linker instruction;
# -------------------------------------------
$(PROJECT).elf : $(O_FILES)
	$(CMPDIR)\gcc -o $(PROJECT).elf $(O_FILES) $(CFLAGS) $(LDFLAGS) 
	-@echo ------------------------------------------ 
	-@echo Compiling and Linking Done
	-@echo ------------------------------------------

# -------------------------------------------
# Any problems with getting this working email
# questions to whatzdat_pimp@hotmail.com .
# This was tested on devkitadv r4 and Visual C++ 6
# on a p3 450/512mb/Windows XP Pro.
# -------------------------------------------
# EOF;

