/*
 * ----------------------------------------------------------------------------------
 * slots.c
 *
 * Functions to display the memory snapshots slots
 * 
 * Copyright (C) Don_Miguel <gp32@rambler.ru>
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 */


#define _arch_dreamcast

#include <kos.h>
#include "file_dialog.h"
#include "screen.h"
#include "menu.h"

#define GP_GT_HCENTER  160
#define GP_GT_VCENTER  120

#define MAX_PATH_NAME_LEN 256

char *tmp_string[MAX_PATH_NAME_LEN];
char g_string[MAX_PATH_NAME_LEN];

void TextOut (int x, int y, char *str, int color);
void TextOut2 (int x, int y, char *str, int bg, int fg);

#define n_fstr 17

dirent_t *p_list;

typedef struct {           
	char name[16];
	int type;
	unsigned int size;
}MYDIRENTRY;
MYDIRENTRY	*pi_list;

unsigned long list_count = 0, read_count = 0;

extern uint16 vbuffer[];

myRECT      rc;

int GetFilesList(char *start_path)
{
	char *pszFileExtension;
	int i, removed=0, ndx=0;
	unsigned long ts;
	file_t d;
	dirent_t *de;
	dirent_t entry;

	list_count = 0;
	read_count = 0;

	d = fs_open(start_path, O_RDONLY | O_DIR);

	if (d == 0) 
	{
		return -1;
	}

	if (pi_list != NULL)
	{
		free(pi_list);
	}

	if (p_list != NULL)
	{
		free(p_list);
	}

	while ( (de = fs_readdir(d)) ) 
	{
		list_count ++;
	}

	list_count += 2; // for . and ..

	fs_close(d);
	p_list = NULL;
	pi_list = NULL;

	p_list = (dirent_t *)malloc(list_count * sizeof(dirent_t));
	pi_list = (MYDIRENTRY*)malloc(list_count * sizeof(MYDIRENTRY));

	d = fs_open(start_path, O_RDONLY | O_DIR);

	if (d == 0) 
	{
		return -1;
	}

	i = 0;

	// Simulate a . and .. directory

	read_count++;
	strcpy (entry.name, ".");
	entry.size = 0;
	p_list[i++] = entry;
	read_count++;
	strcpy (entry.name, "..");
	entry.size = 0;
	p_list[i++] = entry;

	while ( (de = fs_readdir(d)) ) 
	{
		read_count++;
		p_list[i] = *de;
		i ++;
	}

	fs_close(d);

	for (i=0; i<read_count; i++)
	{
		strcpy (pi_list[ndx].name, _strupr(p_list[i].name));

		if (p_list[i].size >= 0)	/* file */
		{
			pszFileExtension = (char *) (strrchr(p_list[i].name, '.') + 1);

			if (strcmp ((const char*)p_list[i].name,".")==0)
			{	// .
				pi_list[ndx].type = 0; //root
				ndx++;
			}
			else if (strcmp ((const char*)p_list[i].name,"..")==0)
			{	// ..
				pi_list[ndx].type = 1; //previous folder
				ndx++;
			} 
			else if (strcmp (pszFileExtension, "SNA") == 0 || 
				strcmp (pszFileExtension, "TAP") == 0 ||
				strcmp (pszFileExtension, "SCR") == 0 ||
				strcmp (pszFileExtension, "SLT") == 0 ||
				//strcmp (pszFileExtension, "TZX") == 0 ||
				strcmp (pszFileExtension, "Z80") == 0 )
			{
				pi_list[ndx].type = 3; //file
				pi_list[ndx].size = p_list[i].size;

				ndx ++;
			}
			else
			{
				removed ++;
			}
		}
		else	/* directory <-- (mAttr.attr & 0x10 == 1) */
		{
			pi_list[ndx].type = 2; //directory
			ndx++;
		}
	}

	list_count -= removed;

	return (int)read_count - removed;
}


int SelectFileDialog(char *start_path, int mode, char *output_path)
{
	int i,t,t1;
	int keydata;
	int gf_pos, sf_pos;
	char current_path[256];
	int nflip;
	unsigned char *ptr;
	uint16	ret;
	cont_cond_t cond;
	uint8 down_timer=0, controller;
	int bSel;
	int bExit = 0;

	rc.w = 2*512;
	rc.h = 2*512;
	rc.x = 0;
	rc.y = 0;

	gf_pos = 0;
	sf_pos = 0;

	while (!bExit)
	{
		if(cont_get_cond(maple_first_controller(), &cond) <0) break;
		if (!(cond.buttons & CONT_START)) 
			bExit = 0;
		else
			bExit = 1;
	}

	strcpy (current_path, start_path);

	if ( GetFilesList(start_path) <= 0 )
	{
		return -1;
	}

	keydata = 0;

	controller = maple_first_controller();

	while(1)
	{
		//keyboard

		if(cont_get_cond(controller, &cond) <0) break;

		if ( !(cond.buttons & CONT_DPAD_UP))
		{
			sf_pos--;
			if (sf_pos<0)
			{
				sf_pos = 0;
				gf_pos--;
				if (gf_pos<0)
				{
					gf_pos = 0;
				}
			}

		}

		if ( !(cond.buttons & CONT_DPAD_DOWN))
		{
			sf_pos++;
			if (sf_pos>=n_fstr*2)
			{
				sf_pos = n_fstr*2-1;
				gf_pos++;
			}
			if (gf_pos+sf_pos>= list_count)
			{
				sf_pos--;
			}

		}
		
		ret=check_buttons(cond.buttons,CONT_DPAD_LEFT);

		if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
		{
			if (gf_pos==0 && sf_pos<n_fstr)
			{
				sf_pos = 0;
			} 
			else 
			{
				sf_pos-=n_fstr;
				if (sf_pos<0)
				{
					gf_pos+=(sf_pos-1);
					if (gf_pos)
					{
						sf_pos+=n_fstr;
					} else
					{
						sf_pos=0;
					}
				}
				if (gf_pos<0)
				{
					gf_pos=0;
				}
			}
		}
		else
		{
			if(ret==KEY_PUSHED)
				down_timer++;			
			else
				down_timer=0;					
		}

		ret=check_buttons(cond.buttons,CONT_DPAD_RIGHT);

		if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
		{
			sf_pos+=n_fstr;
			if ( sf_pos >= n_fstr*2 )
			{
				gf_pos+=n_fstr;
				sf_pos-=n_fstr;
			}
			if (gf_pos+sf_pos>= list_count)
			{
					sf_pos-=(gf_pos+sf_pos-(int)list_count+1);
			}
			if (sf_pos<0)
				sf_pos=0;
		}
		else
		{
			if(ret==KEY_PUSHED)
				down_timer++;			
			else
				down_timer=0;					
		}

		ret=check_buttons(cond.buttons, CONT_START);

		if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
		{
			
			t = pi_list[gf_pos+sf_pos].type;

			switch (t)
			{
				case 0:
					strcpy(current_path,"/cd/");
					break;
				case 1:

					strcpy (g_string, current_path);
					for (i=strlen((char*)&g_string)-2; i>1; i-- )
					{
						if ( g_string[i] == '/')
						{
							g_string[++i] = 0;
							strcpy(current_path,(char*)&g_string);
							break;
						}
					}

					break;

				case 2:	//folder
					if (current_path[strlen(current_path)-1] != '/')
					{
						strcat(current_path, "/");
					}
					strcat(current_path, (const char*)pi_list[gf_pos+sf_pos].name);
					break;

				default: //file
					
					strcpy (output_path, current_path);
					if (current_path[strlen(output_path)-1] != '/')
					{
						strcat(output_path, "/");
					}
					strcat (output_path, pi_list[gf_pos+sf_pos].name);
					free(pi_list);
					free(p_list);
					pi_list = NULL;
					p_list = NULL;
					
					return 0;
			}

			sf_pos = 1;
			gf_pos = 0;

			GetFilesList((char*)current_path);

		}
		else
		{
			if(ret==KEY_PUSHED)
				down_timer++;			
			else
				down_timer=0;					
		}

		if ( !(cond.buttons & CONT_A)) // Button A
		{
			return 1;
		}

		RectFill(vbuffer, 0, 0, 320, 20, 27);
		RectFill(vbuffer, 0, 21, 320, 240, 0xff);

		TextOut2 (4, 9, current_path, 27, 7);

		for (i=0; i<n_fstr*2; i++)
		{
			if (gf_pos+i >= list_count)
			{
				break;
			}

			//cursor
			if (i == sf_pos)
			{
				RectFill(vbuffer, 6+(i/n_fstr)*151, 26+(i-(i>=n_fstr?n_fstr:0))*12, 6+(i/n_fstr)*151+150, 26+(i-(i>=n_fstr?n_fstr:0))*12+12, 123);
				bSel = true;
			}
			else
			{
				bSel = false;
			}

			t = pi_list[gf_pos+i].type;
			
			switch (t)
			{
				case 0: // .
					if (bSel)
					{
						TextOut2 (10+(i/n_fstr)*151, 29+(i-(i>=n_fstr?n_fstr:0))*12, "< . >  Root", 123, 160);
					}
					else
					{
						TextOut (10+(i/n_fstr)*151, 29+(i-(i>=n_fstr?n_fstr:0))*12, "< . >  Root", 160);
					}
					break;

				case 1: // ..
					if (bSel)
					{
						TextOut2 (10+(i/n_fstr)*151,29+(i-(i>=n_fstr?n_fstr:0))*12, "< .. > Previous", 123, 240);
					}
					else
					{
						TextOut (10+(i/n_fstr)*151,29+(i-(i>=n_fstr?n_fstr:0))*12, "< .. > Previous", 240);
					}
					break;

				case 2:	//folder
					sprintf(g_string, "<%s>", pi_list[gf_pos+i].name);
					if (bSel)
					{
						TextOut2 (10+(i/n_fstr)*151,29+(i-(i>=n_fstr?n_fstr:0))*12, g_string, 123, 70);
					}
					else
					{
						TextOut (10+(i/n_fstr)*151,29+(i-(i>=n_fstr?n_fstr:0))*12, g_string, 70);
					}
					break;

				default: //file
					if (bSel)
					{
						TextOut2 (10+(i/n_fstr)*151,29+(i-(i>=n_fstr?n_fstr:0))*12, pi_list[gf_pos+i].name, 123, 17);
					}
					else
					{
						TextOut(10+(i/n_fstr)*151,29+(i-(i>=n_fstr?n_fstr:0))*12, pi_list[gf_pos+i].name, 17);
					}

					t = pi_list[gf_pos+i].size;
					if (t<=999)
					{
						t1 = 'b';
					}
					else if (t <= 999999)
					{
						t /= 1000;
						t1 = 'K';
					}
					else
					{
						t /= 1000000;
						t1 = 'M';
					}
					sprintf(g_string, "%4d%c", t, (char)t1);
					if (bSel)
					{
						TextOut2 (10+(i/n_fstr)*151 + 13*8,29+(i-(i>=n_fstr?n_fstr:0))*12, g_string, 123, 16);
					}
					else
					{
						TextOut(10+(i/n_fstr)*151 + 13*8,29+(i-(i>=n_fstr?n_fstr:0))*12, g_string, 16);
					}
			}
		}

		renderscene(rc);

		Delay(50);

	}

	free(p_list);	
}

void TextOut (int x, int y, char *str, int color)
{
	text_out8b (vbuffer, str, 0, color, x, y, 1);
}

void TextOut2 (int x, int y, char *str, int bg, int fg)
{
	text_out8b (vbuffer, str, bg, fg, x, y, 1);
}
