/* snapshot.c: snapshot handling routines
   Copyright (c) 1999-2003 Philip Kendall

   $Id: snapshot.c,v 1.75 2003/11/07 14:07:33 pak21 Exp $

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

   Author contact information:

   E-mail: pak21-fuse@srcf.ucam.org
   Postal address: 15 Crescent Road, Wokingham, Berks, RG40 2DB, England

*/

#define _arch_dreamcast
#include <kos.h>

#include "config.h"

#include <stdio.h>
#include <stdlib.h>

#include "libspectrum.h"
#include "snapshotread.h"
#include "utils.h"
#include "../z80defs.h"
#include "../z80_mame.h"

#define POKE(addr, b)		(GetMemPtrEx()).MemPoke (addr,b)
#define PEEK(addr)			(GetMemPtrEx()).MemPeek (addr)

extern int errno;
extern unsigned int g_b48kmode;
extern void cpu_writeport(int port, int data); 

extern int errno;

int snapshot_read( const char *filename )
{
  utils_file file;
  libspectrum_snap *snap;
  int error;

  error = libspectrum_snap_alloc( &snap ); if( error ) return error;

  error = utils_read_file( filename, &file );
  if( error ) { libspectrum_snap_free( snap ); return error; }

  error = libspectrum_snap_read( snap, file.buffer, file.length,
				 LIBSPECTRUM_ID_UNKNOWN, filename );
  if( error ) {
    utils_close_file( &file ); libspectrum_snap_free( snap );
    return error;
  }

  if( utils_close_file( &file ) ) {
    libspectrum_snap_free( snap );
    return 1;
  }

  error = snapshot_copy_from( snap );
  if( error ) { libspectrum_snap_free( snap ); return error; }

  error = libspectrum_snap_free( snap ); if( error ) return error;

  return 0;
}

void snapshot_flush_slt (void)
{
  /*int i;
  for ( i=0; i<256; i++ ) {
    if( slt[i] ) free( slt[i] );
    slt[i] = NULL;
    slt_length[i] = 0;
  }
  slt_screen = NULL;
  */
}

int
snapshot_read_buffer( const unsigned char *buffer, size_t length,
		      libspectrum_id_t type )
{
  libspectrum_snap *snap; int error;

  error = libspectrum_snap_alloc( &snap ); if( error ) return error;

  error = libspectrum_snap_read( snap, buffer, length, type, NULL );
  if( error ) { libspectrum_snap_free( snap ); return error; }
    
  error = snapshot_copy_from( snap );
  if( error ) { libspectrum_snap_free( snap ); return error; }

  error = libspectrum_snap_free( snap ); if( error ) return error;

  return 0;
}

int snapshot_copy_from( libspectrum_snap *snap )
{
  int i,j; int error;
  int capabilities;

  z80_set_reg (Z80_AF, libspectrum_snap_a ( snap ) << 8|libspectrum_snap_f ( snap )); 
  z80_set_reg (Z80_AF2, libspectrum_snap_a_ ( snap ) << 8|libspectrum_snap_f_ ( snap )); 
  z80_set_reg (Z80_BC, libspectrum_snap_bc ( snap )); 
  z80_set_reg (Z80_DE, libspectrum_snap_de ( snap )); 
  z80_set_reg (Z80_HL, libspectrum_snap_hl ( snap )); 
  z80_set_reg (Z80_BC2, libspectrum_snap_bc_ ( snap )); 
  z80_set_reg (Z80_DE2, libspectrum_snap_de_ ( snap )); 
  z80_set_reg (Z80_HL2, libspectrum_snap_hl_ ( snap )); 
  z80_set_reg (Z80_IX, libspectrum_snap_ix ( snap )); 
  z80_set_reg (Z80_IY, libspectrum_snap_iy ( snap )); 
  z80_set_reg (Z80_I, libspectrum_snap_i ( snap )); 
  z80_set_reg (Z80_R, libspectrum_snap_r ( snap )); 
  z80_set_reg (Z80_SP, libspectrum_snap_sp ( snap )); 
  z80_set_reg (Z80_PC, libspectrum_snap_pc ( snap )); 
  z80_set_reg (Z80_IFF1, libspectrum_snap_iff1 ( snap )); 
  z80_set_reg (Z80_IFF2, libspectrum_snap_iff2 ( snap )); 
  z80_set_reg (Z80_IM, libspectrum_snap_im ( snap )); 
  z80_set_reg (Z80_HALT, libspectrum_snap_halted ( snap )); 

  if( !g_b48kmode ) 
  {
    cpu_writeport( 0xfffd, libspectrum_snap_out_ay_registerport( snap ) );
	PageControl(libspectrum_snap_out_128_memoryport( snap ) );
  }

  // tstates = libspectrum_snap_tstates( snap );

  for( i=0; i<8; i++ ) 
  {
    if( libspectrum_snap_pages( snap, i ) )
      memcpy( GetPagePtr(i), libspectrum_snap_pages( snap, i ), 0x4000 );
  }

  return 0;
}

