/*	Load / Save snapshots 

***	Copyright:	1994-1996 Gnter Woigk
  				mailto:kio@myokay.net

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.aug.94		started work (on application.c)				KIO !
	05.dec.94		replaced 'core' save&load by '.sna'			KIO !
	06.dec.94		added '.z80' v1.45 loader					KIO !
	10.dec.94		enhanced '.z80' loader to version 2.01		KIO !
	14.jan.95		separated this file from application.c			KIO !
	22.jan.95		save '.z80' added						KIO !
	05.nov.95		paged memory model support				KIO !
	25.may.03		adapted for for the GP32 by _tyrell_
	??.jun.04		adapted for the DC by _tyrell_
*/

#define _arch_dreamcast

#include <kos.h>
#include	"load+save.h"
#include "Z80_mame.h"
#include "osd_cpu.h"
#include "vmutest.hpp"

#define GET48KMEM()		 (GetMemPtrEx()).Speccy48kMem
#define MODE48K()		 (GetMemPtrEx()).b48kMode
#define GETPAGEPTR(page) (GetMemPtrEx()).GetPagePtr(page)
extern int g_b48kmode;
extern unsigned char g_Port7ffdValue;
extern unsigned char ucFePortValue; // border color (last FE port value)
 
short	romID = 0;
unsigned char *buff;
int total_size;


// *****************************************************************


/* -----	Compress .z80 file image -------------------------------------------
		compression sheme:  dc.b $ed, $ed, count, char
		return 0  --> ok --> compressed
		return !0 --> ok --> compressed, but compression increased size
		For security: the compression buffer must be at least  (5/3*sourcesize)  bytes long
*/

int CompressZ80    (unsigned char *source, long sourcesize, unsigned char *dest, long *destsize)
{	unsigned char *q,*z,c,n,*qe;
	q = source;
	z = dest;
	qe = source+ sourcesize;
	
	while (q<qe)
	{	c=*q++;
		if (*q!=c)						// single byte
		{	*z++ = c;
			if ((c==0xed)&&(q+2<=qe)&&(*q==*(q+1)))	// special care for compressible sequence after single 0xed !!!
				*z++=*q++;
			continue;
		}
		for (n=1; n<255 && *q==c; n++) q++;
		if (q>qe) { n -= q-qe; q=qe; };
		if ( (n>=4) || (c==0xed) )			// compress ?
		{	*z++ = 0xed;
			*z++ = 0xed;
			*z++ = n;
			*z++ = c;
		}
		else							// don't compress
		{	for (;n;n--) *z++ = c;
		}
	}
	
	*destsize = z-dest;
	return ( (*destsize<=sourcesize) ? 1 : 0 );
}


/* -----	Decompress .z80 file image -----------------------------------------
		compression scheme:  dc.b $ed, $ed, count, char
		return 0:	ok
		return 1:	error
*/

int DecompressZ80 (unsigned char *source, long sourcesize, unsigned char *dest, long destsize)
{	unsigned char *q,*z,c,n;

	q=source; 
	source+=sourcesize;
	z=dest; 
	dest+=destsize;
	
	while ( z<dest && q<source ) 
	{	c=*q++;						// read next byte
		if (c!=0xed) { *z++=c; continue; }	// single byte

		if (q==source) { *z++=c; continue; } // rare case: last source byte is 0xed

		c=*q++;						// read next byte
		if (c!=0xed) { *z++=0xed; q--; continue; };
		
		n=*q++; c=*q++;				// read count & char
		if ( (z+n)>dest ) return 0;
		if (q>source) return 0;
		memset (z, c, n);				// expand compressed bytes
		z+=n;
	};
	if (q==source&&z==dest) return 1;

	if (z==dest&&q+4<=source&&destsize==0xC000)	// version 1.45 file with end marker ?
		if (*q++==0) if (*q++==0xed) if (*q++==0xed) if (*q++==0) return 1;

	return 0;
}

/*
 * Save some data to the Z80 file
 */

int Buffer2File (unsigned char *pdata, long size)
{	
	// copy data to the temporary buffer, increase total size

	memcpy (&buff[total_size], pdata, size);
	total_size += size;

	return (0);
}


/* -----	save one page in .z80 v2.01 format ---------------------------------------------------
		5.nov.95:	now takes pointer to page as argument
*/
int save_block_z80 ( char ID, unsigned char *page, int compression)
{	
	long	destsize;
	int 	err=0;
	unsigned char *p;
	
	if (compression)
	{
		p = (unsigned char *)malloc (0x4000*5/3+4);
		
		err = CompressZ80 ( page, 0x4000, p+3, &destsize);

		*(p) = (unsigned char)(destsize & 0x00ff);
		*(p+1) = (unsigned char)(destsize >> 8);
		*(p+2) = ID;
		
		err = Buffer2File ( p, destsize+3 );
	}
	else
	{
		p = (unsigned char *)malloc (0x4000*5/3+4);
		*(p) = 0;
		*(p+1) = 0;
		*(p+2) = ID;
		err = Buffer2File ( p, 0x4000 + 3 );
	}

	free (p);

 	return err;
}



// -----	save .z80 file; version 2.01 --------------------------------------------------
int SaveZ80 (char *fileName, char *shortDesc)
{	int err;
	int i;
	int compression = 1;

	if (buff != NULL) free(buff);
	buff = (unsigned char *) malloc (128 * 1024);
	total_size = 0;

	struct {	char 					// Z80 Version 1.45 Header
			a,f,c,b,l,h,pcl,pch,spl,sph,
			i,r,data,e,d,c2,b2,e2,d2,l2,h2,a2,f2,
			yl,yh,xl,xh,iff1,iff2,im,
									// Z80 Version 2.01 enhanced header (if PC=0)
			h2lenl,h2lenh,				// size of header extension: 23 for version 2.01 files
			npcl,npch,					// PC (instead of pcl+pch, which is set to 0)
			model,					// 	Value:	Meaning in v2.01	Meaning in v3.0
									//	0		48k				48k
									//	1		48k + If.1			48k + If.1
									//	2		SamRam			48k + M.G.T.
									//	3		128k				SamRam
									//	4		128k + If.1		128k
									//	5		-				128k + If.1
									//	6		-				128k + M.G.T.
			port_7ffd,				// If in SamRam mode, bitwise state of 74ls259.
									// For example, bit 6=1 after an OUT 31,13 (=2*6+1)
									// If in 128 mode, contains last OUT to 7ffd (paging control)
			if1paged,					// !=0 means: interface1 rom is paged in
			rldiremu,					// Bit 0: 1 if R register emulation on
									// Bit 1: 1 if LDIR emulation on
			port_fffd,					// Last OUT to fffd (soundchip register number)
			soundreg[16];				// Contents of the sound chip registers
		} head;

	memset ( &head, 0, sizeof(head) );			// head.pcl = head.pch = head.if1paged = 0;

	head.h2lenl	= 23;	head.h2lenh = 0;	// flag: version 2.01 header
	
	switch(g_b48kmode)
	{	case 1:	head.model = 0; break;	// ZX Spectrum 48k
		case 0:	head.model = 3; break;	// ZX 128
	}

	// head.port_fffd	= AYregnr;				// ZX 128 only: value of AY register select register
	// gp_mem_func.memcpy ( head.soundreg, AYreg, 16 );		// ZX 128 only: AY registers
	head.port_7ffd = g_Port7ffdValue;				// ZX 128 only: value of page control register

	head.im	= (unsigned char)(z80_get_reg (Z80_IM));
	head.i	= (unsigned char)(z80_get_reg (Z80_I));
	head.r	= (unsigned char)(z80_get_reg (Z80_R) & 0x7F);
	head.a	= (unsigned char)(z80_get_reg (Z80_AF) >> 8);
	head.a2	= (unsigned char)(z80_get_reg (Z80_AF2) >> 8);
	head.f	= (unsigned char)(z80_get_reg (Z80_AF) & 0x00ff);
	head.f2	= (unsigned char)(z80_get_reg (Z80_AF2) & 0x00ff);
	head.b	= (unsigned char)(z80_get_reg (Z80_BC) >> 8);		
	head.c	= (unsigned char)(z80_get_reg (Z80_BC) & 0x00ff);
	head.d	= (unsigned char)(z80_get_reg (Z80_DE) >> 8);		
	head.e	= (unsigned char)(z80_get_reg (Z80_DE) & 0x00ff);
	head.h	= (unsigned char)(z80_get_reg (Z80_HL) >> 8);
	head.l	= (unsigned char)(z80_get_reg (Z80_HL) & 0x00ff);
	head.xh	= (unsigned char)(z80_get_reg (Z80_IX) >> 8);
	head.xl	= (unsigned char)(z80_get_reg (Z80_IX) & 0x00ff);
	head.yh	= (unsigned char)(z80_get_reg (Z80_IY) >> 8);		
	head.yl	= (unsigned char)(z80_get_reg (Z80_IY) & 0x00ff);
	head.b2	= (unsigned char)(z80_get_reg (Z80_BC2) >> 8);
	head.c2	= (unsigned char)(z80_get_reg (Z80_BC2) & 0x00ff);
	head.d2	= (unsigned char)(z80_get_reg (Z80_DE2) >> 8);
	head.e2	= (unsigned char)(z80_get_reg (Z80_DE2) & 0x00ff);
	head.h2	= (unsigned char)(z80_get_reg (Z80_HL2) >> 8);
	head.l2	= (unsigned char)(z80_get_reg (Z80_HL2) & 0x00ff);
	head.sph	= (unsigned char)(z80_get_reg (Z80_SP) >> 8);
	head.spl	= (unsigned char)(z80_get_reg (Z80_SP) & 0x00ff);
	head.npch= (unsigned char)(z80_get_reg (Z80_PC) >> 8);
	head.npcl= (unsigned char)(z80_get_reg (Z80_PC) & 0x00ff);

	head.iff1 = (unsigned char)(z80_get_reg (Z80_IFF1));
	head.iff2 = (unsigned char)(z80_get_reg (Z80_IFF2));
	
	// head.data = ((RR&0x80)>>7) + bordercolor<<1;
	head.data = (( z80_get_reg (Z80_R) & 0x80)>>7) + ((ucFePortValue&0x07)<<1);

	err = Buffer2File ((unsigned char *)&head, 55);

/*	Now the compressed blocks follow. Each block has a 3 byte header:
		dc.w		length of data (without this header; low byte first)
		dc.b		page number of block
		dc.s		compressed data follows

	The pages are numbered, depending on the hardware mode, in the following way:

		Page		48 mode			128 mode			SamRam mode
		0		48K rom			rom (basic)		48K rom
		1		Interface I, Disciple or Plus D rom, according to setting
		2		-				rom (reset)		samram rom (basic)
		3		-				page 0			samram rom (monitor,..)
		4		8000-bfff			page 1			Normal 8000-bfff
		5		c000-ffff			page 2			Normal c000-ffff
		6		-				page 3			Shadow 8000-bfff
		7		-				page 4			Shadow c000-ffff
		8		4000-7fff			page 5			4000-7fff
		9		-				page 6			-
		10		-				page 7			-
		11		Multiface rom		Multiface rom		-

	In 48K mode, pages 4,5 and 8 are saved.
	In SamRam mode, pages 4 to 8 are saved.  
	In '128 mode, all pages from 3 to 10 are saved.

	The 128 has a memory map like:   Rom [switchable];   Ram 5;   Ram 2;   Ram [switchable]
	The paging is all controlled by OUT commands to 7FFDh (32765). This port is 
	write-only so a copy of the last byte sent to it is stored at 5B5Ch (23388).
	
	The byte to output is:
		Bits 0-2:	Ram for top 16k:	0 is the usual bank;   7 contains some scratchpads used by the 128k editor.
		Bit 3:	Screen select:		0 => Use the normal screen in bank 5;	1 => Use the shadow screen in bank 7.
		Bit 4:	ROM for lower 16k:	0 => 128k editor;					1 => 48k BASIC
		Bit 5:	Lock. If set, disables all further paging.
*/
	if (g_b48kmode == 1)
	{			err = save_block_z80 ( 4, (unsigned char *)&(GET48KMEM()[0x8000]), compression );
		if (!err)	err = save_block_z80 ( 5, (unsigned char *)&(GET48KMEM()[0xC000]), compression );
		if (!err)	err = save_block_z80 ( 8, (unsigned char *)&(GET48KMEM()[0x4000]), compression );
	}
	else
	{	
		for (i=0;i<=7 && !err;i++)   err = save_block_z80 ( i+3, GETPAGEPTR(i), compression );
	}

	// Compress and write snapshot to the VMU ...

	compress (buff, total_size, fileName, shortDesc);

	free (buff);
	buff = NULL;

	return err;
}
