/*
 * ----------------------------------------------------------------------------------
 * menu.c
 *
 * Functions to display the options screen, the virtual keyboard, to load and save
 * preferences.
 * 
 * Copyright (C) 2001-2004 Stephane Schmitz
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 * ----------------------------------------------------------------------------------
 * Version  	: 1.0
 * Author	: S.Schmitz
 * History  	: 	03/05/2003, V0.1
 *			- First release
 * Date		: 03/05/2003
 * ----------------------------------------------------------------------------------
 */

#include "screen.h"
#include "menu.h"
#include "dc_vmu.h"
#include "vmutest.hpp"
#include <kos.h>

extern uint16 vbuffer[];

myRECT      rc;

extern unsigned char keyb2[];
extern unsigned char screen2[];
extern unsigned short *palEntries;

// Change options here !

static struct opt options[] = {// {OPT_TYPE_COMMENT, -1, "*** Machine type ***", 3, START_Y, OPT_STATE_ON, 0},
			// {OPT_TYPE_COMMENT, -1, " ", 3, START_Y+1, OPT_STATE_ON, 0},
			{OPT_TYPE_CHECKBOX, 0, "Spectrum 48k", 3, START_Y, OPT_STATE_ON, MACHINE48K,0},
			{OPT_TYPE_CHECKBOX, 0, "Spectrum 128k", 3, START_Y+1, OPT_STATE_OFF, MACHINE128K,0},
			{OPT_TYPE_COMMENT, -1, " ", 3, START_Y+2, OPT_STATE_ON, 0,0},
			// {OPT_TYPE_COMMENT, -1, "*** Joy. emulation ***", 3, START_Y+5, OPT_STATE_ON, 0},
			// {OPT_TYPE_COMMENT, -1, " ", 3, START_Y+6, OPT_STATE_ON, 0},
			{OPT_TYPE_CHECKBOX, 1, "Sinclair1", 3, START_Y+3, OPT_STATE_ON, SINCLAIR1, 0},
			{OPT_TYPE_CHECKBOX, 1, "Sinclair2", 3, START_Y+4, OPT_STATE_OFF, SINCLAIR2, 0},
			{OPT_TYPE_CHECKBOX, 1, "Cursor", 3, START_Y+5, OPT_STATE_OFF, CURSOR, 0},
			{OPT_TYPE_CHECKBOX, 1, "User defined", 3, START_Y+6, OPT_STATE_OFF, CUSTOM_KEYS, 0},
			{OPT_TYPE_COMMENT, -1, " ", 3, START_Y+7, OPT_STATE_ON, 0, 0},
			// {OPT_TYPE_COMMENT, -1, "*** Misc. ***", 3, START_Y+11, OPT_STATE_ON, 0},
			// {OPT_TYPE_COMMENT, -1, " ", 3, START_Y+13, OPT_STATE_ON, 0},
			{OPT_TYPE_ON_OFF, 2, "Border emulation", 3, START_Y+8, OPT_STATE_OFF, BORDER, 0},
			{OPT_TYPE_ON_OFF, 3, "Bilinear filtering", 3, START_Y+9, OPT_STATE_OFF, FILTERING, 0},
			{OPT_TYPE_ON_OFF, 4, "Fast rendering", 3, START_Y+10, OPT_STATE_OFF, RENDERING, 0},
			{OPT_TYPE_KEY, 5, "Button UP", 3, START_Y+11, OPT_STATE_OFF, KEY_UP, 0},
			{OPT_TYPE_KEY, 6, "Button DOWN", 3, START_Y+12, OPT_STATE_OFF, KEY_DOWN, 0},
			{OPT_TYPE_KEY, 7, "Button LEFT", 3, START_Y+13, OPT_STATE_OFF, KEY_LEFT, 0},
			{OPT_TYPE_KEY, 8, "Button RIGHT", 3, START_Y+14, OPT_STATE_OFF, KEY_RIGHT, 0},
			{OPT_TYPE_KEY, 9, "Button FIRE1", 3, START_Y+15, OPT_STATE_OFF, KEY_FIREA, 0},
			{OPT_TYPE_KEY, 10, "Button FIRE2", 3, START_Y+16, OPT_STATE_OFF, KEY_FIREB, 0},
			{OPT_TYPE_ON_OFF, 11, "Use frameskip", 3, START_Y+17, OPT_STATE_OFF, FRAMESKIP, 0},
			{OPT_TYPE_ON_OFF, 12, "Fullscreen mode", 3, START_Y+18, OPT_STATE_OFF, FULLSCREEN, 0}
};

unsigned char *font=charrom;
extern unsigned short *CreateSpectrumPalette (void);

static const int UP_BORDER = 32;
static const int LEFT_BORDER = 32;

#define K_STATUS_NORMAL	0x00
#define K_STATUS_CAPS	0x01
#define K_STATUS_SYMBOL	0x02

struct st_key
{
	unsigned char x,y,value, ss_value, cs_value;
};
	
struct st_key matrix[4][10] = 	{ 	{{0,16,'1'},{24,16,'2'},{48,16,'3'},{72,16,'4'},{96,16,'5'},{120,16,'6'},{144,16,'7'},{168,16,'8'},{192,16,'9'},{216,16,'0','-',4}},
					{{8,64,'Q'},{32,64,'W'},{56,64,'E'},{80,64,'R'},{104,64,'T'},{128,64,'Y'},{152,64,'U'},{176,64,'I'},{200,64,'O'},{224,64,'P','"'}},
					{{16,112,'A'},{40,112,'S'},{64,112,'D'},{88,112,'F'},{112,112,'G'},{136,112,'H'},{160,112,'J'},{184,112,'K'},{208,112,'L'},{232,112,3}},
					{{0,160,1},{24,160,'Z'},{48,160,'X'},{72,160,'C'},{96,160,'V'},{120,160,'B'},{144,160,'N'},{168,160,'M'},{192,160,2},{216,160,' '}}
				};					

struct ini_file_opt
{
	int opt_num, opt_state;
};

int ascii2charrom(char chr)
{
	int x=(int)chr;

	if(x>=97&&x<=122) x+=160;
	else if(x>=65&&x<=90) x-=64;
	return x;
}

void text_out8 (uint16 *screen, const char *string, unsigned char bg, unsigned char fg, int x, int y, int transparant) 
{
	int char_width=DISPLAY_WIDTH/FONT_WIDTH;
	int xpos;
	int xx;
	int yy;
	
	// unsigned char pixel;
	unsigned short pixel;

	signed char *character_data;
	int i;
	int screen_x, screen_y;

	if(y>=(DISPLAY_HEIGHT/FONT_HEIGHT)) return;
	xpos=x;
	for(i=0; (xpos<char_width)&&string[i]; i++) 
	{
		character_data=font+(ascii2charrom(string[i])
				*((FONT_WIDTH/8)*FONT_HEIGHT));
		xx=0;
		yy=0;
		while(yy<FONT_HEIGHT) 
		{
			pixel=*(character_data+yy);
			pixel=pixel>>((FONT_WIDTH-1)-xx);
			if(pixel&1) 
			{
				pixel= *(palEntries+fg);
			} 
			else 
			{
				pixel= *(palEntries+bg);
			}

			if(!(transparant && (pixel==*(palEntries+bg)) && (xx&1) && (yy&1))) 
			{
				screen_x = (BORDER_LEFT+(xpos*FONT_WIDTH))+xx;
				screen_y = (y*FONT_HEIGHT)+yy;
				screen [ (DISPLAY_WIDTH * screen_y) + (screen_x)] = pixel;
			}

			xx++;
			if(xx==FONT_WIDTH) 
			{
				xx=0;
				yy++;
			}
		}
		xpos++;
	}
}

void text_out8b (uint16 *screen, const char *string, unsigned char bg, unsigned char fg, int x, int y, int transparant) 
{
	int xpos;
	int xx;
	int yy;
	
	// unsigned char pixel;
	unsigned short pixel;

	signed char *character_data;
	int i;
	int screen_x, screen_y;

	if(y >= DISPLAY_HEIGHT) return;

	xpos=x;
	for(i=0; (xpos<DISPLAY_WIDTH) && string[i]; i++) 
	{
		character_data=font+(ascii2charrom(string[i])
				*((FONT_WIDTH/8)*FONT_HEIGHT));
		xx=0;
		yy=0;
		while(yy<FONT_HEIGHT) 
		{
			pixel=*(character_data+yy);
			pixel=pixel>>((FONT_WIDTH-1)-xx);

			if(pixel&1) 
			{
				pixel= *(palEntries+fg);
			} 
			else 
			{
				pixel= *(palEntries+bg);
			}

			if(!(transparant && (pixel==*(palEntries+bg)) && (xx&1) && (yy&1))) 
			{
				screen_x = xpos + xx;
				screen_y = y + yy;
				screen [ (DISPLAY_WIDTH * screen_y) + (screen_x)] = pixel;
			}

			xx++;
			if(xx==FONT_WIDTH) 
			{
				xx=0;
				yy++;
			}
		}
		xpos+=FONT_WIDTH;
	}
}

/*
 * Display a string. This version does not use any mask of any kind.
 */

void text_out8c (uint16 *screen, const char *string, unsigned char bg, unsigned char fg, int x, int y, int transparant) 
{
	int char_width=DISPLAY_WIDTH/FONT_WIDTH;
	int xpos;
	int xx;
	int yy;
	
	// unsigned char pixel;
	unsigned short pixel;

	signed char *character_data;
	int i;
	int screen_x, screen_y;
	int pix = 0;

	if(y>=(DISPLAY_HEIGHT/FONT_HEIGHT)) return;
	xpos=x;
	for(i=0; (xpos<char_width)&&string[i]; i++) 
	{
		character_data=font+(ascii2charrom(string[i])
				*((FONT_WIDTH/8)*FONT_HEIGHT));
		xx=0;
		yy=0;
		while(yy<FONT_HEIGHT) 
		{
			pixel=*(character_data+yy);
			pixel=pixel>>((FONT_WIDTH-1)-xx);
			if(pixel&1) 
			{
				pix = 1;
				pixel= *(palEntries+fg);
			} 
			else 
			{
				pix = 0;
				pixel= *(palEntries+bg);
			}

			if(!(transparant && (pixel==*(palEntries+bg)) && (xx&1) && (yy&1))) 
			{
				screen_x = (BORDER_LEFT+(xpos*FONT_WIDTH))+xx;
				screen_y = (y*FONT_HEIGHT)+yy;
				if (pix)
				{
					screen [ (DISPLAY_WIDTH * screen_y) + (screen_x)] = pixel;
				}
			}

			xx++;
			if(xx==FONT_WIDTH) 
			{
				xx=0;
				yy++;
			}
		}
		xpos++;
	}
}

void clearBackground (int bg)
{
	// change background color
	
	RectFill (vbuffer, 32, 32, 320-64, 240-48, bg);
	
	// Change border color
	
	RectFill (vbuffer, 0, 0, 320, 32, bg); // upper border
	RectFill (vbuffer, 0, 240-16, 320, 240, bg); // lower border
	RectFill (vbuffer, 0, 0, 32, 240-16, bg); // left border
	RectFill (vbuffer, 320-32, 0, 320, 240-16, bg); // right border
}

void displayMenuOptions (int bg, int fg, int sel_color)
{
	int i = 0;
	unsigned char *pt_buf;
	char buf[3];
	int xpos=0, ypos=0;

	//clearBackground(bg);
	drawBackground();

	for (i=0; i<NUM_OPTIONS; i++)
	{
		// if (options[i].type != OPT_TYPE_COMMENT)
		{
			if (options[i].disabled == 1)
			{
				// disabled options are displayed in another color
				text_out8c (vbuffer, options[i].opt_text, bg, fg/2, options[i].xpos, options[i].ypos, 0);
			}
			else
			{
				text_out8c (vbuffer, options[i].opt_text, bg, fg, options[i].xpos, options[i].ypos, 0);
			}

			xpos = (options[i].xpos*8 - 2) + BORDER_LEFT;

			// Depending on the option state, display additional informations

			// if on/off option, display on/off and change the colors of "on" or "off"

			if (options[i].disabled == 0)
			{
				if (options[i].type == OPT_TYPE_ON_OFF)
				{
					if (options[i].state == OPT_STATE_ON)
					{
						text_out8c (vbuffer, "on", bg, sel_color, options[i].xpos + 20, options[i].ypos, 0);
						text_out8c (vbuffer, "off", bg, fg, options[i].xpos + 23, options[i].ypos, 0);
					}
					else
					{
						text_out8c (vbuffer, "on", bg, fg, options[i].xpos + 20, options[i].ypos, 0);
						text_out8c (vbuffer, "off", bg, sel_color, options[i].xpos + 23, options[i].ypos, 0);
					}
				}
				else if (options[i].type == OPT_TYPE_CHECKBOX)
				{
					if (options[i].state == OPT_STATE_ON)
					{
						text_out8c (vbuffer, "(X)", bg, sel_color, options[i].xpos + 20, options[i].ypos, 0);
					}
					else
					{
						text_out8c (vbuffer, "( )", bg, sel_color, options[i].xpos + 20, options[i].ypos, 0);
					}
				}
				else if (options[i].type == OPT_TYPE_KEY)
				{
					// Handles the extra keys display
					
					if (options[i].state == 1 || options[i].state == 2 ||
						options[i].state == 3 ||
						options[i].state == 4 ||
						options[i].state == ' ')
					{							
						switch (options[i].state)
						{
							case 1:
								strcpy (buf, "cs");
								break;
							case 2:
								strcpy (buf, "ss");
								break;
							case 3:
								strcpy (buf, "en");
								break;
							case 4:
								strcpy (buf, "bs");
								break;
							case ' ':
								strcpy (buf, "sp");
								break;									
						}
					}
					else
					{
						buf[0] = (char)(options[i].state); 
						buf[1] = '\0';
					}

					text_out8c (vbuffer, buf, bg, sel_color, options[i].xpos + 20, options[i].ypos, 0);
				}
			} // if (options[i].disabled == 0)

		}
		
	} // for ()
}

//-----------------------------------------------------------------------------
//
// check the key mask
//

uint16 check_buttons(uint16 keyvalue,uint16 keymask)
{
	static uint16 key_pressed=0;
	uint16 key= key_pressed & keymask;

	if(key==0 && !(keyvalue & keymask))
	{	key_pressed |= keymask;		
		return KEY_ASSERTED;
	}
	
	if(key!=0 && (keyvalue & keymask))
	{	key_pressed &= ~keymask;
		return KEY_RELEASED;
	}

	if(key!=0 && !(keyvalue & keymask))
	{	return KEY_PUSHED;}

	return 0;
}

// -----------------------------------------------------------------------------
// Options screen
// Return: 0 if B button has been pressed, 1 if select has been pressed
// -----------------------------------------------------------------------------

int draw_options(int bg, int fg, int sel_color) 
{
	int i, x, y;
	int xpos=0, ypos=0;
	int bExit = 0;
	static int opt_num = 0;
	int nflip=0, bKeyPressed=1;
	int width = 0;
	char buf[3], key;
	unsigned char *pt_buf;
	int yoffset, xoffset;
	uint16	ret;
	cont_cond_t cond;
	uint8	down_timer=0;

	// Look for the largest string, which will be used to resize the rectangle around the options, draw the options states etc...

	for (i=0; i<NUM_OPTIONS; i++)
	{
		if (strlen (options[i].opt_text) > width && options[i].type != OPT_TYPE_COMMENT)
		{
			width = strlen (options[i].opt_text);
		}
	}

	width += 1;

	// Skip comments

	while (options[opt_num].type == OPT_TYPE_COMMENT)
	{
		opt_num++;
		if (opt_num>NUM_OPTIONS-1) opt_num = 0;
	}

	// main loop

	while (bExit == 0)
	{

		// change background color

		// GpRectFill(NULL, &gpDraw[nflip], 32, 32, 320-64, 240-48, bg);

		if(cont_get_cond(maple_first_controller(), &cond) <0) break;

		if ( !(cond.buttons & CONT_A))
		{
			return (0);
		} 
		else if ( !(cond.buttons & CONT_START))
		{
			return (1);
		}
		else
		{
			// Display options, and options states (selected, on/off ...)

			if (bKeyPressed)
			{
				bKeyPressed = 0;
				
				displayMenuOptions (bg, fg, sel_color);
				
			} // if (bKeyPressed)

			// coords where texture will be displayed

			rc.w = 2*512;
			rc.h = 2*512;
			rc.x = 0;
			rc.y = 0;

			renderscene (rc);

			xpos = (options[opt_num].xpos*8 - 2) + BORDER_LEFT;
			ypos = (options[opt_num].ypos*8 - 1);

			// Draw a rectangle around the selected option

			RectDraw (vbuffer, xpos, ypos, xpos + ((width+9)*8)+3, ypos + 8, sel_color);

			ret=check_buttons(cond.buttons,CONT_DPAD_UP);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				down_timer=0;
				
				// Clear the previous rectangle before moving it to the next option
				// Clear it in both screen buffers !
				
				bKeyPressed = 1;
				
				opt_num--;
				if (opt_num<0) opt_num = NUM_OPTIONS-1;

				// Skip comments and disabled options

				while (options[opt_num].type == OPT_TYPE_COMMENT || options[opt_num].disabled == 1)
				{
					opt_num--;
					if (opt_num<0) opt_num = NUM_OPTIONS-1;
				}

				displayMenuOptions (bg, fg, sel_color);

			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}

			ret=check_buttons(cond.buttons,CONT_DPAD_DOWN);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				bKeyPressed = 1;
				
				opt_num++;
				if (opt_num>NUM_OPTIONS-1) opt_num = 0;

				// Skip comments and disabled options

				while (options[opt_num].type == OPT_TYPE_COMMENT || options[opt_num].disabled == 1)
				{
					opt_num++;
					if (opt_num>NUM_OPTIONS-1) opt_num = 0;
				}

				displayMenuOptions (bg, fg, sel_color);
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}

			// User selects an option
			// Either check the option (if type is OPT_TYPE_CHECKBOX) or change the on/off flag
			// Make sure the other states of this group are changed

			ret=check_buttons(cond.buttons,CONT_X);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				Delay (150);
				
				bKeyPressed = 1;
					
				// two cases : we "click" an "on/off" option -> change the option state
				// we click on a checkbox option which is part of a group, then unselect the previously
				// selected option of this group, and select the current one

				// (1) ON/OFF option
				
				if (options[opt_num].type == OPT_TYPE_ON_OFF)
				{
					options[opt_num].state = !options[opt_num].state; // reverse option

					if (options[opt_num].opt_num == FULLSCREEN)
					{
						if (options[opt_num].state == OPT_STATE_ON)
						{
							// set border option to off and disable it
							options[8].state = OPT_STATE_ON;
							options[8].disabled = 1;
						}
						else
						{
							//options[8].disabled = 0;
							options[8].disabled = 1; // disable the 'border' option permanently
						}
					}
				}

				// (2) CHECKBOX

				else if (options[opt_num].type == OPT_TYPE_CHECKBOX)
				{
					i = 0;
					while (i < NUM_OPTIONS)
					{
						if (options[i].group == options[opt_num].group)
						{
							if (options[i].state == OPT_STATE_ON)
							{
								options[i].state = OPT_STATE_OFF; 
								break;
							}
						}
						i++;
					}

					options[opt_num].state = OPT_STATE_ON; 
				}

				// (3) USER DEFINED KEY

				else if (options[opt_num].type == OPT_TYPE_KEY)
				{
					key = display_virtual_keyboard2 (1, 4);
					if (key != 0) options[opt_num].state = key;
					Delay (500);
				}

				displayMenuOptions (bg, fg, sel_color);

			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}
			
		}
	}
}	

int get_option_state (int opt_num)
{
	int i=0;

	for (i=0; i<NUM_OPTIONS; i++)
	{
		if (options[i].type != OPT_TYPE_COMMENT)
		{
			if (options[i].opt_num == opt_num)
			{
				return (options[i].state);
			}
		}
	}

	return (-1);
}

void set_option_state (int opt_num, int state)
{
	int i=0;

	for (i=0; i<NUM_OPTIONS; i++)
	{
		if (options[i].type != OPT_TYPE_COMMENT)
		{
			if (options[i].opt_num == opt_num)
			{
				if (opt_num==KEY_UP||opt_num==KEY_DOWN||opt_num==KEY_LEFT||opt_num==KEY_RIGHT
					||opt_num==KEY_FIREA||opt_num==KEY_FIREB)
				{
					if (state == '#') // enter
					{
						options[i].state = 3;
					}
					else if (state == '?') // shift
					{
						options[i].state = 2;
					}
					else if (state == '.') // undefined state
					{
					}
					else if (state == '^') // space
					{
						options[i].state = ' ';
					}
					else 
					{
						options[i].state = state;
					}
				}
				else
				{
					options[i].state = state;
				}
			}
		}
	}
}

int save_options (void)
{
	const uint8 *buf;
	const uint8 *buf2;
	uint buf_len;
	char desc_long[512];
	int i, cnt = 0;
	struct ini_file_opt opt;
	file_t f; 
	uint8 buffer[4]; 
	char inifile[32];
	
	// Fill a buffer with the games options
	// first find the number of options

	for (i=0; i<NUM_OPTIONS; i++)
	{
		if (options[i].type != OPT_TYPE_COMMENT) cnt ++;
	}

	buf = malloc (sizeof (struct ini_file_opt) * cnt);
	buf2 = buf;
	for (i=0; i<NUM_OPTIONS; i++)
	{
		if (options[i].type != OPT_TYPE_COMMENT)
		{
			opt.opt_num = options[i].opt_num;
			opt.opt_state = options[i].state;
			memcpy (buf2, &opt, sizeof (struct ini_file_opt));
			buf2 = buf2 + sizeof (struct ini_file_opt);
		}
	}

	buf_len = sizeof (struct ini_file_opt) * cnt;
		
	strcpy (inifile, "/vmu/");
	strcat (inifile, GetVMUName());
	strcat (inifile, "/speccyal.ini");

	f = fs_open(inifile, O_WRONLY); 
	if (!f) 
	{
		free (buf);
		return 0;
	}

	//start of the header
 
	write_vms_header (f, 0);

	//actual content start here 

	fs_write (f, buf, buf_len); 
	fs_close(f);

	free (buf);

	return (1);
}

void write_vms_header (file_t f, int bKeyFile)
{
	int i;
	vmu_hdr_t header;
	unsigned char bitmap[512];

	strcpy (&header.desc_short[0], "DCSpeccyalK");
	strcpy (&header.desc_long[0], bKeyFile?"DCSpeccyalK custom keys":"DCSpeccyalK options");
	strcpy (&header.app_id[0], "DCSpeccyalK v0.1b");
	header.icon_cnt = 0;
	header.icon_anim_speed = 0;
	header.eyecatch_type = 0;
	header.crc = 0;
	header.data_len = 0;
	for (i=0; i<20; i++) header.reserved[i] = 0;
	for (i=0; i<16; i++) header.icon_pal[i] = 0;
	for (i=0; i<512; i++) bitmap[i] = 0;
	fs_write(f, &header, sizeof (vmu_hdr_t));
	fs_write(f, bitmap, 512);
}

// ------------------------------------------------------------
// This version only saves the custom keys (a profile is saved
// for each game, so the user does not have to enter the keys
// again)
// ------------------------------------------------------------

int save_game_keys (char *game_filename)
{
	file_t vmu; 
	int i, length;
	struct ini_file_opt opt;
	char *pszSavFileExtension;
	char *pszSlashPosition;
	char pszFileName[128];
	char pszVmuFile[128];

	pszSavFileExtension = (char *) (strrchr(game_filename, '.') + 1);
	pszSlashPosition = (char *) (strrchr(game_filename, '/') + 1);
	length = pszSavFileExtension - pszSlashPosition - 1;
	strncpy (pszFileName, pszSlashPosition, length);
	pszFileName[length] = '\0';
	
	// Take the filename, change the extension to 'KEY'. It will be the custom keys file, saved for
	// this game in particular
	
	strcat (pszFileName, ".KEY");
	strcpy (pszVmuFile, "/vmu/");
	strcat (pszVmuFile, GetVMUName());
	strcat (pszVmuFile, "/");
	strcat (pszVmuFile, pszFileName);
	
	vmu = fs_open(pszVmuFile ,O_WRONLY); 
	if (!vmu) 
	{ 
		fs_close (vmu);
		return(0);
	}

	write_vms_header (vmu, 1);

	// Save options into the file

	for (i=0; i<NUM_OPTIONS; i++)
	{
		// Only save the custom keys
		
		if (options[i].type == OPT_TYPE_KEY)
		{
			opt.opt_num = options[i].opt_num;
			opt.opt_state = options[i].state;
			fs_write (vmu, &opt, sizeof (struct ini_file_opt));
		}
	}

	fs_close (vmu);

	return (1);
}

int load_options (void)
{
	int i, cnt=0,j=0;
	uint8 *buf;
	uint buf_len;
	uint8 *compressed_buf;
	uint compressed_len;
	struct ini_file_opt opt;
	int nbread;
	char inifile[32];

	for (i=0; i<NUM_OPTIONS; i++)
	{
		if (options[i].type != OPT_TYPE_COMMENT) cnt ++;
	}

	// loading

	buf_len = sizeof (struct ini_file_opt) * cnt;
	buf = malloc (sizeof (struct ini_file_opt) * cnt);
	compressed_len = buf_len;
	compressed_buf = (uint8*) &buf;

	file_t vmu; 

	strcpy (inifile, "/vmu/");
	strcat (inifile, GetVMUName());
	strcat (inifile, "/speccyal.ini");

	vmu = fs_open(inifile, O_RDONLY); 
	if (vmu) 
	{ 
		fs_seek(vmu, 640, SEEK_SET); 
		nbread = fs_read(vmu, buf, buf_len); 
		fs_close(vmu); 
		i = 0;
		while (i < cnt)
		{
			memcpy(&opt, buf, sizeof(struct ini_file_opt));
			for (j=0; j < NUM_OPTIONS; j++)
			{
				if (options[j].opt_num == opt.opt_num && options[j].type != OPT_TYPE_COMMENT)
				{
					options[j].state = opt.opt_state; 
					break;
				}
			}
			i++;
			buf += sizeof(struct ini_file_opt);
		}

	}
	else
	{
		printf ("Problem au ndc_vmu_load() !\n");
	}

	options[8].disabled = 1; // disable the 'border' option permanently
	options[8].state = OPT_STATE_ON;
	options[9].disabled = 1; // disable the 'bilinear filtering' option permanently
	options[9].state = OPT_STATE_OFF;
	//options[10].disabled = 1; // disable the 'fast rendering' option permanently
	//options[10].state = OPT_STATE_ON;
	options[18].disabled = 1; // disable the 'full screen' option permanently
	options[18].state = OPT_STATE_OFF;

	free (buf);
	return (1);

}

int load_game_keys (char *game_filename)
{
	file_t		vmu, cd;
	long		nbread;
	struct ini_file_opt opt;
	int		i, count = 0, length;
	char *pszSavFileExtension;
	char *pszSlashPosition;
	char pszFileName[128];
	char pszVmuFile[128];
	char pszCdFile[128];
	char keys[10];

	pszSavFileExtension = (char *) (strrchr(game_filename, '.') + 1);
	pszSlashPosition = (char *) (strrchr(game_filename, '/') + 1);
	length = pszSavFileExtension - pszSlashPosition - 1;
	strncpy (pszFileName, pszSlashPosition, length);
	pszFileName[length] = '\0';
	
	// Take the filename, change the extension to 'KEY'. It will be the custom keys file, saved for
	// this game in particular
	
	strcat (pszFileName, ".KEY");
	strcpy (pszVmuFile, "/vmu/");
	strcat (pszVmuFile, GetVMUName());
	strcat (pszVmuFile, "/");
	strcat (pszVmuFile, pszFileName);
	
	*(pszSavFileExtension) = 'K';
	*(pszSavFileExtension+1) = 'E';
	*(pszSavFileExtension+2) = 'Y';

	// First, look on the cd for a .KEY file...

	cd = fs_open(game_filename ,O_RDONLY); 

	if (cd)
	{
		nbread = fs_read (cd, &keys[0], 9);

		// Keys contained are : up,down,left,right,a,b,x,y,start (for compatibilty with DreamSpec)
	
		set_option_state (KEY_UP, (int)keys[0]);
		set_option_state (KEY_DOWN, (int)keys[1]);
		set_option_state (KEY_LEFT, (int)keys[2]);
		set_option_state (KEY_RIGHT, (int)keys[3]);
		set_option_state (KEY_FIREA, (int)keys[4]);
		set_option_state (KEY_FIREB, (int)keys[5]);

		fs_close (cd);
	}
	else
	{
		fs_close (cd);

		// Otherwise look on the vmu ...

		vmu = fs_open(pszVmuFile ,O_RDONLY); 
		
		if (!vmu)
		{
			fs_close (vmu);
			return (0);
		}
		
		fs_seek(vmu, 640, SEEK_SET);	
		do
		{
			nbread = fs_read (vmu, &opt, sizeof (struct ini_file_opt));

			if (nbread == 0)
			{
				fs_close (vmu);
				return (0);
			}
			
			// Set the corresponding option with the value taken from the file

			i = 0;
			while (i < NUM_OPTIONS)
			{
				if (options[i].opt_num == opt.opt_num)
				{
					options[i].state = opt.opt_state; 
				}
				i++;
			}

			count ++;
		}
		while (count < 6); // 6 custom keys
		
		fs_close (vmu);
	}

	return (1);

}

void displayRectangle (int nflip, int cur_col, int cur_row, int color)
{
	if (cur_row != 18)
	{
		RectDraw (vbuffer, (cur_col*8) + BORDER_LEFT - 4,
				(cur_row*8) - 4, 
				(cur_col*8) + BORDER_LEFT + 8 + 4, 
				(cur_row*8) + 8 + 4, color);
	}
	else
	{
		RectDraw (vbuffer, (cur_col*8) + BORDER_LEFT - 4,
				(cur_row*8) - 4, 
				(cur_col*8) + BORDER_LEFT + 8*2 + 4, 
				(cur_row*8) + 8 + 4, color);
	}
}

void drawKeyboard (void)
{
	unsigned char *keyb;
	unsigned char r, g, b;
	int i, j;

	// Display KB
	
	keyb = &keyb2[0];
	keyb += 64;

	for (i=0; i<192; i++)
	{
		for (j=0; j<256; j++)
		{
			r = *keyb; keyb++;
			g = *keyb; keyb++;
			b = *keyb; keyb++;
			vbuffer [( (BORDER_TOP+1+i)*DISPLAY_WIDTH) + j + BORDER_LEFT+1] = RGB565 (r,g,b);
		}
	}
}

void drawBackground (void)
{
	unsigned char *pscr2;
	unsigned char r, g, b;
	int i, j;

	// Display KB
	
	pscr2 = &screen2[64];

	for (i=0; i<240; i++)
	{
		for (j=0; j<320; j++)
		{
			r = *pscr2; pscr2++;
			g = *pscr2; pscr2++;
			b = *pscr2; pscr2++;
			vbuffer [(i*DISPLAY_WIDTH) + j] = RGB565 (r,g,b);
		}
	}
}

char display_virtual_keyboard2 (int bg, int fg)
{	
	static int cur_row=0, cur_col=0;
	int i, j, x, y;
	int xpos=0, ypos=0;
	int bExit = 0;
	static int opt_num = 0;
	int nflip=0;
	int width = 0;
	char buf[3], key;
	int yoffset, xoffset;
	static int k_status;
	cont_cond_t cond;
	uint8 down_timer = 0;
	int ret;

    rc.w = 2*512;
	rc.h = 2*512;
	rc.x = 0;
	rc.y = 0;

	drawKeyboard ();
	fill_border(nflip, k_status);
	display_pressed_key_and_flip (nflip, cur_col, cur_row);
	
	while (bExit == 0)
	{
		renderscene(rc);

		if(cont_get_cond(maple_first_controller(), &cond) <0) break;

		if ( !(cond.buttons & CONT_A)) 
		{
			return (0);
		}

		ret=check_buttons(cond.buttons, CONT_Y);
		if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
		{
			if (k_status & K_STATUS_SYMBOL) k_status = k_status & ~K_STATUS_SYMBOL;
			else 
			{
				k_status = k_status | K_STATUS_SYMBOL;
				// k_status = k_status & ~K_STATUS_CAPS;
			}
						
			Delay (150);

			drawKeyboard ();
			fill_border(nflip, k_status);
			display_pressed_key_and_flip (nflip, cur_col, cur_row);
		}
		else
		{
			if(ret==KEY_PUSHED)
				down_timer++;			
			else
				down_timer=0;					
		}

		ret=check_buttons(cond.buttons, CONT_B);
		if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
		{
			if (k_status & K_STATUS_CAPS) k_status = k_status & ~K_STATUS_CAPS;
			else 
			{
				k_status = k_status | K_STATUS_CAPS;
				// k_status = k_status & ~K_STATUS_SYMBOL;
			}
						
			Delay (150);

			drawKeyboard ();
			fill_border(nflip, k_status);
			display_pressed_key_and_flip (nflip, cur_col, cur_row);
		}
		else
		{
			if(ret==KEY_PUSHED)
				down_timer++;			
			else
				down_timer=0;					
		}

		if ( !(cond.buttons & CONT_START)) 
		{
			bExit = 1;
		}
		else
		{
			ret=check_buttons(cond.buttons,CONT_DPAD_UP);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{				
				drawKeyboard ();
				fill_border(nflip, k_status);
				
				cur_row-- ;
				if (cur_row<0) 
				{
					cur_row = 3;
				}
				
				display_pressed_key_and_flip (nflip, cur_col, cur_row);
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}

			ret=check_buttons(cond.buttons,CONT_DPAD_DOWN);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{				
				drawKeyboard ();
				fill_border(nflip, k_status);
				
				cur_row ++;
				if (cur_row>3) 
				{
					cur_row = 0;
				}
				
				display_pressed_key_and_flip (nflip, cur_col, cur_row);
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}


			ret=check_buttons(cond.buttons,CONT_DPAD_RIGHT);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				drawKeyboard ();
				fill_border(nflip, k_status);
				
				cur_col ++;
				if (cur_col>9) cur_col = 0;
				
				display_pressed_key_and_flip (nflip, cur_col, cur_row);
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}

			ret=check_buttons(cond.buttons,CONT_DPAD_LEFT);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{								
				drawKeyboard ();
				fill_border(nflip, k_status);
				
				cur_col --;
				if (cur_col<0) 
				{
					cur_col = 9;
				}
				
				display_pressed_key_and_flip (nflip, cur_col, cur_row);
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}
		}			
		
	}

	// symbol shift or caps shift pressed ?

	if (k_status & K_STATUS_SYMBOL) 
	{
		return ((matrix[cur_row][cur_col]).ss_value);
	}
	else if (k_status & K_STATUS_CAPS) 
	{
		return ((matrix[cur_row][cur_col]).cs_value);
	}
	else return ((matrix[cur_row][cur_col]).value);
}

void display_pressed_key_and_flip (int nflip, int cur_col, int cur_row)
{	
	Delay (100);
	
	if (cur_row == 3 && cur_col == 9)
	{
		RectFill(vbuffer, LEFT_BORDER+(matrix[cur_row][cur_col]).x, UP_BORDER+(matrix[cur_row][cur_col]).y, LEFT_BORDER+(matrix[cur_row][cur_col]).x+32, UP_BORDER+(matrix[cur_row][cur_col]).y+32, 0);
		BitBlt(vbuffer, LEFT_BORDER+(matrix[cur_row][cur_col]).x+2, UP_BORDER+(matrix[cur_row][cur_col]).y+2, 32, 24, (unsigned char*)keyb2, (matrix[cur_row][cur_col]).x, (matrix[cur_row][cur_col]).y, 256, 192);
	}
	else
	{
		RectFill(vbuffer, LEFT_BORDER+(matrix[cur_row][cur_col]).x, UP_BORDER+(matrix[cur_row][cur_col]).y, LEFT_BORDER+(matrix[cur_row][cur_col]).x+24, UP_BORDER+(matrix[cur_row][cur_col]).y+24, 0);
		BitBlt(vbuffer, LEFT_BORDER+(matrix[cur_row][cur_col]).x+2, UP_BORDER+(matrix[cur_row][cur_col]).y+2, 24, 24, (unsigned char*)keyb2, (matrix[cur_row][cur_col]).x, (matrix[cur_row][cur_col]).y, 256, 192);
	}
}

// ---------------------------------------------------------------------------
// Display a black border around the virtual keyboard
// Also display the caps shift / symbol shift keys status
// ---------------------------------------------------------------------------

void fill_border (int nflip, int k_status)
{			
	RectFill (vbuffer, 0, 0, 320, 33, 0); // upper border
	RectFill (vbuffer, 0, 240-16, 320, 240, 0); // lower border
	RectFill (vbuffer, 0, 0, 33, 240-16, 0); // left border
	RectFill (vbuffer, 320-32, 0, 320, 240-16, 0); // right border
	
	if (k_status & K_STATUS_CAPS) 
	{
		text_out8 (vbuffer, "Caps Shift ON", 0, 15, 0, 1, 0);	// symbol shift
	}
	else
	{
		text_out8 (vbuffer, "Caps Shift OFF", 0, 15, 0, 1, 0);	// symbol shift
	}
	
	if (k_status & K_STATUS_SYMBOL) 
	{
		text_out8 (vbuffer, "Symbol Shift ON", 0, 15, 0, 2, 0);	// symbol shift
	}
	else
	{
		text_out8 (vbuffer, "Symbol Shift OFF", 0, 15, 0, 2, 0);	// symbol shift
	}
}

void RectFill (uint16 *screen, unsigned int x, unsigned int y, unsigned int x2, unsigned int y2, unsigned int bg)
{
	int i, j;

	for (i=y; i<y2; i++)
	{
		for (j=x; j<x2; j++)
		{
			screen  [ (i*DISPLAY_WIDTH) + j ] = palEntries[bg];
		}
	}
}

void RectDraw (uint16 *screen, unsigned int x, unsigned int y, unsigned int x2, unsigned int y2, unsigned int fg)
{	int i;

	for (i=x; i<x2; i++)
	{
		screen [y*DISPLAY_WIDTH + i] = *(palEntries+fg);
	}

	for (i=y; i<y2; i++)
	{
		screen [i*DISPLAY_WIDTH + x] = *(palEntries+fg);
	}

	for (i=x; i<x2; i++)
	{
		screen [ (y2)*DISPLAY_WIDTH + i] = *(palEntries+fg);
	}

	for (i=y; i<y2; i++)
	{
		screen [ i*DISPLAY_WIDTH + x2] = *(palEntries+fg);
	}
}

void BitBlt (uint16 *screen, unsigned int dx, unsigned int dy, unsigned int dw, unsigned int dh, unsigned char *src,
			 unsigned int sx, unsigned int sy, unsigned int sw, unsigned int sh)
{
	int x, y;
	unsigned char r, g, b;
	unsigned char *keyb;

	// Display KB
	
	// copy from position sx and sy of keyb2 and blit it to position (dx,dy) in vbuffer with a width = dw and 
	// height = dh

	for (y=0; y<dh; y++)
	{
		// Initial blitting position
		keyb = src + 64 + ((sy+y-1)*sw*3) + ((sx-1)*3);

		for (x=0; x<dw; x++)
		{
			r = *keyb; keyb++;
			g = *keyb; keyb++;
			b = *keyb; keyb++;
			screen [( (y+dy)*DISPLAY_WIDTH) + x + dx] = RGB565 (r,g,b);
		}
	}
}

void Delay (long delay)
{
	timer_spin_sleep (delay);
}

