/*
 * ----------------------------------------------------------------------------------
 * screen.c
 *
 * Display functions
 * 
 * Copyright (C) 2004 Stephane Schmitz
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 * ----------------------------------------------------------------------------------
 * Version  	: 1.0
 * Author	: S.Schmitz
 * History  	: 	28/04/2004, V0.1
 *			- First release
 * Date		: 28/04/2004
 * ----------------------------------------------------------------------------------
 */

#include "screen.h"
#include "speccyal.h"
#include <png/png.h>
#include "menu.h"
#include "dc_snap.h"

extern int errno;
extern struct options emu_options;

// Spectrum keyboard layout picture

pvr_init_params_t pvr_params = 
{
  /* Enable opaque and translucent polygons with size 16 */
  { PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0 },
  
  /* Vertex buffer size 512K */
  512*1024
};

////uint16 vbuffer[512 * 512] __attribute__ ((aligned (32)));
uint16 vbuffer[320 * 240] __attribute__ ((aligned (32)));
extern unsigned char screen1[];
pvr_ptr_t	texture __attribute__ ((aligned (32)));;

__inline void affiche_poly(myRECT rc)
{
   	pvr_vertex_t vert;
	pvr_poly_cxt_t cxt;
    pvr_poly_hdr_t hdr;

	pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 512, 512, texture, emu_options.filtering ?PVR_FILTER_BILINEAR:PVR_FILTER_NONE);
    pvr_poly_compile(&hdr, &cxt);
    pvr_prim(&hdr, sizeof(hdr));

    vert.argb = PVR_PACK_COLOR(1.0, 1.0, 1.0,1.0);
    vert.oargb = 0;
    vert.flags = PVR_CMD_VERTEX;
    
    vert.x = rc.x;
    vert.y = rc.y;
    vert.z = 2;
    vert.u = 0.0;
    vert.v = 0.0;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = rc.x + rc.w;
    vert.y = rc.y;
    vert.z = 2;
    vert.u = 1.0;
    vert.v = 0.0;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = rc.x;
    vert.y = rc.y + rc.h;
    vert.z = 2;
    vert.u = 0.0;
    vert.v = 1.0;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = rc.x + rc.w;
    vert.y = rc.y + rc.h;
    vert.z = 2;
    vert.u = 1.0;
    vert.v = 1.0;
    vert.flags = PVR_CMD_VERTEX_EOL;
    pvr_prim(&vert, sizeof(vert));		
	
}

__inline void renderscene (myRECT rc)
{
	sq_cpy (vram_s, vbuffer, 512*256*2);
}


/*__inline void renderscene (myRECT rc)
{
	// DMAtise la texture

	//dcache_flush_range(vbuffer, DISPLAY_WIDTH*DISPLAY_HEIGHT*2);
	//pvr_txr_load_dma(vbuffer, texture, DISPLAY_WIDTH*DISPLAY_HEIGHT*2, 1);
	dcache_flush_range(vbuffer, 512*512*2);
	pvr_txr_load_dma (vbuffer, texture, 512*512*2, 1);
	//memcpy (texture, vbuffer, 512*256*2);
	//sq_cpy (texture, vbuffer, 512*256*2);
        
    // dmarre la scne 3d
	#ifdef DEBUG
		//vid_border_color(255, 0, 0);
		pvr_wait_ready();
		//vid_border_color(0, 255, 0);	
	#else
		pvr_wait_ready();			
	#endif

	pvr_scene_begin();    
	pvr_list_begin(PVR_LIST_OP_POLY);
	affiche_poly(rc);        
    
    pvr_list_finish(); 
    pvr_scene_finish(); 

	#ifdef DEBUG
		//vid_border_color(255, 255, 0);
	#endif
}
*/

void affiche_background_poly (myRECT rc, int color)
{
	 pvr_poly_cxt_t cxt;
     pvr_poly_hdr_t hdr;
     pvr_vertex_t vert;

	 pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
     pvr_poly_compile(&hdr, &cxt);
     pvr_prim(&hdr, sizeof(hdr));
    		
	 vert.argb = color;
     vert.oargb = 0;
     vert.flags = PVR_CMD_VERTEX;

     vert.x = rc.x;
     vert.y = rc.y;
     vert.z = 1;
     vert.u = 0.0;
     vert.v = 0.0;
     pvr_prim(&vert, sizeof(vert));

     vert.x = rc.x + rc.w;
     vert.y = rc.y;
     vert.z = 1;
     vert.u = 1.0;
     vert.v = 0.0;
     pvr_prim(&vert, sizeof(vert));

     vert.x = rc.x;
     vert.y = rc.y+rc.h;
     vert.z = 1;
     vert.u = 0.0;
     vert.v = 1.0;
     pvr_prim(&vert, sizeof(vert));

     vert.x = rc.x + rc.w;
     vert.y = rc.y+rc.h;
     vert.z = 1;
     vert.u = 1.0;
     vert.v = 1.0;
     vert.flags = PVR_CMD_VERTEX_EOL;
     pvr_prim(&vert, sizeof(vert));
	
	
}

void render_background (myRECT rc, int color)
{
	// DMAtise la texture
    dcache_flush_range(vbuffer, DISPLAY_WIDTH*DISPLAY_HEIGHT*2);
	pvr_txr_load_dma(vbuffer, texture, DISPLAY_WIDTH*DISPLAY_HEIGHT*2, 1);
        
    // dmarre la scne 3d

	pvr_scene_begin();    
	pvr_list_begin(PVR_LIST_OP_POLY);
	affiche_background_poly (rc, color);
	affiche_poly(rc);
    pvr_list_finish();
    pvr_scene_finish();
}

// -----------------------------------------------------------------------------
// Options screen
// Return: 0 if B button has been pressed, 1 if select has been pressed
// -----------------------------------------------------------------------------

void setup_screen ()
{
	myRECT rc;
	int i,j;
	unsigned char *pscr1 = &screen1[64];
	unsigned char r, g, b;
	int bButton = 0;
	cont_cond_t cond;

	vid_init(DM_320x240_NTSC, PM_RGB565);

	// init all stuff here

    ////pvr_init_defaults();
    ////texture = pvr_mem_malloc (512 * 512 * 2);

	/*LoadPcx2Buffer ("/cd/img/screen1.pcx", &screen1[0]);
	//memcpy (&vbuffer[0], &screen1[0], 320*240*2);

	for (i=0; i<240; i++)
	{
		for (j=0; j<320; j++)
		{
			vbuffer [(i*512) + j] = screen1[ (320*i)+j];
		}
	}
	*/

	// png_to_texture("/cd/img/screen1a.png", texture, PNG_NO_ALPHA);

	for (i=0; i<240; i++)
	{
		for (j=0; j<320; j++)
		{
			r = *pscr1; pscr1++;
			g = *pscr1; pscr1++;
			b = *pscr1; pscr1++;
			vbuffer [(i*DISPLAY_WIDTH) + j] = RGB565 (r,g,b);
		}
	}

	rc.w = 2*512;
	rc.h = 2*512;
	rc.x = 0;
	rc.y = 0;
	renderscene (rc);

	while (!bButton)
	{
		if(cont_get_cond(maple_first_controller(), &cond) <0) break;
		if ( !(cond.buttons & CONT_B) || !(cond.buttons & CONT_A) ||
			!(cond.buttons & CONT_X) || !(cond.buttons & CONT_Y) )
		{
			bButton = 1;
		}
		renderscene (rc);
	}
}	

void release_screen_resources ()
{
	pvr_mem_free(texture);
}

