/*
 * ----------------------------------------------------------------------------------
 * slots.c
 *
 * Functions to display the memory snapshots slots
 * 
 * Copyright (C) 2004 Stephane Schmitz
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 * ----------------------------------------------------------------------------------
 * Version  	: 1.0
 * Author	: S.Schmitz
 * History  	: 	09/06/2004, V0.1
 *			- First release
 * Date		: 09/06/2004
 * ----------------------------------------------------------------------------------
 */

#include "screen.h"
#include "menu.h"
#include "slots.h"
#include "vmutest.hpp"
#include <kos.h>

extern uint16 vbuffer[];
extern unsigned short *palEntries;

myRECT      rc;
int selected = 0;
char slot1[17],slot2[17],slot3[17],slot4[17];
char freeBlocks[10];

void clearSlotsBackground (int bg)
{	
	RectFill (vbuffer, 0, 80, 320, 140, bg);
}

void displaySlotsOptions (int bg, int fg, int sel_color, int bSaveOrLoad)
{
	int ypos = 116, xpos = 70;

	RectFill (vbuffer, 50, 80, 270, 190, fg);
	RectFill (vbuffer, 52, 82, 268, 188, bg);

	if (bSaveOrLoad) 
	{
		text_out8b (vbuffer, "SAVE snapshot", bg, sel_color, 110, 85, 0);
	}
	else
	{
		text_out8b (vbuffer, "LOAD snapshot", bg, sel_color, 110, 85, 0);
	}

	text_out8b (vbuffer, "SLOT 1", bg, selected==0?sel_color:fg, xpos, ypos, 0);
	text_out8b (vbuffer, slot1, bg, fg, xpos+64, ypos, 0);
	ypos += 10;
	text_out8b (vbuffer, "SLOT 2", bg, selected==1?sel_color:fg, xpos, ypos, 0);
	text_out8b (vbuffer, slot2, bg, fg, xpos+64, ypos, 0);
	ypos += 10;
	text_out8b (vbuffer, "SLOT 3", bg, selected==2?sel_color:fg, xpos, ypos, 0);
	text_out8b (vbuffer, slot3, bg, fg, xpos+64, ypos, 0);
	ypos += 10;
	text_out8b (vbuffer, "SLOT 4", bg, selected==3?sel_color:fg, xpos, ypos, 0);
	text_out8b (vbuffer, slot4, bg, fg, xpos+64, ypos, 0);
	ypos += 20;
	text_out8b (vbuffer, "Free Blocks:", bg, fg, xpos, ypos, 0);
	text_out8b (vbuffer, freeBlocks, bg, fg, xpos+104, ypos, 0);
}

/*
 * -----------------------------------------------------------------------------------
 * Display the load/save memory slots
 *
 * INPUT	: bSaveOrLoad = 1 or 0 (display "SAVE" or "LOAD")
 * OUTPUT	: /
 * RETURN	: /
 * DATE		: 09/06/2004
 * -----------------------------------------------------------------------------------
 */

int displaySlots (int bg, int fg, int sel_color, int bSaveOrLoad)
{
	int i, x, y;
	int xpos=0, ypos=0;
	int bExit = 0;
	static int opt_num = 0;
	int bKeyPressed=1;
	uint16	ret;
	cont_cond_t cond;
	uint8	down_timer=0;

	while (!bExit)
	{
		if(cont_get_cond(maple_first_controller(), &cond) <0) break;
		if (!(cond.buttons & CONT_START)) 
			bExit = 0;
		else
			bExit = 1;
	}

	bExit = 0;

	strcpy (slot1, getFileShortDesc ("spkslot0.mem"));
	strcpy (slot2, getFileShortDesc ("spkslot1.mem"));
	strcpy (slot3, getFileShortDesc ("spkslot2.mem"));
	strcpy (slot4, getFileShortDesc ("spkslot3.mem"));
	sprintf (freeBlocks, "%d", GetFreeBlocks());

	if (strcmp (slot1, "") == 0) strcpy (slot1, "(FREE)");
	if (strcmp (slot2, "") == 0) strcpy (slot2, "(FREE)");
	if (strcmp (slot3, "") == 0) strcpy (slot3, "(FREE)");
	if (strcmp (slot4, "") == 0) strcpy (slot4, "(FREE)");

	// main loop

	while (bExit == 0)
	{
		// change background color

		if(cont_get_cond(maple_first_controller(), &cond) <0) break;

		if ( !(cond.buttons & CONT_A))
		{
			return (-1);
		} 
		else if ( !(cond.buttons & CONT_START))
		{
			if (bSaveOrLoad) return (selected); // save
			else return (selected+4); // load
		}
		else
		{
			// Display options, and options states (selected, on/off ...)

			if (bKeyPressed)
			{
				bKeyPressed = 0;
				
				displaySlotsOptions (bg, fg, sel_color, bSaveOrLoad);
				
			} // if (bKeyPressed)

			// coords where texture will be displayed

			rc.w = 2*512;
			rc.h = 2*512;
			rc.x = 0;
			rc.y = 0;

			renderscene (rc);

			// Draw a rectangle around the selected option

			// RectDraw (vbuffer, xpos, ypos, xpos + ((width+9)*8)+3, ypos + 8, sel_color);

			ret=check_buttons(cond.buttons,CONT_DPAD_UP);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				down_timer=0;
				
				bKeyPressed = 1;

				if (selected == 0) selected = 3; else selected--;
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}

			ret=check_buttons(cond.buttons,CONT_DPAD_DOWN);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				down_timer=0;

				bKeyPressed = 1;

				if (selected == 3) selected = 0; else selected++;
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}

			ret=check_buttons(cond.buttons,CONT_DPAD_LEFT);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				down_timer=0;
				
				bKeyPressed = 1;

				bSaveOrLoad = !bSaveOrLoad;
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}

			ret=check_buttons(cond.buttons,CONT_DPAD_RIGHT);

			if( (ret==KEY_ASSERTED) || (down_timer>KEY_DELAY))
			{
				down_timer=0;
				
				bKeyPressed = 1;

				bSaveOrLoad = !bSaveOrLoad;
			}
			else
			{
				if(ret==KEY_PUSHED)
					down_timer++;			
				else
					down_timer=0;					
			}
			
		}
	} // while 

}