/*
 * ----------------------------------------------------------------------------------
 * snaps.c
 * 
 * Copyright (C) 2001 Stephane Schmitz
 *
 * Files formats loading management - taken from speccyal/win32 and modified for the GP32 
 * handheld console
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms 
 * of the GNU General Public License as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA. 
 * ----------------------------------------------------------------------------------
 * Version  	: 1.0
 * Author	: S.Schmitz
 * History  	: 	28/04/2003
 *			- GP32: SNA, Z80 (48k & 128k), TAP, SLT, SCR files supported !
 *			26/04/2003
 *			- Started GP32 conversion (.SNA supported as of now)
 * Date		: 25/04/2003
 * ----------------------------------------------------------------------------------
 */

//#include <stdlib.h>
//#include <stdio.h>

#define _arch_dreamcast

#include <kos.h>
#include "snaps.h"
#include "tapreader.h"
#include "load+save.h"
#include "Z80_mame.h"

#define GETMEM()			GetMemPtr()
#define GET48KMEM()		(GetMemPtrEx()).Speccy48kMem
#define GET128KMEM()		(GetMemPtrEx()).Speccy128kMem
#define MEMPOKE(addr,v)		(GetMemPtrEx()).MemPoke (addr,v)
#define MODE48K()			(GetMemPtrEx()).b48kMode
#define RESETCPU()		z80_reset (NULL)
#define PAGECONTROL(value)		PageControl(value)

extern int g_b48kmode;
char g_szSnapFileName[256];
int bTapFile = 0;
extern unsigned char *CpuMem;
extern unsigned char ucFePortValue;
unsigned char *pSpecMem; // Spectrum memory pointer
char szCurrExt[256],szCurrFile[256],szCurrPath[256],szCurrDir[256];
file_t	hTapFile;
int counter = 0;
void printError (char *msg);

// .SLT levels table

struct SltDataBlockList	*sltDataLevelsList = NULL;
struct SltDataBlockList	*sltInitialDataLevelsListPointer = NULL;
void ReleaseMem (void);
void releaseSltData (struct SltDataBlockList *pSltData);
unsigned int LoadZ80Snapshot (char *szSnapFileName, char *szSavFile, unsigned char *vmuData);
unsigned int ReadDataBlock (file_t hSnaFile, unsigned char mode, unsigned char *vmuData);
unsigned int ReadDecrunchRamDump (file_t hSnaFile, unsigned long dwBytesToRead, unsigned long dwMemPos, unsigned char *vmuData);
unsigned int ReadAndDecrunch128kPage (file_t hSnaFile, unsigned long dwBytesToRead, unsigned char *ucDestBlock, unsigned char *vmuData);
unsigned int BuildSltLevelsTable (char *pszSnapFileName);
int load_tap (unsigned char blk_id, unsigned long start, unsigned long length, char *pszFileName);
int read_block (unsigned long start, unsigned long length, file_t hTapFile);
int skip_block (unsigned long length, file_t hTapFile);

void ReleaseMem (void)
{
	releaseSltData (sltInitialDataLevelsListPointer);
	sltDataLevelsList = NULL;
	sltInitialDataLevelsListPointer = NULL;

	if (hTapFile != NULL)
	{
		fs_close (hTapFile); // to release the lock
	}

	hTapFile = NULL;

	bTapFile = 0;
}

void releaseSltData (struct SltDataBlockList *pSltData)
{
	struct SltDataBlockList	*sltPreviousPointer;

	sltDataLevelsList = pSltData;
	
	while (sltDataLevelsList != NULL)
	{
		sltPreviousPointer = sltDataLevelsList;	
		sltDataLevelsList = sltDataLevelsList->nextBlock;
		free (sltPreviousPointer);
	}
}

int LoadFile (char *szSnapFileName)
{

	char        	*szFileExtension, *pszSavFileExtension, szSavFile[256]="";
	file_t	hSnaFile, h_rfile;
	unsigned long	BytesRead;
	int		bFound = 0, bSaveState = 0;
	static char	szLastLoadedTAP[256] = "";
	long		lFilePosition;
	Z80_Regs 	z80regs;
	unsigned char	regaf;
	struct Snapshot	SnapHeader;
	off_t curpos;

	// Passing an empty string instead of the filename means that the program has requested
	// an access to the .TAP file (and ONLY this file), to load a new level...

	if (strcmp (szSnapFileName, "") == 0)
	{
		szSnapFileName = szLastLoadedTAP;
	}
	else
	{
		bTapFile = 0;
	}

	strcpy (g_szSnapFileName, szSnapFileName);

	// Gets file name extension

	if (strrchr(szSnapFileName, '.') == NULL) 
	{
		return (0);
	}

	szFileExtension = (char *) (strrchr(szSnapFileName, '.') + 1);
	
	// replace file extension with .SAV (save-state loading)

	strcpy (szSavFile, szSnapFileName);
	pszSavFileExtension = (char *) (strrchr(szSavFile, '.') + 1);

	if (pszSavFileExtension == NULL)
	{
		return (0);
	}
	else
	{
		// Ugly string replace function :)
	
		*(pszSavFileExtension) = 'S';
		*(pszSavFileExtension+1) = 'A';
		*(pszSavFileExtension+2) = 'V';
		*(pszSavFileExtension+3) = '\0';
		
		// Try to open an .SAV file
		// If found -> we must load it !
		
		h_rfile = fs_open (szSavFile, O_RDONLY);

		if (h_rfile != 0)
		{
			fs_close(h_rfile);
			
			*(szFileExtension) = 'S';
			*(szFileExtension+1) = 'A';
			*(szFileExtension+2) = 'V';
			*(szFileExtension+3) = '\0';
		}
	}

	if (strcmp (szFileExtension, "SNA") == 0)

	// ------------------------------------------------------------------------------
	// ===============
	// Loads SNA files
	// ===============
	// ------------------------------------------------------------------------------

	{
		if (bTapFile == 1)
		{
			// Error closing open file handle ?

			fs_close (hTapFile);

			hTapFile = NULL;

			bTapFile = 0;
		}

		h_rfile = fs_open (szSnapFileName, O_RDONLY);

  		if (h_rfile == 0)
		{
			printError ("Cannot open .SNA file !\n");
			fs_close (h_rfile);
			return (0);
  		}
	
		BytesRead = fs_read (h_rfile, &SnapHeader, 27);
		
		if (BytesRead != 27)
		{
			printError ("Cannot read .SNA file header !\n");
			fs_close (h_rfile);
			return (0);
		}

		z80_get_context (&z80regs);
					
		RESETCPU();
				
		// Initialize z80 registers to start emulation

		z80regs.I = SnapHeader.regI;
		z80regs.AF2.b.h = SnapHeader.regAP;
		z80regs.AF2.b.l = SnapHeader.regFP;
		z80regs.HL2.b.h = SnapHeader.regHP;
		z80regs.HL2.b.l = SnapHeader.regLP;
		z80regs.DE2.b.h = SnapHeader.regDP;
		z80regs.DE2.b.l = SnapHeader.regEP;
		z80regs.BC2.b.h = SnapHeader.regBP;
		z80regs.BC2.b.l = SnapHeader.regCP;
		z80regs.HL.b.h = SnapHeader.regH;
		z80regs.HL.b.l = SnapHeader.regL;
		z80regs.DE.b.h = SnapHeader.regD;
		z80regs.DE.b.l = SnapHeader.regE;
		z80regs.BC.b.h = SnapHeader.regB;
		z80regs.BC.b.l = SnapHeader.regC;
		z80regs.IX.b.h = SnapHeader.regIXh;
		z80regs.IX.b.l = SnapHeader.regIXl;
		z80regs.IY.b.h = SnapHeader.regIYh;
		z80regs.IY.b.l = SnapHeader.regIYl;
				
		if (SnapHeader.interrupt & 0x04)
		{
			z80regs.IFF1 = 1;
			z80regs.IFF2 = 1;
		}

		z80regs.R = SnapHeader.regR;
		z80regs.AF.b.l = SnapHeader.regF;
		z80regs.AF.b.h = SnapHeader.regA;

		ucFePortValue = SnapHeader.BorderColor;
		
		z80regs.SP.d = (unsigned short)((SnapHeader.regSPh<<8)|(SnapHeader.regSPl));

		z80regs.IM = SnapHeader.InterruptMode;
				
		// Load snapshot RAM dump into spectrum memory 

		BytesRead = fs_read (h_rfile, (void *)&(GETMEM()[SCREEN_LOW]), 49152);
		
		if (BytesRead != 49152)
		{
			printError ("Cannot read .SNA main data block !\n");
			fs_close (h_rfile);
			return (0);
		}

		// PC is pushed onto stack
		// Get it back and increment stack pointer
		
		z80regs.PC.b.h = GETMEM()[z80regs.SP.d+1];
		z80regs.PC.b.l = GETMEM()[z80regs.SP.d];
		if (z80regs.PC.d == 0xffff) z80regs.PC.d = 0xfff4;
		z80regs.SP.d = z80regs.SP.d + 2;
		z80regs.AF.b.l = z80regs.IFF1 == 1 ? z80regs.AF.b.l|PV_FLAG : z80regs.AF.b.l&~PV_FLAG;
		
		z80_set_context (&z80regs);
		
		// Error closing open file handle ?

		fs_close (h_rfile);

		hSnaFile = NULL;

	}

	// ------------------------------------------------------------------------------
	// ===============
	// Loads Z80 files
	// ===============
	// ------------------------------------------------------------------------------

	else if ( (strcmp (szFileExtension,"Z80") == 0) || (strcmp (szFileExtension,"SAV") == 0) 
			|| (strcmp (szFileExtension,"MEM") == 0))
	{
		if (bTapFile == 1)
		{
			fs_close (hTapFile);
			bTapFile = 0;
		}

		LoadZ80Snapshot (szSnapFileName, szSavFile, NULL);

	}

	// ------------------------------------------------------------------------------
	// ===============
	// Loads SLT files
	// ===============
	// ------------------------------------------------------------------------------

	else if (strcmp (szFileExtension,"SLT") == 0)
	{
		// A .SLT file is primary a Z80 snapshot with additional information about the extra levels
		// which will be loaded at run time by using the DDCB opcode trap. Once this opcode has been 
		// encountered, a new level is loaded, by looking in the A register for the level number,
		// HL is pointing towards the destination memory

		// If it's the first time we load a tape file, load the .Z80 part data and build the levels table
		// the following times -> load data levels

		if (bTapFile == 0)
		{
			LoadZ80Snapshot (szSnapFileName, NULL, NULL);

			if (BuildSltLevelsTable (szSnapFileName) == 0)
			{
			}

			strcpy (szLastLoadedTAP, szSnapFileName);
			strcpy (g_szSnapFileName, szSnapFileName);
			bTapFile = 1;
		}
		else
		{
			// Load and decrunch a particular level data

			hSnaFile = fs_open (szSnapFileName, O_RDONLY);
	
  			if (hSnaFile == 0)
  			{
  				return (0);
  			}

			// Look for level info in the table we have built when we opened the .SLT file
			// It contains level numbers and level position in the .SLT file

			sltDataLevelsList = sltInitialDataLevelsListPointer;

			while (sltDataLevelsList != NULL && bFound == 0)
			{
				if (sltDataLevelsList->uiLevelNumber == (z80_get_reg (Z80_AF)>>8) )
				{
					bFound = 1;
				}
				else
				{
					sltDataLevelsList = sltDataLevelsList->nextBlock;
				}
			}				

			// Set file position to the wanted data level block

			curpos = fs_seek (hSnaFile, 0, SEEK_SET);
			curpos = fs_seek (hSnaFile, sltDataLevelsList->ulLevelDataFilePosition, SEEK_CUR);

			if (curpos <= 0)
			{
				fs_close (hSnaFile);
				return (0);
			}

			ReadDecrunchRamDump (hSnaFile, sltDataLevelsList->ulLevelDataLength, z80_get_reg (Z80_HL), NULL);

			fs_close (hSnaFile);

			hSnaFile = NULL;
		}
	}
	
	// ------------------------------------------------------------------------------
	// ===============
	// Loads TAP files
	// ===============
	// ------------------------------------------------------------------------------

	else if (strcmp (szFileExtension,"TAP") == 0)
	{
		// If it's the first time we load a tape file, simply open the file
		// the following times -> load code chunks

		if (bTapFile == 0)
		{
			tape_close ();

			strcpy (szLastLoadedTAP, szSnapFileName);
			strcpy (g_szSnapFileName, szSnapFileName);
			bTapFile = 1;

			tape_open (szSnapFileName);

			// Force autoload -> Z80 trap code will be triggered, and this function will be called

			if (g_b48kmode) LoadZ80Snapshot("/cd/lib/tape_48.z80", szSavFile, NULL);
			else LoadZ80Snapshot("/cd/lib/tape_128.z80", szSavFile, NULL);
		}
		else
		{

			// Load a chunk into memory
			// A register contains the chunk type (00 = header, ff = code)
			// IX contains the destination address
			// DE contains the length to load
			// e.g. when user types LOAD "", A=00, IX=variable address, DE=always 0x0011

			if(!load_tap((unsigned char)(z80_get_reg (Z80_AF)>>8), z80_get_reg (Z80_IX), z80_get_reg (Z80_DE), g_szSnapFileName))
			{
				regaf = z80_get_reg (Z80_AF);
				regaf = (regaf & CL_C_FLAG & CL_Z_FLAG) | (regaf & 0xf0);
				z80_set_reg (Z80_AF, regaf);
			}
			else 
			{
				regaf = z80_get_reg (Z80_AF);
				regaf = (regaf | CL_C_FLAG | CL_Z_FLAG);
				z80_set_reg (Z80_AF, regaf);
			}

			z80_set_reg (Z80_PC, 0x5e2); // load routine ret
		}

	} // if "TAP" ...	

	// ------------------------------------------------------------------------------
	// ===============
	// Loads SCR files
	// ===============
	// ------------------------------------------------------------------------------

	else if (strcmp (szFileExtension, "SCR") == 0)
	{
		if (bTapFile == 1)
		{
			fs_close (hTapFile);

			hTapFile = NULL;
			bTapFile = 0;
		}
		
		hSnaFile = fs_open (szSnapFileName, O_RDONLY);

	
		if (hSnaFile == 0)
		{
			printError ("Cannot open .SCR file !");
			return (0);
		}
		
		// Load screen file into spectrum screen memory 

		BytesRead = fs_read (hSnaFile, (void *)(&(GET48KMEM()[SCREEN_LOW])), 6912);
		
		if (BytesRead != 6912)
		{
			printError ("Error reading screen (.SCR) file !");
			fs_close (hSnaFile);
			return (0);
		}

		fs_close (hTapFile);

		hSnaFile = NULL;
	}

	return (1);
}

// Load a memory snapshot from the VMU memory 

int LoadFileFromVmu (char *fileName)
{
	unsigned char *buff;

	buff = NULL;

	uncompress (&buff, fileName);

	if (buff != NULL) 
	{
		LoadZ80Snapshot (NULL, NULL, buff);
		free (buff);
	}
	else
	{
		return (0);
	}

	return (1);
}

// Load a memory snapshot to the VMU memory

int SaveSnapShot (char *fileName, char *shortDesc)
{	
	SaveZ80 (fileName, shortDesc);

	return (1);
}

// --------------------------------------------------------------------------
// Load a chunk of .TAP file into memory
// --------------------------------------------------------------------------

/*int load_tap (unsigned char blk_id, unsigned long start, unsigned long length, file_t hTapFile)
{
	unsigned long blocklen;
	unsigned char id,ucBlockLenH,ucBlockLenL,ucChecksum;
	unsigned long dwBytesRead;

	// Read block length

	dwBytesRead = fs_read (hTapFile, (void *)&ucBlockLenH, 1);

	if (dwBytesRead != 1)
	{
		printError ("Can't read .TAP file block MSB !");
		return (0);
	}
	
	dwBytesRead = fs_read (hTapFile, (void *)&ucBlockLenL, 1);
		
	if (dwBytesRead != 1)
	{
		printError ("Can't read .TAP file block LSB !");
		return (0);
	}

	blocklen = (ucBlockLenH + (ucBlockLenL << 8)) - 2;	// length of this block
	
	// Header or data ?
	
	dwBytesRead = fs_read (hTapFile, (void *)&id, 1);
		
	if (dwBytesRead != 1)
	{
		printError ("Can't read .TAP file block type !");
		return (0);
	}	

	// A register stores correct id

   	if(length == blocklen)
	{	
		// length okay

      		read_block(start, length, hTapFile);			// read it in

		// Read checksum
		
		dwBytesRead = fs_read (hTapFile, (void *)&ucChecksum, 1);
		
		if (dwBytesRead != 1)
		{
			printError ("Can't read .TAP file checksum !");
			return (0);
		}

		z80_set_reg (Z80_IX, z80_get_reg (Z80_IX) + length);
	}
	else
	{	
		
		// too many bytes requested

		// wrong block type!

   		skip_block(blocklen, hTapFile);
		
		// Read checksum
		
		dwBytesRead = fs_read (hTapFile, (void *)&ucChecksum, 1);
		
		if (dwBytesRead != 1)
		{
			printError ("Can't read .TAP file checksum !");
			return (0);
		}

		z80_set_reg (Z80_IX, z80_get_reg (Z80_IX) + length);

		return(0);		// failed
	}

   return(1);
}

int read_block (unsigned long start, unsigned long length, file_t hTapFile)
{
	int loop;
	unsigned char ucByte;
	unsigned long dwBytesRead;

	for(loop=0;loop<length;loop++) 
	{
		dwBytesRead = fs_read (hTapFile, (void *)&ucByte, 1);
		
		if (dwBytesRead != 1)
		{
			printError ("Can't read .TAP file (read_block)!");
			return (0);
		}

		MEMPOKE (start, ucByte);

		start++;
	}

	return (1);
}

int skip_block (unsigned long length, file_t hTapFile)
{
	int loop;
	unsigned char ucWaste;
	unsigned long dwBytesRead;

	for(loop=0;loop<length;loop++) 
	{
		dwBytesRead = fs_read (hTapFile, (void *)&ucWaste, 1);
		
		if (dwBytesRead != 1)
		{
			printError ("Can't read .TAP file (skip_block)!");
			return (0);
		}
	}
	return (1);
}
*/

int load_tap (unsigned char blk_id, unsigned long start, unsigned long length, char *pszFileName)
{
	int i = 0;
	struct stTapBlock *block;
	file_t fd;
	long curpos;
	long read;
	unsigned char ch;

	block = getNextBlock ();
	
   	if (block != NULL && block->dataLength == length)
	{
		// copy block content to memory

		// open .tap file, seek block position, read content, copy content to memory

		fd = fs_open (pszFileName, O_RDONLY);
		if (!fd)
		{
			return (0);
		}

		curpos = fs_seek (fd, block->dataPos, SEEK_SET);
		if (curpos != block->dataPos)
		{
			fs_close (fd);
			return (0);
		}

		for (i=0; i<length; i++)
		{
			read = fs_read (fd, &ch, sizeof(char));
			MEMPOKE (start++, ch);
		}

		fs_close (fd);

		z80_set_reg (Z80_IX, z80_get_reg (Z80_IX) + length);
	}
	else
	{	
		z80_set_reg (Z80_IX, z80_get_reg (Z80_IX) + length);

		return(0);		// failed
	}

   return(1);
}

// --------------------------------------------------------------------------
// Load a Z80 snapshot (or a SLT, which is a Z80 with additional infos)
// --------------------------------------------------------------------------

unsigned int LoadZ80Snapshot (char *szSnapFileName, char *szSavFile, unsigned char *vmuData)
{
	file_t			hSnaFile;
	struct Z80Snapshot_v145	Z80Header_v145;
	struct Z80Snapshot_v201	Z80Header_v201;
	struct Z80Snapshot_v3	Z80Header_v3;
	unsigned long	BytesRead, dwFileSizeLow;
	unsigned int	bMode128 = 0, bError = 0;
	Z80_Regs 	z80regs;

	counter = 0; // index used to get bytes from the vmu data

	if (vmuData == NULL)
	{
		hSnaFile = fs_open (szSnapFileName, O_RDONLY);

		if (hSnaFile == 0)
		{
			return (0);
		}

		BytesRead = fs_read (hSnaFile, &Z80Header_v145, sizeof(struct Z80Snapshot_v145));

		if (BytesRead != sizeof(struct Z80Snapshot_v145))
		{
			fs_close (hSnaFile);
			return (0);
		}
	}
	else
	{
		memcpy (&Z80Header_v145, &vmuData[counter], sizeof(struct Z80Snapshot_v145));
		counter += sizeof(struct Z80Snapshot_v145);
	}
	
	z80_get_context (&z80regs);

	// RESETCPU();

	z80regs.I = Z80Header_v145.regIR;
	z80regs.AF2.b.h = Z80Header_v145.regAP;
	z80regs.AF2.b.l = Z80Header_v145.regFP;
	z80regs.HL2.b.h = Z80Header_v145.regHLPh;
	z80regs.HL2.b.l = Z80Header_v145.regHLPl;
	z80regs.DE2.b.h = Z80Header_v145.regDEPh;
	z80regs.DE2.b.l = Z80Header_v145.regDEPl;
	z80regs.BC2.b.h = Z80Header_v145.regBCPh;
	z80regs.BC2.b.l = Z80Header_v145.regBCPl;
	z80regs.HL.b.h = Z80Header_v145.regHLh;
	z80regs.HL.b.l = Z80Header_v145.regHLl;
	z80regs.DE.b.h = Z80Header_v145.regDEh;
	z80regs.DE.b.l = Z80Header_v145.regDEl;
	z80regs.BC.b.h = Z80Header_v145.regBCh;
	z80regs.BC.b.l = Z80Header_v145.regBCl;
	z80regs.IX.b.h = Z80Header_v145.regIXh;
	z80regs.IX.b.l = Z80Header_v145.regIXl;
	z80regs.IY.b.h = Z80Header_v145.regIYh;
	z80regs.IY.b.l = Z80Header_v145.regIYl;

	if (Z80Header_v145.IFF_1)
	{ 
		z80regs.IFF1 = 1;
		z80regs.IFF2 = 1;
	}

	z80regs.R = (Z80Header_v145.regRR | ((Z80Header_v145.DataByte1 & 0x01) << 7) );
	z80regs.AF.b.h = Z80Header_v145.regA;
	z80regs.AF.b.l = Z80Header_v145.regF;
	z80regs.SP.b.h = Z80Header_v145.regSPh;
	z80regs.SP.b.l = Z80Header_v145.regSPl;
	z80regs.PC.b.h = Z80Header_v145.regPCh;
	z80regs.PC.b.l = Z80Header_v145.regPCl;

	ucFePortValue = Z80Header_v145.DataByte1 & 0x07;

	//Z80Cpu1->IntMode=Z80Header_v145.DataByte2 & 0x03;

	z80regs.IM = Z80Header_v145.DataByte2 & 0x03;

	// File is a v2.01 or v3.00 if PC = 0

	if (Z80Header_v145.regPCl==0 && Z80Header_v145.regPCh==0)
	{
		// In this case, read v2.01 header 

		if (vmuData == NULL)
		{
			BytesRead = fs_read (hSnaFile, &Z80Header_v201, sizeof(struct Z80Snapshot_v201));
			
			if (BytesRead != sizeof(struct Z80Snapshot_v201))
			{
				fs_close (hSnaFile);
				return (0);
			}
		}
		else
		{
			memcpy (&Z80Header_v201, &vmuData[counter], sizeof(Z80Header_v201));
			counter += sizeof(Z80Header_v201);
		}

		// Read v3 header ?
		
		if ((unsigned long)(Z80Header_v201.HeaderBlockLengthHigh<<8|Z80Header_v201.HeaderBlockLengthLow)==54)
		{
			if (vmuData == NULL)
			{
				BytesRead = fs_read (hSnaFile, &Z80Header_v3, sizeof(struct Z80Snapshot_v3));
			
				if (BytesRead != sizeof(struct Z80Snapshot_v3))
				{
					fs_close (hSnaFile);
					return (0);
				}
			}
			else
			{
				memcpy (&Z80Header_v3, &vmuData[counter], sizeof(Z80Header_v3));
				counter += sizeof(Z80Header_v3);
			}
			
			// Is it a 48k game snapshot ?

			if (	Z80Header_v201.HardwareMode != 0 && 
				Z80Header_v201.HardwareMode != 1 && 
				Z80Header_v201.HardwareMode != 3)
			{
				bMode128 = 1;
			}

		}
		else
		{
			// Is it a 48k game snapshot ?

			if (Z80Header_v201.HardwareMode != 0 && Z80Header_v201.HardwareMode != 1)
			{
				bMode128 = 1;
			}
		}
	
		z80regs.PC.b.h = Z80Header_v201.regPCh;
		z80regs.PC.b.l = Z80Header_v201.regPCl;	

		// Ok, we can read and decrunch data blocks

		if (bMode128 == 0)
		{
			// 48k mode
			// 48k spectrum ram dumps = 3 pages

			bError = ReadDataBlock(hSnaFile, 0, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 0, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 0, vmuData);					
		}
		else
		{
			if (MODE48K())
			{			
				if (vmuData == NULL)
				{
					fs_close (hSnaFile);
				}

				hSnaFile = NULL;
				
				return (0);
			}

			PAGECONTROL(Z80Header_v201.BitwiseState);

			// 128k mode
			// 128k spectrum ram dumps = 8 pages

			bError = ReadDataBlock(hSnaFile, 1, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 1, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 1, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 1, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 1, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 1, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 1, vmuData);
			if (bError == 0) bError = ReadDataBlock(hSnaFile, 1, vmuData);
		}

		if (bError == 1)
		{
			if (vmuData == NULL)
			{
				fs_close (hSnaFile);
			}

			hSnaFile = NULL;

			return (0);
		}

	}
	else
	{
		// File is v1.45 or below if PC <> 0

		// Is data compressed ?

		if (Z80Header_v145.DataByte1 & 0x20)
		{
			// Decrunch data

			// Gets file size and reads (dwFileSize-header size) bytes into
			
			dwFileSizeLow = fs_total (hSnaFile);

			if (dwFileSizeLow <= 0)
			{
				return (0);
			}

			ReadDecrunchRamDump(hSnaFile, dwFileSizeLow-sizeof(struct Z80Snapshot_v145), SCREEN_LOW, vmuData);

		}
		else
		{
			// Data is not in a compressed format
		
			if (vmuData == NULL)
			{
				BytesRead = fs_read (hSnaFile, (void *)(&(GET48KMEM()[SCREEN_LOW])), 49152);
				
				if (BytesRead != 49152)
				{
					fs_close (hSnaFile);
					return (0);
				}
			}
			else
			{
				memcpy ((void *)(&(GET48KMEM()[SCREEN_LOW])), &vmuData[counter], 49152);
				counter += 49152;
			}


		} // Z80Header_v145.DataByte1 & 0x20

	} // Z80Header_v145.regPCl==0 && Z80Header_v145.regPCh==0				

	z80_set_context (&z80regs);

	// Close file

	if (vmuData == NULL)
	{
		fs_close (hSnaFile);
	}

	hSnaFile = NULL;

	return (1);
}

unsigned int ReadDataBlock (file_t hSnaFile, unsigned char mode, unsigned char *vmuData)
{
	unsigned long	dwBytesRead, dwMemPos;
	struct DataBlock	GameBlock;
	unsigned char 	*ucDestBlock;

	if (vmuData == NULL)
	{
		dwBytesRead = fs_read (hSnaFile, (void *)&GameBlock, sizeof(struct DataBlock));

		if (dwBytesRead != sizeof(struct DataBlock))
		{
			fs_close (hSnaFile);
			return (1);
		}
	}
	else
	{
		memcpy ((void *)&GameBlock, &vmuData[counter], sizeof(struct DataBlock));
		counter += sizeof(struct DataBlock);
	}

	// 48k or 128k snapshot ?

	if (mode == 0)
	{

		// -----------------------------------
		// 48K Mode
		// -----------------------------------

		// Checks page number

		if (GameBlock.ucPageNumber == 4)
		{
			dwMemPos=0x8000;
		}
		else
		if (GameBlock.ucPageNumber == 5)
		{
			dwMemPos=0xc000;
		}
		else
		if (GameBlock.ucPageNumber == 8)
		{
			dwMemPos=0x4000;
		}
		else
		if (GameBlock.ucPageNumber == 0)
		{
			dwMemPos=0x0000;
		}

		// Is data block compressed ?

		if ((unsigned long)(GameBlock.DataLengthHigh<<8|GameBlock.DataLengthLow) == 0xffff)
		{
			// No compression!
			
			if (vmuData == NULL)
			{
				dwBytesRead = fs_read (hSnaFile, (void *)(&(GET48KMEM()[dwMemPos])), BLOCKSIZE);
				
				if (dwBytesRead != BLOCKSIZE)
				{
					fs_close (hSnaFile);
					return (1);
				}
			}
			else
			{
				memcpy ((void *)(&(GET48KMEM()[dwMemPos])), &vmuData[counter], BLOCKSIZE);
				counter += BLOCKSIZE;
			}
				
		}
		else
		{
			// Decrunchs data block

			return (ReadDecrunchRamDump (hSnaFile, (unsigned long)(GameBlock.DataLengthHigh<<8|GameBlock.DataLengthLow), dwMemPos, vmuData));
		}
	}
	else
	{
	
		// -----------------------------------
		// 128K Mode
		// -----------------------------------

		// Checks page number

		switch (GameBlock.ucPageNumber)
		{
			case 3:
					ucDestBlock = GET48KMEM() + 0xc000;
				break;
			case 4:
					ucDestBlock = GET128KMEM();
				break;
			case 5:
					ucDestBlock = GET48KMEM() + 0x8000;
				break;
			case 6:
					ucDestBlock = GET128KMEM() + 0x4000;
				break;
			case 7:
					ucDestBlock = GET128KMEM() + 0x8000;
				break;
			case 8:
					ucDestBlock = GET48KMEM() + 0x4000;
				break;
			case 9:
					ucDestBlock = GET128KMEM() + 0xc000;
				break;
			case 10:
					ucDestBlock = GET128KMEM() + 0x10000;
				break;
		}

		// Compressed or uncompressed block ?

		if ((unsigned long)(GameBlock.DataLengthHigh<<8|GameBlock.DataLengthLow) == 0xffff)
		{
			// No compression!

			if (vmuData == NULL)
			{
				dwBytesRead = fs_read (hSnaFile, (void *)(ucDestBlock), BLOCKSIZE);
				
				if (dwBytesRead != BLOCKSIZE)
				{
					fs_close (hSnaFile);
					return (1);
				}
			}
			else
			{
				memcpy ((void *)(ucDestBlock), &vmuData[counter], BLOCKSIZE);
				counter += BLOCKSIZE;
			}
		}
		else
		{
			// Compressed data, decrunchs data block

			return (ReadAndDecrunch128kPage (hSnaFile, (unsigned long)(GameBlock.DataLengthHigh<<8|GameBlock.DataLengthLow), ucDestBlock, vmuData));
		}

	}

	return (0);
}

// Reads and uncompress a .Z80 snapshot RAM dump

unsigned int ReadDecrunchRamDump (file_t hSnaFile, unsigned long dwBytesToRead, unsigned long dwMemPos, unsigned char *vmuData)
{
	unsigned long		BytesRead, RamDumpPos = 0, DestPos = 0;
	unsigned char		*lpMem;
	unsigned int		EndBlock = 0, ED00 = 0;
	unsigned char		ByteToCopy, NumberOfCopies;
	long 				oldPos;
	off_t				curpos;
	// allocate a buffer to put snaphost ram dump

	// Allocates memory

	lpMem = (unsigned char *)malloc (dwBytesToRead);

	if (lpMem == NULL)
	{
		return (1);
	}

	// Loads compressed data into allocated memory buffer

	// curpos = fs_seek (hSnaFile, 0, SEEK_CUR);
	//gp_str_func.sprintf (buff, "BytesToRead = %d, oldPos = %d", dwBytesToRead, oldPos);
	//printError (buff);

	if (vmuData == NULL)
	{
		BytesRead = fs_read (hSnaFile, (void *)(&lpMem[0]), dwBytesToRead);
				
		if (BytesRead != dwBytesToRead)
		{
			free (lpMem);
			return (1);
		}
	}
	else
	{
		memcpy ((void *)(&lpMem[0]), &vmuData[counter], dwBytesToRead);
		counter += dwBytesToRead;
	}

	// Decrunchs ram dump

	while (EndBlock==0 && RamDumpPos <= dwBytesToRead)
	{
		// ED code ? if so let's see if next byte is ED too -> sequence of bytes
		
		if ( lpMem[RamDumpPos] == 0xed)
		{			
						
			if ( lpMem[RamDumpPos+1] == 0xed)
				{
					RamDumpPos+=2;

					// Copy sequence of bytes into spectrum memory

					ByteToCopy	= lpMem[RamDumpPos+1];
					NumberOfCopies	= lpMem[RamDumpPos];
					memset ((void *)((GetMemPtrEx()).Speccy48kMem + dwMemPos + DestPos), ByteToCopy, NumberOfCopies);
					DestPos += NumberOfCopies;
					RamDumpPos += 2;

					// End block marker ? in case where a byte is directly following
					// a single ED

					if (ED00 == 1)
					{
						if (  (lpMem[RamDumpPos+1] == 0xed) && (lpMem[RamDumpPos+2] == 0xed)
						&& ( lpMem[RamDumpPos+3] == 0x00) )
						{
							EndBlock = 1;
							RamDumpPos += 4;
						}
					}
					ED00 = 0;
				}
				else
				{
					// No compressed memory, simply copy ED byte

					(GetMemPtrEx()).Speccy48kMem[dwMemPos + DestPos] = lpMem[RamDumpPos];
					DestPos++;
					RamDumpPos++;
					ED00 = 1;
				}			
		}
		else
		{
			// Normal byte, simply copy it into the Spectrum memory

			(GetMemPtrEx()).Speccy48kMem[dwMemPos + DestPos] = lpMem[RamDumpPos];
			DestPos++;
			RamDumpPos++;
		}
	}

	// Unlocks and frees allocated memory

	free (lpMem);

	return (0);
}

// Reads and uncompress a 128k .Z80 snapshot page block

unsigned int ReadAndDecrunch128kPage (file_t hSnaFile, unsigned long dwBytesToRead, unsigned char *ucDestBlock, unsigned char *vmuData)
{
	unsigned long		BytesRead, RamDumpPos = 0, DestPos = 0;
	unsigned char 		*lpMem;
	unsigned int		EndBlock = 0, ED00 = 0;
	unsigned char		ByteToCopy, NumberOfCopies;
	
	// allocate a buffer to put snapshot ram dump

	// Allocates memory

	lpMem = (unsigned char *)malloc (dwBytesToRead);

	if (lpMem == NULL)
	{
		return (1);
	}

	if (vmuData == NULL)
	{
		BytesRead = fs_read (hSnaFile, (void *)(lpMem), dwBytesToRead);
				
		if (BytesRead != dwBytesToRead)
		{
			free (lpMem);
			return (1);
		}
	}
	else
	{
		memcpy ((void *)(lpMem), &vmuData[counter], dwBytesToRead);
		counter += dwBytesToRead;
	}

	// Decrunchs ram dump

	while (EndBlock==0 && RamDumpPos <= dwBytesToRead)
	{
		// ED code ? if so let's see if next byte is ED too -> sequence of bytes
		
		if ( lpMem[RamDumpPos] == 0xed)
		{			
						
			if ( lpMem[RamDumpPos+1] == 0xed)
				{
					RamDumpPos+=2;

					// Copy sequence of bytes into spectrum memory

					ByteToCopy	= lpMem[RamDumpPos+1];
					NumberOfCopies	= lpMem[RamDumpPos];
					memset ((void *)(ucDestBlock+DestPos), ByteToCopy, NumberOfCopies);
					DestPos += NumberOfCopies;
					RamDumpPos += 2;

					// End block marker ? in case where a byte is directly following
					// a single ED

					if (ED00 == 1)
					{
						if (  (lpMem[RamDumpPos+1] == 0xed) && (lpMem[RamDumpPos+2] == 0xed)
						&& (lpMem[RamDumpPos+3] == 0x00) )
						{
							EndBlock = 1;
							RamDumpPos += 4;
						}
					}
					ED00 = 0;
				}
				else
				{
					// No compressed memory, simply copy ED byte

					* (ucDestBlock+DestPos) = lpMem[RamDumpPos];
					DestPos++;
					RamDumpPos++;
					ED00 = 1;
				}			
		}
		else
		{
			// Normal byte, simply copy it into the Spectrum memory

			*(ucDestBlock+DestPos) = lpMem[RamDumpPos];
			DestPos++;
			RamDumpPos++;
		}
	}

	// Unlocks and frees allocated memory

	free (lpMem);

	return (0);
}

/*
	----------------------------------------------------------------------------------
	Build the SLT levels table
	----------------------------------------------------------------------------------
	INPUT:	pszSnapFileName: the .SLT file to open
	OUTPUT:	/
	PARMS:	/
	RETURN:	/
	DATE:	29/01/2003
	HISTORY: 27/04/2003 - Adapted for the GP32
	----------------------------------------------------------------------------------
*/

unsigned int BuildSltLevelsTable (char *pszSnapFileName)
{
	struct Z80Snapshot_v145	Z80Header_v145;
	struct Z80Snapshot_v201	Z80Header_v201;
	struct Z80Snapshot_v3	Z80Header_v3;
	struct DataBlock		GameBlock;
	unsigned long		BytesRead;
	long			FilePos;
	unsigned long		wLength = 0;
	unsigned char		ucWaste2[6];
	unsigned int		bMode128 = 0;
	unsigned int		uBlocks, i = 0, uiLevelNumber;
	struct SltDataBlock	slt_data_block;
	long			lLength;
	unsigned long		ulPreviousDataBlockLength;
	long			lFirstDataBlockPosition;
	struct SltDataBlockList  	*sltPreviousDataLevelPointer = NULL;
	file_t			hSltFile = NULL;
	unsigned int		index = 0;
	off_t			curpos;

	// Read the content of the .SLT file 

	hSltFile = fs_open (pszSnapFileName, O_RDONLY);

	if (hSltFile == 0)
	{
		printError ("ERROR at GpFileOpen !");
		return (0);
	}

	BytesRead = fs_read (hSltFile, &Z80Header_v145, sizeof(struct Z80Snapshot_v145));
	
	if (BytesRead != sizeof(struct Z80Snapshot_v145))
	{
		printError ("ERROR at GpFileRead !");
		fs_close (hSltFile);
		return (0);
	}

	// File is a v2.01 or v3.00 if PC = 0

	if (Z80Header_v145.regPCl==0 && Z80Header_v145.regPCh==0)
	{
		// In this case, read v2.01 header 

		BytesRead = fs_read (hSltFile, &Z80Header_v201, sizeof(struct Z80Snapshot_v201));
	
		if (BytesRead != sizeof(struct Z80Snapshot_v201))
		{
			printError ("ERROR at GpFileRead !");
			fs_close (hSltFile);
			return (0);
		}

	 	// Read v3 header ?
		
		if ((unsigned long)(Z80Header_v201.HeaderBlockLengthHigh<<8|Z80Header_v201.HeaderBlockLengthLow) == 54)
		{
			BytesRead = fs_read (hSltFile, &Z80Header_v3, sizeof(struct Z80Snapshot_v3));
		
			if (BytesRead != sizeof(struct Z80Snapshot_v3))
			{
				printError ("ERROR at GpFileRead !");
				fs_close (hSltFile);
				return (0);
			}
			
			// Is it a 48k game snapshot ?

			if (	
				Z80Header_v201.HardwareMode != 0 && 
				Z80Header_v201.HardwareMode != 1 && 
				Z80Header_v201.HardwareMode != 3
			   )
			{
				bMode128 = 1;
			}


		}
		else
		{
			// Is it a 48k game snapshot ?

			if (Z80Header_v201.HardwareMode != 0 && Z80Header_v201.HardwareMode != 1)
			{
				bMode128 = 1;
			}
		}

		// Main skipping loop

		// 7 blocks to skip for the 128k spectrum, 3 for the 48k

		if (bMode128 == 1)
		{
			uBlocks = 7;
		}
		else
		{
			uBlocks = 3;
		}

		while (i <= uBlocks)
		{
			// Read block header

			BytesRead = fs_read (hSltFile, &GameBlock, sizeof(struct DataBlock));
		
			wLength = (unsigned long)(GameBlock.DataLengthHigh<<8|GameBlock.DataLengthLow);

			if (BytesRead != sizeof(struct DataBlock))
			{
				printError ("ERROR at GpFileRead (skip) !");
				fs_close (hSltFile);
				return (0);
			}

			// Is data block compressed ?

			if (wLength == 0xffff)
			{
				// No compression!

				curpos = fs_seek (hSltFile, BLOCKSIZE, SEEK_CUR);

				if (curpos <= 0)
				{
					printError ("ERROR at GpFileSeek (skip) !");
					fs_close (hSltFile);
					return (0);
				}

			}
			else
			{
				// compression!

				curpos = fs_seek (hSltFile, wLength, SEEK_CUR);

				if (curpos <= 0)
				{
					printError ("ERROR at GpFileSeek (skip) !");
					fs_close (hSltFile);
					return (0);
				}

			}

			i++;
		}

		// Skip 3 null bytes (added for compatibility) and the 'SLT' characters

		BytesRead = fs_read (hSltFile, &ucWaste2[0], 6);
		
		if (BytesRead != 6)
		{
			printError ("ERROR at GpFileRead ('SLT') !");
			fs_close (hSltFile);
			return (0);
		}

		// Read and display level data

		i = 0;
		index = 0;
		releaseSltData (sltInitialDataLevelsListPointer);
		sltDataLevelsList = (struct SltDataBlockList *)malloc (sizeof (struct SltDataBlockList));
		sltInitialDataLevelsListPointer = sltDataLevelsList;

		do
		{
			BytesRead = fs_read (hSltFile, &slt_data_block, sizeof(struct SltDataBlock));
		
			if (BytesRead != sizeof(struct SltDataBlock))
			{
				printError ("ERROR at GpFileRead (end) !");
				fs_close (hSltFile);
				return (0);
			}

			if (slt_data_block.DataTypeHigh != 0 || slt_data_block.DataTypeLow != 0) // if not end of table
			{	
				index++;

				// Level number

				uiLevelNumber = (slt_data_block.DataIdentifierLow<<8) | slt_data_block.DataIdentifierHigh;

				// Level length

				lLength = ((slt_data_block.DataLengthLow2<<8 |slt_data_block.DataLengthLow1)<<16) | (slt_data_block.DataLengthHigh2<<8 |slt_data_block.DataLengthHigh1);

				sltDataLevelsList->uiLevelNumber = uiLevelNumber;
				sltDataLevelsList->ulLevelDataLength = lLength;
				sltDataLevelsList->nextBlock = (struct SltDataBlockList *)malloc (sizeof (struct SltDataBlockList));
				sltPreviousDataLevelPointer = sltDataLevelsList;
				sltDataLevelsList = sltDataLevelsList->nextBlock;

			}
		}
		while (slt_data_block.DataTypeHigh != 0 || slt_data_block.DataTypeLow != 0); // end of table

		sltPreviousDataLevelPointer->nextBlock = NULL; // end of data levels table

		curpos = fs_tell (hSltFile);

		if (curpos <= 0)
		{
			printError ("ERROR at GpFileSeek (end) !");
			fs_close (hSltFile);
			return (0);
		}

		lFirstDataBlockPosition = curpos;

		// We now have the data blocks beginning position
		// We must add the current file position to each data block length to find each data block position in the file

		i = 0;
		sltDataLevelsList = sltInitialDataLevelsListPointer;
		ulPreviousDataBlockLength = 0;

		while (i < index && sltDataLevelsList != NULL)
		{

			sltDataLevelsList->ulLevelDataFilePosition = lFirstDataBlockPosition + ulPreviousDataBlockLength;

			ulPreviousDataBlockLength += sltDataLevelsList->ulLevelDataLength;
			i++;
			sltDataLevelsList = sltDataLevelsList->nextBlock;
		}

	}
	else
	{
		// File is v1.45 or below if PC <> 0

	} // Z80Header_v145.regPCl==0 && Z80Header_v145.regPCh==0				

	fs_close (hSltFile);

	hSltFile = NULL;

	return (1);
}

void printError (char *msg)
{
	printf ("Error in snaps.c : %s\n", msg);
}
