/*
Copyright (C) 2004 Stephane Schmitz

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#ifndef TAPREADER
#define TAPREADER

#include <stdio.h>

// A Block header

struct stHeader
{
	unsigned char blockType; // 0,1,2,3
	char filename[10];
	unsigned char dataBlockLenL;
	unsigned char dataBlockLenH;
	unsigned char param1L;
	unsigned char param1H;
	unsigned char param2L;
	unsigned char param2H;
};

// Entry used in the levels list

struct stTapBlock
{
    int bHasHeader;
	int bIsHeader;	// is it a header or a data block ? if yes, header is filled
	struct stHeader header;
	long dataLength;
	long dataPos;
	struct stTapBlock *ptrNextEntry;
	struct stTapBlock *ptrPreviousEntry;
};

// Constructor / destructor

int newTapReader (char *pszFilename);
void releaseTapReader (void);
struct stTapBlock *getNextFreeEntry ();
struct stTapBlock *readLevels ();
int addBlock (int bHasHeader, long length, file_t fd);
int addHeader (long length, file_t fd);
struct stTapBlock *searchBlock (long length, unsigned char blockType);
struct stTapBlock *searchBlockByLength (long length);
struct stTapBlock *getNextBlock ();
struct stTapBlock *getPreviousBlock ();
struct stTapBlock *blockIteratorInit ();
struct stTapBlock *blockIteratorNext ();
struct stTapBlock *headerIteratorInit ();
struct stTapBlock *headerIteratorNext ();
char *getBlockDescription (struct stTapBlock *block);

// Tape deck simulation

int tape_open (char *filename);
int tape_close ();
int tape_rewind ();
int tape_play ();
int tape_stop ();

#endif // #ifndef TAPREADER
