/*
 * (c) 2002 Viktor Gars
 */

#define _arch_dreamcast
#include <kos.h>
#include "vmu.hpp"
#include "dc_vmu.h"

CVMU vmu;
extern char g_szSnapFileName[];

int GetFreeBlocks2 ();
int GetFreeBlocks2 ()
{
	vmu.Refresh();
	return (vmu.GetFreeBlocks (vmu.GetFirstVMUnit()->GetVMU()));
}

char *GetVMUName2 ();
char *GetVMUName2 ()
{
	vmu.Refresh();
	return (vmu.GetVMUName (vmu.GetFirstVMUnit()->GetVMU()));
}

void WriteFile(maple_device_t *dev, unsigned char *src_data, int src_size, char *fileName, char *shortDesc)
{
	//Neded datatypes
	vmu_pkg_t pkg;
	uint8 *pkg_out, *tmp_data, *tmp_data2;
	int pkg_size(0),tmp_size(0);
	char filePath[256];
	file_t fd;
	off_t size;

	//Create space for the compressed data
	
	tmp_data = new uint8[src_size+1000];
	tmp_size = src_size+1000;

	//Compress data
	vmu.BZCompress(tmp_data,(int*)&tmp_size,(uint8*)src_data, src_size);

	// check available size

	strcpy (filePath, "/vmu/");
	strcat (filePath, GetVMUName2());
	strcat (filePath, "/");
	strcat (filePath, fileName);

	// If wanting to save on a free slot : check that compressed file size is not > than available vmu space
	// otherwise check than compressed file size is not greater than occupied slot + 
	// free vmu space.

	fd = fs_open (filePath, O_RDONLY);
	if (!fd)
	{
		if ( (tmp_size/512) > GetFreeBlocks2())
		{
			printf ("not enough free blocks !\n");
			fs_close (fd);
			delete tmp_data;
			return;
		}
	}
	else
	{
		size = fs_total (fd);
		
		if ( (tmp_size/512) > ((size/512) + GetFreeBlocks2()))
		{
			printf ("not enough free blocks !\n");
			fs_close (fd);
			delete tmp_data;
			return;
		}
	}
	
	fs_close (fd);

	//make space for the compressed data + the size of the uncompressed size variable
	tmp_data2 = new uint8[tmp_size+sizeof(src_size)];

	//copy the uncompressed data size to the start of the memory
	memcpy(tmp_data2, &src_size, sizeof(src_size));

	//copy the compressed data into src_data (behind the uncompressed size
	memcpy(&tmp_data2[sizeof(src_size)],tmp_data,tmp_size);

	//free memory
	delete tmp_data;

	//get the total size of the data
	src_size=tmp_size+sizeof(src_size);

	//Here we create the file we write to the VMU
	//Add labels to the save file
	if (strcmp (shortDesc, "") == 0)
	{
		strcpy(pkg.desc_short, "(BASIC)");
	}
	else
	{
		// In case where the short description would be longer than 16 chars (possible if
		// the filename is very long) -> limit it to 15 chars + null terminator
		if (strlen (shortDesc) > 15)
		{
			strncpy (pkg.desc_short, shortDesc, 15);
			pkg.desc_short[15] = '\0';
		}
		else
		{
			strcpy (pkg.desc_short, shortDesc);
		}
	}
	strcpy(pkg.desc_long,"DCSpeccyalK v0.2 Snapshot");
	strcpy(pkg.app_id,"DCSPECCYAL");

	//Icon properties
	pkg.icon_cnt = 0;
	pkg.icon_anim_speed=0;
	pkg.eyecatch_type = VMUPKG_EC_NONE;

	//Here we put on the raw data that we want to write to the VMU
	pkg.data_len=src_size;
	pkg.data=tmp_data2;

	//Build vmu file and save it to pkg_out, the size is saved to pkg_size
	vmu_pkg_build(&pkg,&pkg_out,&pkg_size);

	//Write the file on the VMU
	vmu.Write(dev, fileName, pkg_out, pkg_size);

	//Clear memory
	delete pkg_out;
	delete tmp_data2;
}

void ReadFile(maple_device_t *dev, unsigned char **dest_data, char *fileName)
{
	vmu_pkg_t pkg;
	int dest_size(0), src_size(0);
	uint8 *src_data(NULL);

	if (!vmu.FileExist (dev, fileName)) return;

	//Read from VMU dev and file "Test.sys", place data into pkg
	vmu.Read(dev, fileName, &pkg);

	//Create space in memory for the data that we have read into memory
	src_size=pkg.data_len;
	src_data=new uint8[src_size];

	//copy the first part of the data, the uncompressed size
	memcpy(&dest_size,pkg.data,sizeof(src_size));
	
	//copy the compressed data from the pkg to src_data
	memcpy(src_data,&pkg.data[sizeof(src_size)],src_size);
	
	//Create space in memory for the uncompressed data
	*dest_data=new uint8[dest_size];
	
	//Decompress the data
	vmu.BZDecompress(*dest_data, &dest_size, src_data, src_size);

	strcpy (g_szSnapFileName, "/cd/");
	strcat (g_szSnapFileName, pkg.desc_short);
	strcat (g_szSnapFileName, ".Z80");

	//Clear memory
	delete src_data;
	//delete dest_data;
}

extern "C" { 
	void compress (unsigned char *data, int length, char *fileName, char *shortDesc)
	{
		vmu.Refresh();
		WriteFile(vmu.GetFirstVMUnit()->GetVMU(), data, length, fileName, shortDesc);
	}
}

extern "C" { 
	void uncompress (unsigned char **dest_data, char *fileName)
	{
		vmu.Refresh();
		ReadFile(vmu.GetFirstVMUnit()->GetVMU(), dest_data, fileName);
	}
}

extern "C" { 
	char *getFileShortDesc (char *filename)
	{
		static char shortname[17];

		strcpy (shortname, "");

		vmu.Refresh();
		get_file_short_desc (shortname, vmu.GetFirstVMUnit()->GetVMU(), filename);
		return (&shortname[0]);
	}
}

extern "C" { 
	int GetFreeBlocks ()
	{
		vmu.Refresh();
		return (vmu.GetFreeBlocks (vmu.GetFirstVMUnit()->GetVMU()));
	}
}

extern "C" { 
	char *GetVMUName ()
	{
		vmu.Refresh();
		return (vmu.GetVMUName (vmu.GetFirstVMUnit()->GetVMU()));
	}
}
