/*
Speccyal - The 48k Spectrum emulator for Win9x
Multi-Z80 CPU emulator by Neil Bradley (neil@synthcom.com)
Emulation of the AY-3-8910 / YM2149 sound chip based on various code snippets 
by Ville Hallik, Michael Cuddy, Tatsuyuki Satoh, Fabrice Frances, Nicola Salmoria.

Copyright (C) 1999-2000 Stephane Schmitz 

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; either 
version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; 
if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
MA 02111-1307, USA. 
*/

#ifndef Z80CPUDEFS
#define Z80CPUDEFS

#define LSB_FIRST	1

/* ================================================================ */
/* = Flags															*/
/* ================================================================ */

#define C_FLAG			0x01
#define N_FLAG			0x02
#define PV_FLAG			0x04
#define BIT3_FLAG		0x08
#define H_FLAG			0x10
#define BIT5_FLAG		0x20
#define Z_FLAG			0x40
#define S_FLAG			0x80

// To clear FLAGS

#define CL_C_FLAG		0xfe
#define CL_N_FLAG		0xfd
#define CL_PV_FLAG		0xfb
#define CL_BIT3_FLAG	0xf7
#define CL_H_FLAG		0xef
#define CL_BIT5_FLAG	0xdf
#define CL_Z_FLAG		0xbf
#define CL_S_FLAG		0x7f

/* ================================================================ */
/* = Other vars														*/
/* ================================================================ */

#define SPEC48KFRAME	69888
#define ROMSIZE48K		16384
#define SPECMEMSIZE		0xffff

/* ================================================================ */
/* = Memory map														*/
/* ================================================================ */

#define ROM_LOW		0
#define ROM_HIGH	16383
#define SCREEN_LOW	16384
#define SCREEN_HIGH	22527
#define ATTR_LOW	22528
#define ATTR_HIGH	23295

/* ================================================================ */
/* = Joysticks														*/
/* ================================================================ */

#define Fuller		0x7f
#define Kempston	0x1f
#define Sinclair1	0xeffe
#define Sinclair2	0xf7fe

#define Full_FIRE	0x80	// Fuller fire
#define Full_UP		0x01	// Fuller up
#define Full_DOWN	0x02	// Fuller down
#define Full_LEFT	0x04	// Fuller left
#define Full_RIGHT	0x08	// Fuller right
#define Kem_FIRE	0x10	// Kempston fire
#define Kem_UP		0x08	// Kempston up
#define Kem_DOWN	0x04	// Kempston down
#define Kem_LEFT	0x02	// Kempston left
#define Kem_RIGHT	0x01	// Kempston right
#define Sin1_FIRE	0x01	// Sinclair1 fire
#define Sin1_UP		0x02	// Sinclair1 up
#define Sin1_DOWN	0x04	// Sinclair1 down
#define Sin1_LEFT	0x10	// Sinclair1 left
#define Sin1_RIGHT	0x08	// Sinclair1 right
#define Sin2_FIRE	0x10	// Sinclair2 fire
#define Sin2_UP		0x08	// Sinclair2 up
#define Sin2_DOWN	0x04	// Sinclair2 down
#define Sin2_LEFT	0x01	// Sinclair2 left
#define Sin2_RIGHT	0x02	// Sinclair2 right

/* ================================================================ */
/* = Rows of spectrum keyboard keys								  = */
/* ================================================================ */

#define SZXCV		0	// SHIFT Z X C V
#define ASDFG		1	// A S D F G
#define QWERT		2	// Q W E R T
#define N12345		3	// 1 2 3 4 5
#define N09876		4	// 0 9 8 7 6
#define POIUY		5	// P O I U Y
#define ELKJH		6	// ENTER L K J H
#define SPSSMN		7	// SPACE SYMBOL SHIFT M N

/* ================================================================ */
/* = Used to SET/RESET bits (SET/RES opcodes)						*/
/* ================================================================ */

#define BIT_0		0x01
#define BIT_1		0x02
#define BIT_2		0x04
#define BIT_3		0x08
#define BIT_4		0x10
#define BIT_5		0x20
#define BIT_6		0x40
#define BIT_7		0x80

typedef enum {
      REG_AF=0,
      REG_BC,
      REG_DE,
      REG_HL,
      REG_IX,
      REG_IY,
      REG_PC,
      REG_SP,
      REG_AF2,
      REG_BC2,
      REG_DE2,
      REG_HL2,
      REG_IFF1,        
      REG_IFF2,        
      REG_R,
      REG_I,
      REG_A,
      REG_F,
      REG_B,
      REG_C,
      REG_D,
      REG_E,
      REG_H,
      REG_L,
      REG_IXH,
      REG_IXL,
      REG_IYH,
      REG_IYL,
      REG_A2,
      REG_F2,
      REG_B2,
      REG_C2,
      REG_D2,
      REG_E2,
      REG_H2,
      REG_L2,
      REG_PCH,
      REG_PCL,
      REG_SPH,
      REG_SPL,
	  REG_IM
} z80register;

typedef enum {
      DIK_A=0,
      DIK_S,
      DIK_D,
      DIK_F,
      DIK_G,
      DIK_LSHIFT,
	  DIK_RSHIFT,
      DIK_Z,
      DIK_X,
      DIK_C,
      DIK_V,
      DIK_Q,
      DIK_W,
      DIK_E,        
      DIK_R,        
      DIK_T,
      DIK_1,
      DIK_2,
      DIK_3,
      DIK_4,
      DIK_5,
      DIK_6,
      DIK_7,
      DIK_8,
      DIK_9,
      DIK_0,
      DIK_P,
      DIK_O,
      DIK_I,
      DIK_U,
      DIK_Y,
      DIK_RETURN,
      DIK_L,
      DIK_K,
      DIK_J,
      DIK_H,
      DIK_SPACE,
      DIK_LCONTROL,
	  DIK_RCONTROL,
      DIK_M,
      DIK_N,
      DIK_B
} spec_keys;

typedef union
{
	struct { unsigned char l,h; } B;
	unsigned short W;
	unsigned long L;
} DoubleReg;

/* ================================================================ */
/* = CPU Registers													*/
/* ================================================================ */

struct Registers
{
	unsigned char I; // Interrupt vector system
	unsigned char R; // Used to refresh dynamic memory locations
	unsigned char RIFF1, RIFF2;
	int HALTED;
	DoubleReg BC;
	DoubleReg AF;
	DoubleReg AF2;
	DoubleReg DE;
	DoubleReg HL;
	DoubleReg IX;
	DoubleReg IY;
	DoubleReg BCP;	
	DoubleReg DEP;
	DoubleReg HLP;
	DoubleReg SP; // Stack pointer
	DoubleReg PC; // Program counter
};

#ifndef USHORT
#define USHORT unsigned short
#endif

#ifndef UWORD
#define UWORD unsigned short
#endif

#ifndef UCHAR
#define UCHAR unsigned char
#endif

#ifndef UBYTE
#define UBYTE unsigned char
#endif

#define MAX_FILE	256
#define MAX_DESC	1024

extern char IntMode;

struct Snapshot
{
	unsigned char		regI;
	unsigned char		regLP,regHP,regEP,regDP,regCP,regBP;
	unsigned char		regFP,regAP;
	unsigned char		regL,regH,regE,regD,regC,regB;
	unsigned char 		regIYl,regIYh,regIXl,regIXh;
	unsigned char		interrupt;
	unsigned char		regR,regF,regA;
	unsigned char 		regSPl;
	unsigned char 		regSPh;
	unsigned char		InterruptMode;
	unsigned char		BorderColor;
};

struct Z80Snapshot_v145
{
	unsigned char		regA,regF,regBCl,regBCh;
	unsigned char		regHLl,regHLh,regPCl,regPCh,regSPl,regSPh;
	unsigned char		regIR,regRR,DataByte1,regDEl,regDEh;
	unsigned char		regBCPl,regBCPh,regDEPl,regDEPh,regHLPl,regHLPh;
	unsigned char 		regAP,regFP,regIYl,regIYh,regIXl,regIXh;
	unsigned char		IFF_1,IFF_2,DataByte2;
};

struct Z80Snapshot_v201
{
	unsigned char	HeaderBlockLengthLow,HeaderBlockLengthHigh;
	unsigned char	regPCl,regPCh;
	unsigned char	HardwareMode;
	unsigned char	BitwiseState;
	unsigned char	IF1RomPaged;
	unsigned char	R_LDIR_Emulation;
	unsigned char	LastOutFffd;
	unsigned char	SoundChipRegisters[16];		
};

struct Z80Snapshot_v3
{	
	unsigned char	LowTStateCounterLow,LowTStateCounterHigh;
	unsigned char	HighTStateCounter;
	unsigned char	SpectatorFlagByte;
	unsigned char	MGTRomPaged;
	unsigned char	MFRomPaged;
	unsigned char	IsRom;
	unsigned char	IsRom2;
	unsigned char	KeyboardMappings[10];
	unsigned char	MappingKeys1[10];
	unsigned char	MGTType;
	unsigned char	DiscInhibitButtonStatus;
	unsigned char	DiscInhibitFlag;
};

struct DataBlock
{	
	unsigned char	DataLengthLow,DataLengthHigh;
	unsigned char	ucPageNumber;
};

// .TAP files structures

struct TAPHeaderStruct
{
	unsigned char	TAPBlockType; // 0,1,2 or 3 (program, number array, character array, code files)
	unsigned char	TAPFileName[10];
	unsigned char	TAPCodeBlockLenH, TAPCodeBlockLenL;
	unsigned char	TAPParam1H,TAPParam1L;
	unsigned char	TAPParam2H,TAPParam2L;
	unsigned char	TAPChecksum;
};

// .SLT files structures

struct SltDataBlock
{	
	unsigned char	DataTypeHigh;
	unsigned char	DataTypeLow;
	unsigned char	DataIdentifierHigh;
	unsigned char	DataIdentifierLow;
	unsigned char	DataLengthHigh1;
	unsigned char	DataLengthHigh2;
	unsigned char	DataLengthLow1;
	unsigned char	DataLengthLow2;
};

struct SltDataBlockList
{	
	unsigned long	ulLevelDataFilePosition;
	unsigned int	uiLevelNumber;
	unsigned long	ulLevelDataLength;
	struct SltDataBlockList		*nextBlock;
};

#endif

